/*
   This is the main program. see README
   exact diagonalization of Heisenberg/Hubbard/tJ-systems
   
   ToDo: Simplify the program to make the code better readable!!!!

   startvec = \sum_j c^+_{i_j,s} ... |0> = 4u 2d 3u |0>    
   
   complex? phase factor later added during computation of H(i,j)
    multiply phase to base in a way that Hamiltonian and Vector is reell!?

   set parameters
   <OP> as pointer of OP-function
   static int (*a)(int b); o.ae.

   tU -> ...u..u. * ..d..d divide into up and down part (2 tables)   
   tJ -> ...uu... ...--d.d => (N over u)*(N-u over d)
   JJ -> ...
   
   Warn: operation x<<32 leads to errors for 32bit (no shift)
  
   result of (~0)>>1==~0 != (~0u)>>1  depends on compiler?
   
*/

#include <string.h>
#include <unistd.h> /* gethostname */
#include "block.h"
#include "spins.h"
#include "tbase.h"
#include "ansatz.h"
#include "pipe.h"
#include "vector.h"

//#ifdef CONFIG_MPI
//#include <mpi.h>   /* included by block.h */
//#endif

#ifndef TMPDIR        /* 2018-11 added for more flexibility */
#define TMPDIR "tmp/"  /* alternatives: "./" "../tmp/" "/tmp/" */
#endif

#ifdef CFG_CPUSET // switch on if sched_setaffinity will benefit
/* 2017-01: stackoverflow: use -D_GNU_SOURCE and __USE_GNU is set internaly */
#ifndef __USE_GNU
#  warning "add Makefile: DFLAGS+= -DCFG_CPUSET=1 -D_GNU_SOURCE"
#endif
/* 2017.01.21 sched.h needs __USE_GNU to define CPU_SETSIZE for icc16 */
# include <unistd.h>  // declare sleep()
# include <sched.h>   // declare sched_setaffinity() 
# ifndef _FEATURES_H
#  warning "_FEATURES_H is not defined, try top add #include <features.h>"
# endif
#endif

#include "version.h"

#include "spinstri.h"	/* spinstri.c::outtri() */
#include "mem.h"	/* mymalloc();myfree() */
#include "symmetry.h"	/* lattice symmetries */
#include "diag.h"
#include "myprint.h"	/* maple-Ausgabe */
#include "xval.h"	/* expectation values */
#include "h_file.h"	/* read/write Hamilton matrix on Htmp */
#include "error.h"	/* err_fatal */
#include "hilbert.h"
#include "modeldef.h"

/* IFz(x) produced a cpp-error on cray, so we changed it back 1998-148 */
/* ToDo: add PARAz for XXZ model */
#ifndef H_JZ        /* 2018-01-05 moved from .h to .c, IFz renamed to H_JZ */
#define H_JZ    1   /* 1 = JJ-model incl. Jz-term, 0 for faster XY-model */
#endif              /*     wjz[ij]*SizSjz  last 0-test? */

/* single-site-U-part may differ, some samples as f[.,o,x,3]=[empty,u,d,u+d]:
    (nu*nd)           [ 0, 0, 0, 1]   = default
    (nu-.5)(nd-.5)    [ 1,-1,-1, 1]/4 =(2nu-1)(2nd-1)/4=(4nu*nd-2nu-2nd+1)/4
    replace name by H_Uv1 H_Uv2 or similar
    2018-08-23 Ham2+3 simplifications, use sun+=(Ham2)*... instead if+endif
 */
#ifndef Ham2        /* results differ in U*(N/4-(nu+nd)/2) */
#define Ham2	0   /* 0 = U nu nd ,  1 = U(nu-.5)(nd-.5) */
#endif
#ifndef Ham3
#define Ham3	0   /* 1 = with JH-Oles-term cc and cpcp, better sym */
#endif 

/* 2-sites: JJ={-3/4,1/4}J JJ+Ham1={-1/2,1/2}J Jshift? old1801: Ham1 */
/* SiSj + (ham1-1)*ninj/4 ??? shift energies +1/4 */
/* J=4t^2/U (half filled tU U*(n-1)^2/2) */
/* Heff= -t^2/U(n1+n2-n1n2) +4t^2/U SiSj  for n=1 half-filled 2t^2/U shift */
/* 2018-01-05 rename Add34 by H_Jadd additional diagonal term = shift E */
#ifndef H_Jadd   /* ToDo replace Ham1 by H_Jadd=0.25? */
#define H_Jadd 0.0   /* SiSj+H_Jadd*ninj shifting energy, default is 0.0 */
		    /* may be +0.75 or -0.25 (or +0.25) is a good idea */
#endif

/* against egcs-2.90.29 980515 (egcs-1.0.3 release)
 * warning: initialization of non-const reference "class a_cplx &" from
 *    rvalue "a_cplx" */
/* ToDo: remove h_xypara+h_zpara, replaced by .wjz[] */
double  h_xypara=1.0;
double  h_zpara=1.0;	/* H=h_zpara*SizSjz+h_xypara*(SixSjx+SiySjy),
				 z-anisotropy Feb99, xy Jun99 */
int     algo;	/* choice of diagonalization + basis generation algorithm
        algo.bit0..2:  0=lanzcos,1=richter,2=wenzel2x2 
        algo.bit3:     8=generate_base_by_H(experimental newns)
        algo.bit4..5:  16=recursive_ns 32=(parallel)seq_linpack_scan
        algo.bit6:     64=MPI_Abort after l1-base-generation
*/
//int     chkpt_mode=0; /* 1=setup after crash possible, 2=continue old: save_mode */
int     checkpoint=0;   /* checkpoint for crash status 0=no, 6=latest */
int	pew;		/* number eigenvalue to check */
int     x_p;		/* output format for vector |x>   (0..3) */
ULong	n1;		/* Matrixsize , n2=Linsize=obsolete */
int     n1_bits;        /* bits needed to store n1, used for MPI packing */
double  h_nz=0;    /* raw number of nonzero matrix elements (statistics) */
/* 2017-02-15 replace variable number of fixed storage-block-length $HBLen
 *            by fixed number $ah_blks of variable storage blocks 
 *   this was needed to fix a unbalanced problem for partly stored matrix
 */
ULong node_ah_blks=0; /* for ah_nz[ah_blks], incl. 0-blks ...
      node_ah_blks      ... round up to mod pt_n and get_max_mpi */
ULong stored_ah_blks[B_NUM]; /* mem+dsk ToDo17: bad false sharing */
ULong min_stored_ah_blks=0;  /* mem+dsk min_ah_blks_of_all_threads */
ULong max_stored_ah_blks=0;  /* mem+dsk max_ah_blks_of_all_threads (balance) */
ULong min_Mstored_ah_blks=0; /* mem     min_ah_blks_of_all_threads */
/*    min_stored_ah_blks       replace min_b_sn1/AH 2017-10 */
int *ah_nz=NULL; /* malloc ah_nz[pt_i*ah_blks/pt_n+ah_blk], num nz-elements per AH blk */
/* thread locality lost? but its not the bottleneck? */
/* ToDo17: write/read round up to AH or 1K-multiple to avoid fragmentation */
/* ToDo17 ah_blks_mem, ah_blks_dsk num stored ah*ah_nz[] blks */
/* ToDo: 1st blk for estimation, than store mem mem dsk mem mem dsk ... */
//ULong   min_b_sn1; /* pseudo minimum stored n1 per block of all blocks (MPI) */
#ifndef CFG_FTLM   /* ~ Finite-Temperature Lanczos Method */
double ew[NEW];	   /* eingenvalues (Eigenwerte)  (max 4)	*/
#else 
double ew[MAX_ITR];
#endif
ULong  *l2b;		/* Hilfs-Tabellen storeh2 */
ULong opt_perf_test=0; /* 0: normal operation, reduce n1 to p.e. 1e6 */
int   opt_pack_nz=0; /* pack same nz to get symmetric off-diag-block size */

/* ToDo: norm2 remove it, not necessary (tiny speedup) but complex code */
// ATnorm2 *norm2=NULL;	/* vector idx->sym-idx, norm(sym-idx)	*/
// minfo_t m_n2 = {NULL,0,0,"tmp/n2.map",NULL}; /* norm2 companion */
// int base2ok=0;			/* flag: norm2 not yet inititialized */

// ToDo: 2017-01 add next 3 parameters
double opt_maxfile=0.0; /* 2017-01-22 maxfilesize for matrix on scratch */
double opt_maxmem=-1.0; /* 2017-01-22 added + reduce hfmax to disk only 
 * max_taskmem=-1 (default) tries to auto configure  or dont use?
 * max_taskmem=0  dont_use?
 * max_taskmem=1e6  mem/task in Bytes 1e6=1MB (reduced by buf+vec+cfg)
 *    long pages = sysconf(_SC_PHYS_PAGES);    // not standard, _SC_AVPHYS_PAGES
 *    long page_size = sysconf(_SC_PAGE_SIZE); // posix-1 unistd.h
 *  but needs the number of tasks per node (getenv(SLURM_MEMORY?))
 * usually you have constant tasks per nodes, so one number per machine
 * usually it ends with available_mem/cpu_core 
 * its mainly to store maximum of sparse matrix in memory, avoid recomputation
 * ToDo: but what about TASKS_PER_NODE>1 ? use ENV?
 *    or check MP_TASKS_PER_NODE=4 MP_NODES=31 lrzGarching ?
 *  set hmmax_per_node xor hmmax_total !?! ToDo per core!
 * ToDo2017: simplify hfmax=file_max to num blocks of AH*NZX
 */
ULong hmem_nzmax=0; /* derived from maxmem,AH,NZX,... 2017 */
ULong hdsk_nzmax=0; /* old hfmax = bad, new use hdsk_nzmax and hmem_nzmax
   max_num_hxy[]-elements on dsk, 0 = unlimited, see opt_maxmem */
/* ToDo: 2013-04 better use system limits? joblimit.vmem=... or ulimit -v ... */
 /* ULong maxmem; */ /* maximum memory for vectors -> no paging! removed 2013 */
/* Warning: false sharing = different threads write to the same cache line */
struct s_thread_memH {
  char  *ptr;  /* STOREH=0 pointer to allocated memory, set extern */
  size_t len;  /* bytes allocated to memH, extern, static after init */
//  size_t used; /* number of stored valid data in bytes, used <= len */
  size_t ofs;  /* offset pointing to next read/write memory block */
  ULong  blk;  /* block number handled, mainly debugging purpose */
  ULong  wrt;  /* block numbers stored (to memory and disk?) */
  int    full;  /* mem filled flag, added 2017-02-24, avoid rd/wr to rest
                   or ToDo: blk>=wrt ? ToDo: not 0|1 but mem-end? */
  char dummy_fill_cache_line[64-24]; /* avoid false cache sharing on cores */
} memH[B_NUM]; /* one struc per cpu-core */
int mopen(int blk, int wrt) { /* ToDo: ptr=0? len=0? */
  memH[blk].ofs=0; memH[blk].blk=0;               /* reset pointers */
  if (wrt) { memH[blk].wrt=0; memH[blk].full=0;
             // memH[blk].used=0;
           } /* reset stored */
  return(blk); }
int mclose(int blk) { /*memH[blk].ofs=memH[blk].len;*/ return(blk); }
/* names mread/mwrite conflict with lam-mpi, renamed in 2.33 */
/* pointer is at last block, mixed r/w not allowed */
/* return len on success, return 0 on failure */
/* output ...\n (not \n...), better looking for multitask/thread-output */
/* ToDo: output last blk node_n1/numAH too !? */
size_t memHwrite(void *from, size_t len, int blk) { /* blk = threadnum */
  // static int nout=0; // not thread safe, ToDo better use
  const int nout = memH[blk].blk;  // 2017-10-19
  const int tblk = mpi_i*pt_n + blk; /* total thread num for log2 output */ 
  /* output (static) pointer + size once to reduce output (diffs) of runs */
  if ((vvv&16) && nout==0 && ((tblk&(tblk-1))==0 || tblk+1==mpi_n*pt_n))
    printf("DBG.mHwr %3d.%-2d Hblk %10p sz= %9lu\n", mpi_i, blk,
      (void *)memH[blk].ptr, (ULong)memH[blk].len);
//  if (blk+1==pt_n) nout++;
  if ((vvv&16) && (nout&(nout-1))==0 
               && ((tblk&(tblk-1))==0 || tblk+1==mpi_n*pt_n))
    printf("DBG.mHwr %3d.%-2d Hblk %4lu @ %9lu l= %8lu\n", mpi_i, blk, 
      memH[blk].blk, (ULong)memH[blk].ofs, (ULong)len);
  if (!memH[blk].ptr) return 0; // ini malloc failed
  if (len>sizeof(struct thxy)) {ERR(9999);return 0;} // max size overflow?
  if (memH[blk].full) return 0; // do not fill with very small blks
  if (memH[blk].ofs+len > memH[blk].len) { // does not fit = full mem
    memH[blk].full=1;
    // if start with \n, 1st \n, but text not before next \n (srun -l 2017-09)
    if ((tblk&(tblk-1))==0 || tblk+1==mpi_n*pt_n)
      // 0 1 2 4 8 16 ... log2 output, matrix write
      printf("DBG.mHwr %3d.%-2d no space left at blk %5lu ofs %6lu %5.1f%%\n",
        mpi_i, blk, memH[blk].blk, (ULong)memH[blk].ofs,
        memH[blk].ofs*100./memH[blk].len); // ToDo: .len blk-specific?
        /* ToDo should show maxblk = mpi_n1/nAH here + %% */
    memH[blk].blk++; /* not stored, but counted for log2 debugging output */
    return 0; /* not stored, no space left */ }
  memcpy(memH[blk].ptr+memH[blk].ofs, from, len);
  memH[blk].ofs += len;   /* set pointer to the next free AHNZ-block */
  memH[blk].blk += 1; /* counts on read + write (blk counter) */
  memH[blk].wrt += 1; /* counts on write (fill state in variable blocks) */
//  if (memH[blk].used < memH[blk].ofs) /* ToDo: needed? memHofs is enough! */ 
//      memH[blk].used = memH[blk].ofs;
  return len; }
/* faster direct access without copying, called by h_read() */
/* ToDo17 needs len=ah_nz[blk]*8B as argument */
/* assuming serial sequenz of lengths $len is same as on write (ToDo hash+chk) */
/* until 2017-10-18 memHnext buggy, read empty very small data block if full */
void *memHnext(size_t len, int blk) { void *ptr=NULL; // blk=thread_num
  const int nout= memH[blk].blk;
  const int tblk = mpi_i*pt_n + blk; /* total thread num for log2 output */
  /* every task/thread have other num of AH-blocks stored, more small blks */
  if ((vvv&16) && nout==0 && ((tblk&(tblk-1))==0 || tblk+1==mpi_n*pt_n))
    printf("DBG.mHrd %3d.%-2d Hblk %10p full=%d writtenBLKs=%4lu\n",
     mpi_i, blk, (void *)memH[blk].ptr, memH[blk].full, memH[blk].wrt);
//  if (blk+1==pt_n) nout++; /* ToDo: use blk-num instead of nout */
  if ((vvv&16) && ((tblk&(tblk-1))==0 || tblk+1==mpi_n*pt_n)
     && ((nout&(nout-1))==0 || memH[blk].blk+1==memH[blk].wrt))
    /* %-2d = left bounded number */
    printf("DBG.mHrd %3d.%-2d Hblk= %4lu ahnz= %6d\n",
      mpi_i, blk,                    memH[blk].blk,
       ah_nz[blk*(node_ah_blks/pt_n)+memH[blk].blk]); // one array per node
      // ah_nz[0..node_ah_blks]
  if (memH[blk].ptr && memH[blk].blk < memH[blk].wrt) { /* 2017-10-27 */
    ptr= memH[blk].ptr +memH[blk].ofs;      // pointer to stored data
    memH[blk].ofs+=len;} // +Hsize shift offset to next data
  memH[blk].blk++;
  return ptr; }

int cfg_sisj;	/* option for correlations, bit0=SiSj bit1=SS bit2=ZiZj(?) */

tbase1 p_mask[NN][NN];		/* bit masks for parity()  */

/* max/min/sum hnz_blocks per thread (balance in storeh/geth_block) */
ULong bhnz_max, bhnz_min, /* used to output balance, ToDo which one? */
      bhnz_sum;           /* blocks of rows of nonzero matrix elements */
ULong mpi_traffic=0,     /* MPI statistics, data volume (/band_width) */
      mpi_xy_traffic[16*16]; 
      /* mpi_xy_traffic: MPI stat summed size pkts from/to task max16 */
int   mpi_syncs=0,       /* number of MPI transactions (*latency) */
      mpi_maxlen=0,      /* maximum len data block in units scplx/ulong */
      /* mpi statistics packets with packetsize up to: 1k 4k 16k 64k 256k 1M */
      mpi_stats_npkts[7]={0,0,0,0,0,0,0};   /* num packets size<1K<<(2i) */
/* old: mpi_z_pkts=0; */ /* number of zero blocks, wasted communications */

char nodename[22]="\0HOSTNAME............"; /* 2016-04-08 set global for err() */
/* ToDo2017: check num_same_nodename 
 *    or check MP_TASKS_PER_NODE=4 MP_NODES=31 lrzGarching */

/* local variables for pthreads (tasks), communication to/from tasks */
typedef struct {
  int num;     /* to identify itself (ro) */
  void *data;  /* to send data-block (set by thread 0, read by workers) */
  int  task;   /* master set 1 to run task + waits for 0, task sets 0 @end */
  void (*worker)(void *data); /* parallel working function */
  /* mutex for each thread? */
} main_pts;
volatile main_pts main_pt[B_NUM]; /* [un]locked by mutex1 */

#ifdef CONFIG_PTHREAD
/* send jobtype (task) to threads and restart them */
/* 2016-10 g++-6.2 + pthread + blas segfaults for stacksize 8K (ulimit -s) */
void parallel_start( int thread_i, void (*worker)(void *data), void *data ){
    pthread_mutex_lock(&pt_mutex1);
    // printf("DBG: send work to task %d\n", thread_i);
#pragma omp critical
{
    main_pt[thread_i].data=data;
    main_pt[thread_i].worker=(/*volatile*/ void (*)(void *))worker;
    main_pt[thread_i].task=1; /* signal worker thread i to run (last!) */
}
    /* ToDo: do next only once after storing all tasks
             or send signal only to the job */
    pthread_cond_broadcast(&pt_cond1); /* data send to threads */
    pthread_mutex_unlock(&pt_mutex1); /* redefined as dummys if _OPENMP */
#ifdef _OPENMP
#pragma omp flush (main_pt)
#endif
}

/* waiting for threads */
void parallel_wait( int thread_i ){
#ifndef _OPENMP
  pthread_mutex_lock(&pt_mutex1);
  while ( main_pt[thread_i].task )  /* volatile! */
    pthread_cond_wait(&pt_cond2,&pt_mutex1); /* unlock+wait_for_child+lock */
  pthread_mutex_unlock(&pt_mutex1);
#else
  while ( main_pt[thread_i].task )  /* volatile! */
    usleep(10000); /* ToDo  */
#endif
}
#else /* ~CFG_PTHREAD dummy parallel functions (serialized pthreads) */
void parallel_start( int thread_i, void (*worker)(void *data), void *data ){
  worker(data); }
void parallel_wait( int thread_i ){}
#endif /* CFG_PTHREAD */

// papi.h: PAPI_VENDOR_SICORTEX
// gcc412 -dM -E - </dev/null # __SICORTEX__ __MIPSEL _R4000 _MIPS_ARCH_5KF
// pathcc -dM -E - </dev/null # __SICORTEX__ __MIPSEL _R4000 _MIPS_ARCH_5KF
// gcc492 -dM -E - </dev/null #     ~        __MIPSEL _R4000 _MIPS_ARCH_5KF
#if defined(__SICORTEX__) || defined(_MIPS_ARCH)
/* 2017-02-24 obsolete, because own_IAlltoall used, its MPI-2 = emul Ia2a */
/* Alltoallv is 1.4x(600tasks) to 15x(5700tasks) slower on SiCortex2009 */
#define USE_Alltoallv_Emulation 1
#else
#define USE_Alltoallv_Emulation 0 /* better on IB.mpich3, IB.ompi18 */
#endif

#if CONFIG_MPI   /* 2017-02-01 */
#define BFACTOR 120l /* HBLen to send, BFACTOR=1+asymbuf for recv data in % */
/* lc40=4+36 k=1 AH=8  mpi*pt=2*2 maxx=9=2*4+1 but max_nzAH_recv= 11*AH */
/* lc40=4+36 k=1 AH=32 mpi*pt=2*2 maxx=9=2*4+1 but max_nzAH_recv= 10*AH */
/* for mpibuf_vy[B_NUM*(NZ_BUFSZ1+NZ_BUFSZ2)] */
#define MPI_NZ_BUFSZ1 (NZXMAX*NUM_AHEAD_LINES)         /* 100% send */
#define MPI_NZ_BUFSZ2 (NZXMAX*NUM_AHEAD_LINES*BFACTOR/100) /* 120% recv */
/* static (=OMP_SHARED) but each thread has its own area [B_NUM] */
/* ToDo: [B_NUM][...+64] to ensure cpus dont have same mem cached */
/* ToDo17: bsend,brecv should be B_NUM times size but for thread 0 only */
/* rename to mpibuf_,   this eats most buffmem (nm --size-sort -S spin) */
/* if to big may cause linker error:
 *    relocation truncated to fit: R_X86_64_PC32 */
scplx   nsend_vy[B_NUM*MPI_NZ_BUFSZ1], /* buffer <nHv> send per node */
        nrecv_vy[B_NUM*MPI_NZ_BUFSZ2]; /* buffer <nHv> recv */
XY_TYPE nsend_iy[B_NUM*MPI_NZ_BUFSZ1],  /* buffer cfg2idx send */
        nrecv_iy[B_NUM*MPI_NZ_BUFSZ2]; /* ??? jj=uint32, but max=4G/node */
tbase   nsend_cfg[B_NUM*MPI_NZ_BUFSZ1], /* send remote symcfg, big on big NN */
        nrecv_cfg[B_NUM*MPI_NZ_BUFSZ2]; /* used in nhm_line() only */
/* mpi_HB is a table of offset+size to data for each node/task */
/* MPI_MAX_NUM should be greater than a CPU cache line (block.h 8K) */
int   bsort_bck[B_NUM][MPI_NZ_BUFSZ1]; /* sort back index, per thread */
/* end of fat arrays */
/* MPI_MAX_NUM is very big 2017-01 8*50K*4B=1.6MB, BNUM+1 fix gcc492 issue */

/* we need a send for remote data (not symmetric if SH not compressed */
/* 2017.01.19 now all threads collect data to thread 0, which does mpi */
/*   packet size will be B_NUM times greater now,
 *   smaller buffers for max speed!? (ToDo: check v2.54 vs. v2.60) */
/* 2017.1.18 add shared space for thread0.mpi only, uint16: 64K*4B=256KB */
/* ToDo17 need to put arays to node_pipe[2] for pipelined communication */
  int   nsend_ofs[MPI_MAX_NUM+1]; /* offset v2.36, +[n]=size */
  int   nsend_cnt[MPI_MAX_NUM+1]; /* size per desttsk +[n]=overflow */
  int   nrecv_ofs[MPI_MAX_NUM+1]; /* offset  2017.1.14 */
  int   nrecv_cnt[MPI_MAX_NUM+1]; /* size per rcv_node 2017.1 */
  /* MPI_MAX_NUM is very big 2017-01 8*50K*4B=1.6MB, BNUM+1 o fix gcc492 issue  */
  /* no false sharing on big enough MPI_MAX_NUM (vs. CPU cache line) */
  /* partly master thread needs access to thread arrays */
  int   bsend_ofs[B_NUM+1][MPI_MAX_NUM+1]; /* offset v2.36, +[n]=size */
  int   bsend_of2[B_NUM][MPI_MAX_NUM+1]; /* help counter 2017.1.14 */
  int   bsend_cnt[B_NUM][MPI_MAX_NUM+1]; /* size per dest_node  */
/* ompi-1.7 MPI-3: Ialltoall,Ialltoallv
 * ompi-1.2 MPI-? Isend,Irecv,Wait,Alltoall,Alltoallv */
MPI_Status  mpi_status; /* used for My_MPI_Wait_Ia2a[v] */
MPI_Request a2a_request; /* used for MPI_Ialltoall[v] */
#if 1 && MPI_VERSION >= 3 /* MPI-3 has non-blocking collective operations */
/* #define My_MPI_Alltoallv(d1,c1,o1,t1,d2,c2,o2,t2,cm) \
              MPI_Alltoallv(d1,c1,o1,t1,d2,c2,o2,t2,cm)  */
#define My_MPI_Ialltoallv   MPI_Ialltoallv
#define My_MPI_Ialltoall    MPI_Ialltoall
#define My_MPI_Test_Ia2a    MPI_Test
#define My_MPI_Wait_Ia2a    MPI_Wait
#else
MPI_Request snd_request[MPI_MAX_NUM], /* only for intern emulation */
            rcv_request[MPI_MAX_NUM];
int My_MPI_Ialltoallv( /* better use MPI_IAlltoallv MPI-3.0 */
  const void *d1,const int *cnt1,const int *ofs1,MPI_Datatype t1,
  const void *d2,const int *cnt2,const int *ofs2,MPI_Datatype t2,
  MPI_Comm cm, MPI_Request *req)
{ int n1, n2, len1, len2, z1, z2, rc=0, id; void *p1, *p2;
  MPI_Type_size(t1, &z1); MPI_Type_size(t2, &z2); /* get data-size */
  for (id=0;id<mpi_n;id++){ /* pre2017-Alltoallv-emulation */
    n1=(mpi_i      +id)%mpi_n;  
    n2=(mpi_i+mpi_n-id)%mpi_n; /* sendto + rcvfrom-node */
    len1=cnt1[n1];      p1=&((char *)d1)[ ofs1[n1]*z1 ];
    len2=cnt2[n2];      p2=&((char *)d2)[ ofs2[n2]*z2 ];
    rc|=MPI_Irecv(p2, len2, t2, n2, 0, cm, &(rcv_request[n2]));
    rc|=MPI_Isend(p1, len1, t1, n1, 0, cm, &(snd_request[n1]));
//    rc|=MPI_Sendrecv(p1,len1,t1,n1,0,
//                     p2,len2,t2,n2,0,cm,&status);  // for node00..
  }
  return(rc); }
int My_MPI_Ialltoall( /* better use MPI_IAlltoall MPI-3.0 */
  const void *d1,const int len1,MPI_Datatype t1,
  const void *d2,const int len2,MPI_Datatype t2,MPI_Comm cm,MPI_Request *req)
{ int n1, n2, z1, z2, rc=0, id; void *p1, *p2;
  MPI_Type_size(t1, &z1); MPI_Type_size(t2, &z2); /* get data-size */
  /* correct? or use type MPI_BYTE only! */
  for (id=0;id<mpi_n;id++){ /* pre2017-Alltoallv-emulation */
    n1=(mpi_i      +id)%mpi_n;
    n2=(mpi_i+mpi_n-id)%mpi_n; /* sendto + recvfrom */
    p1=&((char *)d1)[ n1*z1 ];
    p2=&((char *)d2)[ n2*z2 ];
    rc|=MPI_Irecv(p2, len2, t2, n2, 0, cm, &(rcv_request[n2]));
    rc|=MPI_Isend(p1, len1, t1, n1, 0, cm, &(snd_request[n1]));
  } // for node00.. /* ToDo17: do not call self-sendrecv */
  return(rc); }
/* My-Wait needed for internal request-arrays */
int My_MPI_Wait_Ia2a(MPI_Request *req, MPI_Status *status)
{ int n1, n2, rc=0, id;
  for (id=0;id<mpi_n;id++){ /* wait in same order as A2a */
    n1=(mpi_i+id)%mpi_n;  n2=(mpi_i+mpi_n-id)%mpi_n; /* sendto + recvfrom */
    rc|=MPI_Wait(&(snd_request[n1]),status);
    rc|=MPI_Wait(&(rcv_request[n2]),status);
  } // for node00.. /* ToDo17: do not call self-sendrecv */
  return(rc); }
#endif
#endif /* CONFIG_MPI */

/* spins.h: inline double norm(double r){return(r*r);} */
char *sclock(long t){  /* 0..960s 16m..960m 16h... => 00:00:00 */
   static char s[64]; if (t==~0l) t=itime();
   char s2[24];
   /* this format is a mixture for simple parse and human read */
   /* this format is human readable but not simply enough to parse */
   if(t>4*60*60l)sprintf(s2,"%ldh",(t+1799)/3600); else
   if(t>   4*60l)sprintf(s2,"%ldm",(t+29)  /  60); else sprintf(s2,"%lds",t); 
   /* 0.1-9999.9min is best compromize, 5s..150h=6d */
   /* a recursive call like sprintf(s,"%s",s) is dangerous */
   sprintf(s,"%6.2f =%s",t/60.,s2);
   return(s);}

char *outc(cplx r){  /* output [complex] float number */    
    static char s1[100];
#if VecType & 4
    double phase=arg(r)/M_PI; if (phase<0) phase+=2;
    /* sprintf(s1,"(%g,%g)",real(r),imag(r)); -> |r|*exp(phase*I*Pi) */
    if (Cabs(r)<0.0001 || Cabs(r)>9.99999)
         sprintf(s1,"%12.5e*phase(%5.3f)",Cabs(r), phase);
    else sprintf(s1,"%12.9f*phase(%5.3f)",Cabs(r), phase);
#else /* % 12.9f=-0.123456789 vs. % 12.5e=-1.12345e-09 */
    if (fabs(r)<0.0001 || fabs(r)>9.99999)
         sprintf(s1,"% 12.5e",D(r));
    else sprintf(s1,"% 12.9f",D(r)); /* 12.9f or 12.5e */
#endif
    return(s1);}

void nl(FILE *f){ fprintf(f,"\n");return;}
/* output eigenvalues */
void outew(FILE *f, const double *ew, const int n_ew){
   int k;
   for (k=0;k<n_ew;k++) fprintf(f," %12.8f", ew[k]);
   return;
}


/* ---------------------------------------------------- *
 *      statistics about hamiltonien  (sym,num_nz,...)	*
 *	-	-	-	-	-	-	
 *  double counts to 2^53=9e15 (54bit+10bit) exp e+307(9bit)
 *  
 */
struct tstatH {    
  /* max. and 2nd max number nz elements of a line (weight to?) */
  int nzxmin,  nzxmax,  nzxmin2,  nzxmax2,  /* nz per line */
      nzxminw, nzxmaxw, nzxmin2w, nzxmax2w; /* density or weight of nzx */
  cplx sum12, sum21;  /* sum-upperright,lowerleft r to detect errors */
/* dbl hh12, hh21;  +=<r|r> removed, some elements of H may not be summed
# s=1 N=27 triangle 53+1 n1=2 2012-01-26
#   min,max zaehlt nicht summierte Elemente, deshalb sum += r**2 != 0
# sum_h12   hh12 :  1.414213538*phase(1.000)  3.99999960e+00
# sum_h21   hh21 :  1.414213538*phase(1.000)  1.99999980e+00
# sum h21-h12^*  :  0.00000e+00*phase(0.000) -1.99999980e+00 ToDo: wegnehmen! 
# Stoerungstheorie (Perturbation theory)?
#   1. lowest diag, 2. follow biggest non-diag to nearest diag, ... ?? ToDo
 */
  double traceH,       /* Trace of H for manual error detection */
         minH, minH12, /* 2018-08 minH + minAbsH12 */
         maxH, maxH12; /* maximum absolute diagonal and off-diagonal element */
  /* statistics for data transfers within and between nodes */
  /* MPI_Allreduce dont allow LONG_LONG */
  double hnz_nd,  /* number of non-diagonal elements (v&4 only?) */
         hnz_d,	  /* number of diagonal elements     */
         hnz_local_thread,
         hnz_inter_thread,
         hnz_inter_node; /* ToDo: [nodes] for min/max/mean%??? */
  }  statH[B_NUM]; /* ToDo: -> statH[B_NUM] distinct cache lines */

/* ToDo: move to spins_dbg.c # -Os */
void statH_ini(){ /* reset statistic counters */
 int i;
 for(i=0;i<pt_n;i++){ /* thread safe */
   statH[i].nzxmax=0;
   statH[i].nzxmin=999999;
   statH[i].nzxmax2=0;
   statH[i].nzxmin2=999999;
   statH[i].nzxmaxw=0;
   statH[i].nzxminw=0;
   statH[i].nzxmax2w=0;
   statH[i].nzxmin2w=0;
   statH[i].sum12=C(0,0);
   statH[i].sum21=C(0,0);
   statH[i].traceH=0.0;
   statH[i].minH=  999999.0;
   statH[i].minH12=999999.0;
   statH[i].maxH=  0.0;
   statH[i].maxH12=0.0;
   statH[i].hnz_d=0;
   statH[i].hnz_nd=0;
   statH[i].hnz_local_thread=0;
   statH[i].hnz_inter_thread=0;
   statH[i].hnz_inter_node  =0;
 }
}
/* called by hamilton_puth_line() on each line for statistic analysis
 * also check for potential errors
 * ToDo: replace by tbx=bx%pt_n (thread block) and gbx=tbx+mpi_i*pt_n (global block)
 *            or    tbx=bx & (1<<pt_n_bits)-1;
 * called per lenAHblk
 * ToDo: move to spins_dbg.c # -Os
 */
void add_statH(thbuf *hbuf){
  double rr; char sbuf2[SZccfg];
  ULong xy, x, y;
  int node_x, node_y, thread_x, thread_y, ii, n_old; /* was bx, by */
  cplx r;
  thread_x=hbuf->block;
  node_x=mpi_i;
  if (thread_x>=pt_n)     F1("thread_x>=pt_n thread_x=%d", thread_x);
  if (!hbuf->lenAHblk) return;
  //  x_ofs=EOLen= n*len + (len-1),   x=x_start= n*len = startAH_line_i
  x= hbuf->ofsAHblk - hbuf->ofsAHblk % NUM_AHEAD_LINES; // or (x/len)*len
  if (x != hbuf->idxAHblk * NUM_AHEAD_LINES) ERR(9999); // 2017-02-17
  // not hermitian???
  // ToDo: check x > n1+lenAHblk !?
  if (x>=n1 || x>=node_n1) return; // happens if i+ahead > n1
// if (x>=n1) F1("x>=n1 x=%.0f",1.*x)
// if (x>=node_n1) F1("x>=node_n1 x=%.0f",1.*x)
  if (x>=b_ofs[thread_x+1]) {
     static int reported=0;
     if (!reported)
       printf("\n %d.%d: x>=b_len   x=%.0f (may happen if i+AH>=n1)",
          mpi_i, thread_x, 1.*x);
     reported=1;
     return; // happens on extra calls for balancing???
  }
  x+=mpi_ofs[node_x]+b_ofs[thread_x]; /* x=0..n1-1, ToDo: remove? */
  if ((vvv&64) && mpi_i==0 && x<x_max)
    printf("\nvvv&64 add_statH.addh x_max=%ld hbuf.n=%d n1=%ld:",
           x_max, hbuf->num_nzAH, (long)n1);

  /* synchronize mpi output? */
  n_old=0;
  for (ii=0;ii<hbuf->num_nzAH;ii++) {  /* for each hnz element, eats CPU, vvv&4 only? */
    static int err[B_NUM]={0};
    xy     = hbuf->jj[ii];   /* 0..dest_node_n1-1 */
    node_y = hbuf->abj[ii]; /* destination node 0..mpi_n-1 */
    if (node_y>=mpi_n) ERR(9999); /* never happens? ~init! */
    r      = hbuf->rr[ii];
    y      = xy & ~XY_FLAG;
    for (thread_y=0; thread_y<pt_n; thread_y++)
      if (y<b_ofs[thread_y+1]) break; /* bad: we need remote table here */
    y     += mpi_ofs[node_y]; /* 0..n1 get total idx from node_local_idx */

    if ( // 1 ||   /* debug */
         // err[hbuf->block] || /* debug error followers too */
       y>=n1 || node_y>=mpi_n ||
       y>=mpi_ofs[node_y+1] ||
       y< mpi_ofs[node_y  ]) {
      if (err[hbuf->block]<16) { err[hbuf->block]++;
        if (mpi_i*pt_n+hbuf->block>0 && err[hbuf->block]<=1)
           usleep((mpi_i*pt_n+hbuf->block)*20000); // 20ms
  // ToDo17 on Err output lenAH vs. NAH, ofsAH as multiple of AH
  //             x/y as offsets + b_ofs + node_ofs vs. max
        /* better give more dbg info for that bad case */
        /* p.e. last thread n1-segment may be shorter than others */
  printf("\nERR.statH.y %2d.%2d lenAHblk=%5d     AH=%5d (same?)",
          mpi_i, hbuf->block, hbuf->lenAHblk, NUM_AHEAD_LINES);
  printf("\nERR.statH.y %2d.%2d ofsAHblk=%5lu = %5lu*%d+%5lu (ofs==len-1)?",
     mpi_i, hbuf->block, hbuf->ofsAHblk,  
                         hbuf->ofsAHblk/NUM_AHEAD_LINES,NUM_AHEAD_LINES,
                         hbuf->ofsAHblk%NUM_AHEAD_LINES);
  printf("\nERR.statH.y %2d.%2d nz_i=%5d of nzx*AH= %5d= %5.2f*%d",
     mpi_i, hbuf->block, ii, hbuf->num_nzAH,
     hbuf->num_nzAH*1./NUM_AHEAD_LINES,NUM_AHEAD_LINES);
  printf("\nERR.statH.y %2d.%2d nz_i=%5d cfgj=%s sgn=%+.0f n2=%2d j=%6lu",
     mpi_i, hbuf->block, ii, ccfg2(hbuf->acj[ii],sbuf2),
     D(hbuf->asj[ii]), (int)(hbuf->an2[ii]), (ULong)hbuf->jj[ii]);
  printf("\nERR.statH.y %2d.%2d nz_x=%5d cfgx=%s sgn=%+.0f n2=%2d x=%6lu",
     mpi_i, hbuf->block, n_old, ccfg2(hbuf->acj[n_old],sbuf2),
     D(hbuf->asj[n_old]), (int)(hbuf->an2[n_old]), (ULong)hbuf->jj[n_old]);
  printf("\nERR.statH.y %2d.%2d nz_i=%5d rr= % .2e %+.2ei (zero?)",
     mpi_i, hbuf->block, ii, D(r), D(Im(r)));
  printf("\nERR.statH.y %2d.%2d max0 %2d.%2d.%6lu per node",
     mpi_i, hbuf->block, mpi_i, 0, mpi_ofs[1]-mpi_ofs[0]);
  printf("\nERR.statH.y %2d.%2d maxX %2d.%2d.%6lu per node",
     mpi_i, hbuf->block, mpi_i,  hbuf->block, node_n1);
  printf("\nERR.statH.y %2d.%2d maxX %2d.%2d.%6lu mpi_ofs.diff",
     mpi_i, hbuf->block, mpi_i,  hbuf->block,
     (mpi_ofs[hbuf->block+1]-mpi_ofs[hbuf->block]));
  printf("\nERR.statH.y %2d.%2d  x=  %2d.%2d.%6lu +mpi_ofs",
      mpi_i, hbuf->block, node_x, thread_x,
      x-mpi_ofs[node_x]/*-b_ofs[thread_x]*/);
  printf("\nERR.statH.y %2d.%2d  y=  %2d.%2d.%6lu +mpi_ofs",
      mpi_i, hbuf->block, node_y, thread_y, 
      y-mpi_ofs[node_y]/*-b_ofs[thread_y]*/);
      }
    }

    if ((vvv&64) && (x<x_max || x+1==n1) && (Norm(r)>1e-5 || x==y)){
      /* show 1..x_max H-lines, 2015 skip zero off-diag elements */
      /* printf("\n x=%8lu y=%8lu h_xy=%s",x,y,outc(r)); --- old --- */
      /* try to find exact expressions for maple/maxima input */ 
      if (ii==0) usleep((mpi_i*pt_n+hbuf->block+1)*20000); /* kind of sync */
      // 2018-08-23 add skip + last Hline
      if (x_max+1<n1 && x+1==n1 && y==x) printf("\n ... skip output");
      printf("\n addh(%5lu,%5lu, ",x,y);
      outmpl(stdout,D(r));			/* +(a/b)^(1/2) */
#if VecType & 4
      printf(","); outmpl(stdout,Im(r));	/* +(a/b)^(1/2) */
#endif
      printf("):");				/* for maple */
    }
    if (thread_y==thread_x && node_x==node_y)
                          statH[thread_x].hnz_local_thread++; else
    if (node_y==node_x)   statH[thread_x].hnz_inter_thread++; else
                          statH[thread_x].hnz_inter_node++;
    if (x> y) statH[thread_x].sum12+=r;	     /* summe of half H's */
    if (x< y) statH[thread_x].sum21+=r;	     /* should be conjugated value */
    if (x==y) statH[thread_x].traceH+=Re(r); /* should be base invariant */
    rr=Cabs(r);				/* cplx => real != sqrt(Norm()) */
    // if (n1<32 && vvv && x==y && rr)  /* 2015-12 */
    //  printf("\nDBG: Hdiag %2d %s", (int)x, outc(r));
    if (!(rr<0.8e38)) { /* maxF32=3.4e38 2016-01 p.e. float=(double)1e50 highS */
      /* if using f32 + Tnorm2=f64 (big-s1-systems, try to use double only
       *    found it on N=12 S=5/2 12+48 n1= 25751  2016-01 */
      /* no idea how to solve it, for N=12 S=5/2 */
      printf("\n statH x=%5lu hbuf.ii=%d y=%5lu r=%e",x,ii,y,D(r));
      E0(                 "rr>1e38  (f32 overflow)");
      if (!(rr<1e300)) F0("rr>1e300 (f64 overflow)");
#if  (VecType & 4) == 0  /* float32 */
      F0(" have f32 only, please try f64 or other solutions");
#endif
    } /* (!(rr<1e36)) */
    /* if(rr) makes only sence if nz-elements are packed,
     * ToDo17: add max_nzx-loop checking y is not already counted for x
     *      ... v&4 only (eats time?)? */
    if (x==y && statH[thread_x].minH>rr)  /* min diagonal element incl. 0 */
                statH[thread_x].minH=rr;
    if (x==y && statH[thread_x].maxH<rr)  /* max diagonal element */
                statH[thread_x].maxH=rr;
    if (rr) {  /* ToDo17: why only nonzero of nz-Elements <=> nz-storage */
      if (x!=y && statH[thread_x].minH12>rr) /* min off-diagonal element */
                  statH[thread_x].minH12=rr;
      if (x!=y && statH[thread_x].maxH12<rr) /* max off-diagonal element */
                  statH[thread_x].maxH12=rr;
      if (x==y)   statH[thread_x].hnz_d++;   /* count nz diagonal elements */
      else        statH[thread_x].hnz_nd++;  /* count off-diagonal elements */
    }
    if (xy&XY_FLAG) { /* EOL, start next matrix line */
//printf("\nDBG nzx=%lu n_old=%lu ii=%d x=%lu",(ULong)ii+1-n_old,n_old,ii,x);
      if (statH[thread_x].nzxmax==ii+1-n_old) {
          statH[thread_x].nzxmaxw++; }
      if (statH[thread_x].nzxmin==ii+1-n_old) {
          statH[thread_x].nzxminw++; }
      if (statH[thread_x].nzxmax2==ii+1-n_old) {
          statH[thread_x].nzxmax2w++; }
      if (statH[thread_x].nzxmin2==ii+1-n_old) {
          statH[thread_x].nzxmin2w++; }
      if (statH[thread_x].nzxmax<ii+1-n_old) {
          statH[thread_x].nzxmax2= statH[thread_x].nzxmax;
          statH[thread_x].nzxmax2w=statH[thread_x].nzxmaxw; 
          statH[thread_x].nzxmax=ii+1-n_old;
          statH[thread_x].nzxmaxw=1; }
      if (statH[thread_x].nzxmin>ii+1-n_old) {
          statH[thread_x].nzxmin2= statH[thread_x].nzxmin;
          statH[thread_x].nzxmin2w=statH[thread_x].nzxminw; 
          statH[thread_x].nzxmin=ii+1-n_old;
          statH[thread_x].nzxminw=1; }
      if (statH[thread_x].nzxmax> ii+1-n_old
       && statH[thread_x].nzxmax2<ii+1-n_old ) {
          statH[thread_x].nzxmax2=ii+1-n_old;
          statH[thread_x].nzxmax2w=1; }
      if (statH[thread_x].nzxmin< ii+1-n_old
       && statH[thread_x].nzxmin2>ii+1-n_old ) {
          statH[thread_x].nzxmin2=ii+1-n_old;
          statH[thread_x].nzxmin2w=1; }
      x++; n_old=ii+1;
    }
  }
  return;
}

/* 2015-05-28 bug 11.05.15 n1=3 nzx=3.67
   sum_h12/n1   :  1.85037e-17*phase(0.500)
   sum_h21/n1   :  1.86159e-17*phase(0.535)
   sum_h21-12^* :  1.11191e-16*phase(0.518) not hermitian + abort
   fix: hermitian + return error if not hermitian (old void)
   return: 0=OK, 1=(n1==0) 2=not_hermitian 
 */

/* ToDo: move to spins_dbg.c # -Os */
int statH_out(){		/* output statistics vvv&2 */
  int i;
  if (n1==0) return 1;
  for(i=1;i<pt_n;i++){      /* sum all threads to statH[0] */
    statH[0].hnz_local_thread += statH[i].hnz_local_thread;
    statH[0].hnz_inter_thread += statH[i].hnz_inter_thread;
    statH[0].hnz_inter_node   += statH[i].hnz_inter_node;
    statH[0].sum12  += statH[i].sum12;
    statH[0].sum21  += statH[i].sum21;
    statH[0].traceH += statH[i].traceH;
    if (statH[0].maxH    > statH[i].minH)
        statH[0].maxH    = statH[i].minH;
    if (statH[0].maxH12  > statH[i].minH12)
        statH[0].maxH12  = statH[i].minH12;
    if (statH[0].maxH    < statH[i].maxH)
        statH[0].maxH    = statH[i].maxH;
    if (statH[0].maxH12  < statH[i].maxH12)
        statH[0].maxH12  = statH[i].maxH12;
    statH[0].hnz_nd   += statH[i].hnz_nd;
    statH[0].hnz_d    += statH[i].hnz_d;
    if (statH[0].nzxmax  == statH[i].nzxmax)
        statH[0].nzxmaxw += statH[i].nzxmaxw;
    if (statH[0].nzxmin  == statH[i].nzxmin)
        statH[0].nzxminw += statH[i].nzxminw;
    if (statH[0].nzxmax  <  statH[i].nzxmax) {
        statH[0].nzxmax  =  statH[i].nzxmax;
        statH[0].nzxmaxw =  statH[i].nzxmaxw; }
    if (statH[0].nzxmin  >  statH[i].nzxmin) {
        statH[0].nzxmin  =  statH[i].nzxmin;
        statH[0].nzxminw =  statH[i].nzxminw; }
  }
#ifdef CONFIG_MPI
  { double add[7], sum[7];
    add[0]=0; add[1]=0; // unused 2012-01-26
    add[2]=statH[0].traceH;
    add[3]=Re(statH[0].sum12); add[4]=Re(statH[0].sum21);
    add[5]=Im(statH[0].sum12); add[6]=Im(statH[0].sum21);
    MPI_Allreduce(add, sum, 7, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    statH[0].traceH=sum[2];
    statH[0].sum12 =C(sum[3],sum[5]);
    statH[0].sum21 =C(sum[4],sum[6]); }
  { double add[7], sum[7];
    add[0]=statH[0].hnz_local_thread;
    add[1]=statH[0].hnz_inter_thread;
    add[2]=statH[0].hnz_inter_node;
    add[3]=statH[0].hnz_nd;  add[4]=statH[0].hnz_d;
    MPI_Allreduce(add, sum, 5, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    statH[0].hnz_local_thread=sum[0];
    statH[0].hnz_inter_thread=sum[1];
    statH[0].hnz_inter_node  =sum[2];
    statH[0].hnz_nd=sum[3];  statH[0].hnz_d =sum[4]; }  
  { double add[2], sum[2];
    add[0]=statH[0].maxH;   add[1]=statH[0].maxH12;
    MPI_Allreduce(add, sum, 2, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
    statH[0].maxH=  sum[0]; statH[0].maxH12=sum[1]; }  
  { double add[2], sum[2];
    add[0]=statH[0].minH;   add[1]=statH[0].minH12;
    MPI_Allreduce(add, sum, 2, MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD);
    statH[0].minH=  sum[0]; statH[0].minH12=sum[1]; }  
  { int add[7], sum[7];
    add[1]=statH[0].nzxminw; add[2]=statH[0].nzxmaxw; // num min/max = weight
    add[0]=statH[0].nzxmax;
    MPI_Allreduce(add, sum, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
    if (statH[0].nzxmax!=sum[0]) add[2]=0;
    statH[0].nzxmax=sum[0]; add[0]=statH[0].nzxmin;
    MPI_Allreduce(add, sum, 1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
    if (statH[0].nzxmin!=sum[0]) add[1]=0;
    statH[0].nzxmin=sum[0]; add[0]=0; /* unused */
    MPI_Allreduce(add, sum, 3, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
    statH[0].nzxminw=sum[1]; statH[0].nzxmaxw=sum[2]; }
#endif
  /* MPI_ToDo: sum over all nodes, may fail for multithreads */
  if (!mpi_i) {
    double sum_hnz=statH[0].hnz_local_thread
                  +statH[0].hnz_inter_thread
                  +statH[0].hnz_inter_node  ;
    printf("\n vvv&2 statistics for sparse matrix H:");
    printf("\n hnz transfer local.1,pthread.%d,mpi.%d= %7.1e %7.1e %7.1e",
      pt_n, mpi_n, statH[0].hnz_local_thread*1.,
                   statH[0].hnz_inter_thread*1., 
                   statH[0].hnz_inter_node  *1.);
    printf("\n hnz transfer local.1,pthread.%d,mpi.%d= %6.0f%% %6.0f%% %6.0f%%",
      pt_n, mpi_n, statH[0].hnz_local_thread*100./sum_hnz,
                   statH[0].hnz_inter_thread*100./sum_hnz, 
                   statH[0].hnz_inter_node  *100./sum_hnz);
    /* printf("\n hnz_ndiag>eps : %9.0f",statH[0].hnz_nd); */
    /* printf("\n hnz_diag      : %9.0f",statH[0].hnz_d); */
    /* ToDo: counting long.numlines[nzx=1..1+min(nu,nd)*2*Nw/N]++ is cheaper 
     *    to compute than min2,max2 or more sorting (use [... 1+Nw]) */
    printf("\n hnz_x_min,w  : %9d %14d",statH[0].nzxmin,statH[0].nzxminw);
    /* min2,max2 correct for mpi=1 pt=1 only */
if (mpi_n==1 && pt_n==1) {
    printf("\n hnz_x_min2   : %9d %14d",statH[0].nzxmin2,statH[0].nzxmin2w);
    printf("\n hnz_x_max2   : %9d %14d",statH[0].nzxmax2,statH[0].nzxmax2w);
}
    printf("\n hnz_x_max,w  : %9d %14d  (diag + off-diag)",
                                          statH[0].nzxmax,statH[0].nzxmaxw);
    /* v2.27 x_max ca. min(u,d)*nw+1?, x_min ca. minAO(u,d)+1? */
    /* after-sym-diagonal counted packed, reduces nz-num, change back!? */
    /* raw_off_diag may be diag after symmetrized and are packed 
     * ... but a zero element is left, eating storage and some CPU */
    printf("\n hnz/n1 * n1  : %9.3f * %12lu  (diagonal packed + min_nzx=1)",
                                     1+statH[0].hnz_nd*1./n1, n1);
    printf("\n hnz/n1 * n1  : %9.3f * %12lu  (raw not packed)",
                                            h_nz*1./n1, n1);
    /* 2014-06 trace normalized to mean eigenvalue, for a1,b1 range check? */
    printf("\n minHdiag     : %12.9f", statH[0].minH);
    printf("\n minH12       : %12.9f", statH[0].minH12);
    printf("\n maxHdiag     : %12.9f", statH[0].maxH);
    printf("\n maxH12       : %12.9f", statH[0].maxH12);
    printf("\n trace_H/n1   : %12.9f", statH[0].traceH/n1);
    printf("\n sum_h12/n1   : %s", outc(statH[0].sum12/n1));
    printf("\n sum_h21/n1   : %s", outc(statH[0].sum21/n1));
    /* debugging: */
    printf("\n sum_h21-12^* : %s", outc(statH[0].sum21-Conj(statH[0].sum12)));
    /* if its not hermitian, convergation is not ensured */
    /* 2015-05-28 +1 against near zero rounding errors */
    if(    Cabs(statH[0].sum12 -Conj(statH[0].sum21))
    >1e-8*(Cabs(statH[0].sum12)+Cabs(statH[0].sum21)+1))E0("H not hermitian");
    if(    Cabs(statH[0].sum12 -Conj(statH[0].sum21))
    >1e-2*(Cabs(statH[0].sum12)+Cabs(statH[0].sum21)+1)){E0("assume bad bug");
      return 2;}  /* or perf_test, fixable? */
  }
  return 0;
}
/* --------------------------------------------------------------- */  




#ifndef NUM_AHEAD_LINES
#define NUM_AHEAD_LINES 1  /* minimum, bigger is better, nhm_buffer_lines */
#endif
/* see spins.h   thread buffer:
   typedef struct shbuf { ... nhn[NAH],el[NZX*NAH]} thbuf; */
/* thbuf *buf1; buf1=alloc_hbuf() ... reset_hbuf(buf1) ... free_hbuf(buf1) */
int reset_hbuf(thbuf *buf){
  if (!buf) {ERR(9502); return -1;} /* must be alloc before reset */
  buf->num_nzAH=0; /* ini num_nz_elements_in_buf, returned/used by nhv */
  buf->nzAH_overflow=0; /* 2017-01-31 ERR(9200) debug */
  buf->max_nzAH_overflow=0; /* 2017-01-31 ERR(9200) debug, total_max */
  buf->max_nzAH_recv=0;     /* 2017-02-21 ERR(9200) debug, total_max */
  buf->lenAHblk=NUM_AHEAD_LINES; /* set before nhv called */ 
  buf->block=-1;  /* line belongs to block-nr, -1 means unused, before nhv */
  buf->v0=(scplx *)NULL; /* store/stat or nHn, set before nhv called */
  buf->v1=(scplx *)NULL; /* v1+=H*v0 if !0, set by the nhv caller */
//  buf->sci.o=0;   /* option: 0=fullbase,1=minsymcfg,2=isminsymcfg */
//  buf->scj.o=0;
//  buf->sci.num=0;
//  buf->scj.num=0;
  buf->rx=0.0;  /* <nHv>, used in xval.c to compute <v0|OP|v0> */
  /* nhm_line DBG needs initialization 2013-02 valgrind */
  {int i;for(i=0;i<NUM_AHEAD_LINES;i++)buf->nhv[i]=buf->nhn[i]=0;}
  return 0;
}

thbuf *alloc_hbuf(){
  thbuf *buf;
  /* ToDo: dynamic mem alloc for hbuf->el if needed, or MPIbuf pt0 only */
  buf=(thbuf *)mymalloc( sizeof(thbuf), NULL); /* no msg,  */
  if (!buf) {
    printf("\nERROR alloc_hbuf: malloc(thbuf=%.2e) failed, mpi=%d",
            1.*sizeof(thbuf), mpi_i);
    return 0;
  }
#if 0
  buf->sci.ib= (tbase *)NULL;  /* force SIGSEGFAULT if [] accessed */
  buf->sci.is=(tphase *)NULL;
  buf->scj.ib= (tbase *)NULL;  /* force SIGSEGFAULT if [] accessed */
  buf->scj.is=(tphase *)NULL;
  /* really needed? */
  buf->scj.ib= (tbase *)malloc(sizeof(tbase) *MaxSBase);
  buf->scj.is=(tphase *)malloc(sizeof(tphase)*MaxSBase);
#endif
  reset_hbuf(buf); /* set initial values */
  buf->nn = nn; /* local copy */
  buf->psym = &sym; /* pointer to global sym */
#if defined(_MIPS_ARCH) && defined(CONFIG_PTHREAD) /* SiCortex? bad multithread */
  buf->lsym = sym;          /* copy to local buffer */
  buf->psym = &(buf->lsym); /* set pointer to (thread-) local buffer */
#endif
  return buf;
}

void *free_hbuf( thbuf *hbuf ){
#if 0
  if (hbuf->sci.ib) free(hbuf->sci.ib); /* avoid memory leak */
  if (hbuf->sci.is) free(hbuf->sci.is);
#endif
  return myfree( (void *) hbuf, NULL );
}


/* store elements of one H-line to the HBlock files on disk 
 * input: hbuf.i(b_ofs[blk]...,>node_ofs) + hbuf.n + hbuf.jj+abj+rr[0..n-1]
 * output: H-file + Statistics
 * called by _nhv if m1&2 = storeh, 
 * ToDo: return !0 if not written, for output write/store fail at XX% 2016-01
    or do we have ofs in hbuf to output ofs/n1 as % ???
 */
void hamilton_puth_line(thbuf *hbuf){ /* renamed from nhmline v2.26 */
  int ii, blk, by;
  ULong xy; cplx r; HR_TYPE rr; /* h_put-variables */
  blk=hbuf->block; /* thread_id */
  /* b_hnz - number of hamilton-elements per compute block in thread */
  /* hdsk_nzmax - maximum file space for H per task!? */
#if DBG_NHM
  if(vvv&16)
  printf("\nDBG.puth  %3d.%-2d bofs,iAH,nzAH,nzSZ= %4lu %4lu %4d %4lu",
    mpi_i,blk,blk*node_ah_blks/pt_n, hbuf->idxAHblk,
    hbuf->num_nzAH, (ULong)(hbuf->num_nzAH*(size_t)Hsize/HBLen));
#endif
  if ((!memH[blk].full)  //  2017-01-24 fix mem full, mem+dsk
  || ((b_hnz[blk]+hbuf->num_nzAH < hdsk_nzmax+hmem_nzmax || opt_maxfile<0.0)
                                              && !h_error)){
    for(ii=0;ii<hbuf->num_nzAH;ii++) {
      /* store a matrix element (xy=y_idx,r=Hij) into the file buffer b_hbuf
       * and write full buffer to the file; keywords: hput puth put_h h_put */
      xy=hbuf->jj[ii]; /* 0..(b_len[by]-1)[+XY_FLAG] checked in statH */
      by=hbuf->abj[ii];  /* 0..(pt_n*mpi_n-1) */
      r =hbuf->rr[ii];
#if  DBG_NHM
      if (by>=mpi_n) ERR(9999); /* 2016-04 ~init abj */
//      if ((xy&~XY_FLAG)>=node_n1) {sleep(4);ERR(9999);}
#endif
#if HRMAX>0    /* use a 8-16bit index table to double to safe disk space */
      { int j; /* get hr_index or store as new hr_index else */
        /* Note: Norm(hr-r)<1e-13 is more compact than hr==r, no speed down */
        for(j=0;j<hr_len[blk];j++) if(Norm(hr[blk][j]-r)<=1e-13) break;
        if (j==hr_len[blk]) { /* not found */
           hr[blk][j]=r; hr_len[blk]++;
           if(hr_len[blk]>=HRMAX) err_fulltable(); /* OVL */ }
        rr=(HR_TYPE)j;		/* store index to value */ }
#else    
      rr=M_d2f(r);		/* store value direct, ToDo: d2f needed? ToDo */
#endif    
      /* blk is threadnum to ensure cpu local memory,
       *   struct thxy b_hbuf defined in block.c and used in h_file.c (not good) */
      if (b_bnz[blk]>=HBLen) F2("OO b_hbuf blk=%d i=%d", blk, b_bnz[blk]);
      b_hbuf[blk].nz[ b_bnz[blk] ].rr=rr;
      b_hbuf[blk].nz[ b_bnz[blk] ].by=(BY_TYPE)by;
      b_hbuf[blk].nz[ b_bnz[blk] ].xy=(XY_TYPE)xy;
      /* store in Hblock-buffer, ToDo: store only complete lines? */
      b_bnz[blk]++; /* count buffered elements, store at buffer too? */
      if (xy&XY_FLAG) b_bn1[blk]++; 
/* ToDo 2014-07: better fill rest with zeros (dummy) 
     and store buf if next line will not fit to the buffer or last line? 
     usually HBLen=32k(*8) nline=13-29=25 possible-overhead<29/32k
 */  // first step 
    } /* for ii = 0 .. hbuf->num_nzAH */
    if (hbuf->num_nzAH*1LL*Hsize/HBLen > ~(size_t)0) ERR(9999);// 2017.04 uint64_t 1LL -m32
    /* flush buffer to disk-file, 2017-04 add nzAH*1L*... for gcc -m32 */
    ah_nz[blk*(node_ah_blks/pt_n)+hbuf->idxAHblk]= hbuf->num_nzAH;
    if (memHwrite((void *)&b_hbuf[blk],hbuf->num_nzAH*1LL*Hsize/HBLen,blk)==
                                       hbuf->num_nzAH*1LL*Hsize/HBLen
      /* ToDo: wopen on first try to avoid unneccessary file creation */
      || (h_error==0 && opt_maxfile!=0 && my_fwrite((void *)&b_hbuf[blk],
                                  hbuf->num_nzAH*1LL*Hsize/HBLen,1,blk)==1)) {
      b_snz[blk] += b_bnz[blk];
      stored_ah_blks[blk] += 1; /* mem+dsk blockwise 2017-02 replace b_sn1 */
    } // h_error=1;
    b_bn1[blk]=0; /* reset local counter */
    b_bnz[blk]=0; /* reset counter buffered_nonzero-entries */
  } /* if b_hnz < hdsk_nzmax, store to mem+dsk */
  /* hdsk_nzmax is set by hand or by malloc error above */
  if (b_hnz[blk]               <=hdsk_nzmax+hmem_nzmax
   && b_hnz[blk]+hbuf->num_nzAH>=hdsk_nzmax+hmem_nzmax  && hdsk_nzmax>0
//   && mpi_i*pt_n+blk<4   // it may be essential seeing this errors too!
   ) { // assuming that this error will appear on all nodes, ToDo: MPI_send err
     if (mpi_i*pt_n+blk<8) printf(
   "\nWARN: %2d.%2d nz==nzmax, %.1e+%.1e "
   "b_hnz=%4lu*HB+%5lu %6.2f%% %4lu*AH nzAH=%.0f",
         // %% added 2016-01 not sure if node_n1 is correct or b_ofs[blk+1]-b_ofs[blk]-1
         mpi_i, blk, 1.*hmem_nzmax,1.*hdsk_nzmax,
         b_hnz[blk]/HBLen,b_hnz[blk]%HBLen,
         stored_ah_blks[blk]*pt_n*100./node_ah_blks, 
         hbuf->idxAHblk, hbuf->num_nzAH*1.);
      if(mpi_n*pt_n>1 && !mpi_i)
  printf("\nWARNING stored_nz/HBLen>0 and rest_n1/AH>0 may be buggy 2017\n");
    // strange effects seen, of randomly big bfactors, or crashs SH=1..99% 
    // better store AH*NZX-blks
  } // if hdsk_nzmax
  if (b_hnz[blk]+hbuf->num_nzAH<b_hnz[blk]) ERR(610); /* counter overflow, warn */
  b_hnz[blk] += hbuf->num_nzAH; /* count all nonzero matrix elements */
  /* check if all elements of this matrix line have been stored */
  return;
}

#ifndef CKPT_HELP
#define CKPT_HELP 0  /* 2018.1 def/undef to 0/1 better checkpointing */
#endif

#if CKPT_HELP  /* ToDo: better only all 10 or 100 steps or dyn 1/minute */
/* can be used for debugging to, gdp --pid ... ? */
void ckpt_help() /* MPI_Bcast(signal) + MPI_Barrier + sleep30 */
  /* ToDo: PTHREAD? check extra Bcast time needed on big clusters ? */
  /* 2015-02-26: code helping to get clean MPI-state for checkpointing */
{ int signal=sig1;/* spins.h: sighand.h: extern volatile int sig1; */
  MPI_Bcast(&signal,1,MPI_INT,0,MPI_COMM_WORLD); sig1=signal; /* Barrier? */
  if (sig1&15) { /* volatile.int.sig1 set by SIGUSR1 in sighand.c for ckpt */
    MPI_Barrier(MPI_COMM_WORLD); /* wait for all tasks */
    if (!mpi_i) printf("\nHLP: start 30s checkpoint window ...\n");
    /* ToDo: auto trigger ckpt-script by writing to a fifo here? */
    sleep(30); /* safe checkpoint window without MPI-messages */
    if (!mpi_i) printf(  "HLP: ... end of safe checkpoint window\n");
    /* restore hostnames, PIDs, clock tdelay+=32+ckpt etc. here? */
    sleep(2); sig1=0; } }/* wait for all restored tasks */
#else
  void ckpt_help(){} /* dummy */
#endif /* CKPT_HELP */


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * do b2i = scfg2idx with buffered cfgs,
 *   minsymcfg is already called within nhm (ToDo: do it to here)
 * called by hamilton_nhv()            v0 may used to get |v1>+=|H|v0>, <v0|H|v0>
 *       and wop_block()   v2.36 m1=0, v0 used to get <v1|Op|v0>
 * if lenAHblk=1 it is per line, but since v2.30 its for
 *    lenAHblk lines.
 * must be called by each mpi-process and thread to avoid deadlock v2.33
 * careful use MPI within threaded part here!
 * ToDo: call minsymcfg/sq(norm) from here cached/streamed
 *  m1&4 means l1 is not present, dont use cfg2idx for storeh2
 *  diagonal term must be stored first to get sci.norm2
 */

#ifndef DBG_NHM  /* set this in Makefile, so we cant forget and put to tgz */
#define DBG_NHM 0  /* 0=no debug, 1=debug ToDo: DBG_NHM=16 limit output */
#endif
void minsymcfg_blk_lNbrk(thbuf *hbuf); /* ToDo: declare in hilbert.h 2015-04 */
void minsymcfg_blk_bn(thbuf *hbuf); /* ToDo: declare in hilbert.h 2015-04 */
void minsymcfg_blk_bnVS(thbuf *hbuf); /* ToDo: declare in hilbert.h 2015-04 */
void minsymcfg_blk_bnVS_no_ud(thbuf *hbuf); /* ToDo: declare in hilbert.h 2015-04 */
// int blk_scfg2idx(tbase *cfg, ULong *idx, int n); /* defined in hilbert.c */

/* make code better readable by extern debug_function */
/* ToDo: move to spins_dbg.c # -Os */
#if DGB_NHM
void hbuf_debug(char *txt, thbuf *hbuf, int m){
  int i0; char sbuf2[SZccfg];
  if(mpi_i>=8 && (vvv&4)) usleep(10000*16); /* wait for others */
  if(mpi_i< 8 && (vvv&4)) { int i0; usleep(10000*mpi_i); /* sort mpi output */
    printf("\n%s %2d.%d nzAH= %5d maxAH= %d", txt, hbuf->block,
        hbuf->num_nzAH, hbuf->lenAHblk); // >= nahead
    for (i0=0;i0<hbuf->num_nzAH;i0++) {  /* NZXMAX*NUM_AHEAD_LINES */
      if (i0==8 && hbuf->num_nzAH>10) i0= hbuf->num_nzAH - 2;
      if (m==0)
        printf("\n%s %2d.%d i=%5d %s %8.4f %3s", txt, mpi_i, hbuf->block,
           i0, ccfg2(hbuf->acj[i0],sbuf2), D(hbuf->rr[i0]),
           ((hbuf->jj[i0] & XY_FLAG)?"X++":"   "));
      if (m==1)
        printf("\n%s %2d.%d i=%5d %s sj=%+d n2=%3d rr=% 8.4f %3s", txt,
             mpi_i, hbuf->block, i0, /* i0 not shown for diff ahead=1 vs. n? */
             ccfg2(hbuf->acj[i0],sbuf2), /* minsymcfg */
             (int)D(hbuf->asj[i0]),  /* sign */
             (int)hbuf->an2[i0],     /* norm2 + rr */ 
             D(hbuf->rr[i0]), ((hbuf->jj[i0] & XY_FLAG)?"X++":"   "));
      if (m==2)
        printf("\n%s %2d.%d i=%5d %s bj=%d jj= %6d rr=% 8.4f %3s", txt,
             mpi_i, hbuf->block, i0, /* i0 not shown for diff ahead=1 vs. n? */
             ccfg2(hbuf->acj[i0],sbuf2),     /* minsymcfg */
                  hbuf->abj[i0],      /* bj= 0 .. mpi_n */ 
        (int)hbuf->jj[i0] & ~XY_FLAG, /* jj= 0..node_n1-1 */ 
           D(hbuf->rr[i0]), ((hbuf->jj[i0] & XY_FLAG)?"X++":"   "));
    } 
    if (m==2 && hbuf->v0) /* show <i2|H|v0> */
      for (i0=0,ii=hbuf->jj[i0];i0<hbuf->lenAHblk;i0++)
        if (i0<8 || i0>hbuf->lenAHblk-4)
        if (hbuf->nhn[i0] || D(hbuf->nhv[i0]))
        printf("\nDBG nhm_line blk=%d %d ii=%2d nhn,nhv[i]= % 6.4f % 8.6f",
             mpi_i, hbuf->block, ii+i0, hbuf->nhn[i0], D(hbuf->nhv[i0]));
    usleep(10000*(8-mpi_i)); /* sync-back */
  }
}
#else  /* DBG_NHM dummy */
#define hbuf_debug(x,y,m) {}
#endif /* DBG_NHM */

/* ToDo: move to spins_dbg.c # -Os */
#if DBG_NHM /* by SH + i100 */
void dbg_mpi(const char *ss,int *nsend_cnt,int *nsend_ofs,
                            int *nrecv_cnt,int *nrecv_ofs, thbuf *hbuf) {
  if (mpi_n<=8 && (vvv&4) && hbuf->idxAHblk<2) {
    int id;  usleep((mpi_i+1)*10000);
    printf("\n%s: src=%2d. - dest-ofs:", ss, mpi_i);
    for (id=0;id<=mpi_n;id++) printf(" %6d", nsend_ofs[id]);
    printf("\n%s: src=%2d. - dest-cnt:", ss, mpi_i);
    for (id=0;id< mpi_n;id++) printf(" %6d", nsend_cnt[id]);
    printf("\n%s: dst=%2d. - recv-cnt:", ss, mpi_i);
    for (id=0;id< mpi_n;id++) printf(" %6d", nrecv_cnt[id]);
    printf("\n%s: dst=%2d. - recv-ofs:", ss, mpi_i);
    for (id=0;id<=mpi_n;id++) printf(" %6d", nrecv_ofs[id]);
    usleep((mpi_n-mpi_i+1)*10000);
  for (id=1;id<=mpi_n;id++)  // happens 2017-01-31 gcc412 mpi=7
    if (nrecv_ofs[id] != nrecv_ofs[id-1] + nrecv_cnt[id-1]) ERR(9999); } }
#else
#define dbg_mpi(ss,s1,s2,r1,r2,hbuf)  /* dummy */
#endif

static inline int phase_eq(tphase p1,tphase p2){ // copy of hilbert.c
#if VecType & 4 /* remove rounding errors! 1998-170 not exact!!! */
  return( (Norm(p1-p2)<2e-9)? 1: 0);} /* Norm faster as fabs() ????? */
#else
  return( (p1==p2)? 1: 0);}
#endif
         
// ToDo17: split nhm_line to cfg2scfg_nz_blk + MPIscfg2i2vi_nz_blk(incl.remoteMPI)?
//              +nhn[](should be done by op_H_already!)=compact
//              XY_FLAG to ofs[] (zero-nzblks!) + XY_FLAG ggf. in storedH only
//        see asyncMPI, rename to nhm_AHblk?
// static  
/* called by used now in xval.c too v2.36 */
void nhm_line(thbuf *hbuf, int m1) { /* cfg-to-idx via MPI for a H-line */ 
  int line_start, /* hold index to first element of a line 0..hbuf->num_nzAH-1 */
      ii,         /* index from 0 to (hbuf->num_nzAH-1 < NZXMAX*NUM_AHEAD) */ 
      i2=0;       /* counts through 0..NUM_AHEAD-1 (H-line index) */
  cplx sum=0;
  static scplx r0b[B_NUM][NUM_AHEAD_LINES+1]; /* buffer for <nHv>[] */
  scplx  *r0 = r0b[hbuf->block];
  /* 2013-02-07 index 0..ahead-1 */
  hbuf->nhn[i2]=0; /* 2013-01-17 fix diag2 2x2 */
  hbuf_debug("DBG.cfg2i.i",hbuf,0);
  if (HBLrVS < ((hbuf->num_nzAH+VS-1)/VS)*VS) // 2017-02-18
    F3("HBLrVS=%d VS=%d nzAH=%d",HBLrVS,VS,hbuf->num_nzAH);  
  if (hbuf->num_nzAH && !node_n1) { /* zero length rest block, 2016-03 */
    printf("\nERROR.nhm_line: %2d.%d node_n1=%lu, nzAH=%d but 0 expected!\n",
      mpi_i, hbuf->block, node_n1, hbuf->num_nzAH);
    hbuf->num_nzAH=0; } /* not good, but workaround */
  if (hbuf->num_nzAH != hbuf->lenAHblk) { int j; /* ~diag_only 2017-01-28 */
    // in: acj[]=cfg  out: {acj=scfg,an2=norm2,asj=sign}[nzAH] sq40=98%CPU
    if(sym.wud) minsymcfg_blk_dflt(hbuf); /* block operation hbuf->num_nzAH=ca.1400, also slow */
           else minsymcfg_blk_dflt_no_ud(hbuf); /* 15% faster JJ40 */
               /* link to minsymcfg_blk_bnVS_no_ud(hbuf) dflt=bnVS */
    for (j=0,line_start=0,ii=0;ii<hbuf->num_nzAH;ii++) {  /* NZXMAX*NUM_AHEAD_LINES */
      /* phase==0: 32+8 true=0..3% sym_k-depending, important for scfg2i() */
      if (line_start!=ii) { /* first nz element per line is diagonal-element */
        /* avoid cfg-search-errors for sgn0-cfg not in l1, minimize traffic */
        if (phase_eq(hbuf->asj[ii],(tphase)0) /* || hbuf->rr[ii]==0 */) { 
            hbuf->acj[ii]= hbuf->acj[line_start]; /* dummy diag element */
            // ToDo2017.02 can this be TBASE0 ?
            // hbuf->an2[ii]= hbuf->an2[line_start];
            hbuf->rr[ii]=0;
        } else {
          hbuf->rr[ii] *= hbuf->asj[ii]; /* symcfg sign */
          if (hbuf->an2[line_start]!=hbuf->an2[ii]) /* see n2stats, true<1%? */
            hbuf->rr[ii] *= sqrt(hbuf->an2[line_start]/(double)hbuf->an2[ii]);
        }
      }
      if (hbuf->jj[ii] & XY_FLAG) { line_start=ii+1;
        if  (line_start != hbuf->nzofs[j++]) ERR(9998); // 2017-01-30
      } /* ToDo: use ofs_ii[] */
    } /* ii=0..hbuf->num_nzAH */
  } /* if hbuf->num_nzAH==1 else */
  /* compression: add two equal elements gives only small boost (one core) */
  /* sq: 32+8  hnz/n1= 27.1663  -8.22686824 162s 164s         compressed */
  /* sq: 32+8  hnz/n1= 27.2569  -8.22686824 159s 164s +0.3% uncompressed */
  /* sq: 15+15 hnz/n1= 31.582 31.79  (3..51)   compressed 3m20s     */
  /* sq: 15+15 hnz/n1= 31.834 32.04 (15..61) uncompressed 3m15s +1% */
  /* spin1/s1chain.gpl: N=20 20+20 19.30 vs. 19.35 +0.25% */
  /* s1chain+AddSS: N=16 8+8 n1=19 nzx=11.00 vs. 1927.95 Bug! SS*Nw */
  /* sq+AddSS: 4+36 n1=632 nzx=115.54 vs. 159.02 */
  if (opt_pack_nz) { int i1;  // test 2017-01-25, ToDo fill new gaps? nz_ofs[]
    for (line_start=0,ii=0;ii<hbuf->num_nzAH;ii++) { // do it earlier!
      for (i1=line_start;i1<ii;i1++) {
        if (hbuf->acj[i1]==hbuf->acj[ii]) {
            hbuf->rr[i1]+= hbuf->rr[ii]; // ToDo set acj to diag_cfg
            hbuf->rr[ii]=0.0; }
      } if (hbuf->jj[ii] & XY_FLAG) { line_start=ii+1; }
    }
  }
  // now acj[ii] is diag-cfg-dummy and scfg2i is safe, no errors 2017-02-01
    /* ToDo: earliest point to compute mpi_A2a_send_cnt_array */
    /* 3us loop? versus mpi_n*4B IAlltoall? */
  if (hbuf->num_nzAH==hbuf->lenAHblk) { /* diag only 2017-01-28 zizj */
    for(i2=0;i2<hbuf->num_nzAH;i2++) hbuf->nhn[i2]=D(hbuf->rr[i2]); /* ??? */
  } else { /* more than just the diagonal element hbuf->num_nzAH!=1 */
    for (i2=0,line_start=0,ii=0;ii<hbuf->num_nzAH;ii++) {  /* NZXMAX*NUM_AHEAD_LINES */
      if (hbuf->acj[line_start]==hbuf->acj[ii]) // scfg(Op(cfg))==cfg
          hbuf->nhn[i2]+=D(hbuf->rr[ii]); /* Hii ~cplx */
      if (hbuf->jj[ii] & XY_FLAG) { line_start=ii+1; i2++; hbuf->nhn[i2]=0; }
    } /* ii=0..hbuf->num_nzAH */
  } /* if hbuf->num_nzAH==1 else */
  hbuf_debug("DBG.cfg2i.o",hbuf,1);
  if (m1&4) return; /* for storeh2 only, no l1, don't do scfg2idx  */

  /* ToDo: like geth if (mpi_n==1) ..., better give rnd_xorshift results on test? */
  if (!l1) for (ii=0;ii<hbuf->num_nzAH;ii++) hbuf->abj[ii]=mpi_i; /* speed_test 2016-04-07 */
  if (l1) /* 2015-10-08 benchmark, l1 uninitialized */
  for (ii=0;ii<hbuf->num_nzAH;ii++) {  /* hbuf->num_nzAH <= NZXMAX*NUM_AHEAD_LINES */
    /* ToDo: bad for performance, but needed at the moment for b2i() ??? */
    if (hbuf->rr[ii]==0) { /* replace by dummy data, 0...0.03% hit? */
        hbuf->abj[ii] = mpi_i;    /* this node, avoid communication */
    } else {
        /* scfg2node, needs mpi_l1[mpi_n] ToDo: use blk_scfg2node() */
        hbuf->abj[ii] = scfg2node(hbuf->acj[ii]); /* O(log2(mpi_n)) */
    }
  } /* for ii=0..n */
  if (mpi_n==1 && l1/* 2015-10-08 */) { /* only local data */
    // int blk_scfg2idx(tbase *cfg, XY_TYPE *idx, int n, thbuf *hbuf)
#if 1 /* 2017-04 tested ok for i64 */
    int ix=0; hbuf->nzofs[ix] = 0;
    for (ii=0;ii<hbuf->num_nzAH;ii++)
      if (hbuf->jj[ii] &   XY_FLAG) {
          hbuf->jj[ii] &= ~XY_FLAG; ix++; hbuf->nzofs[ix]= ii+1; }
    if (blk_scfg2idx(hbuf->acj, hbuf->jj, hbuf->num_nzAH, hbuf) ) {
      if (blk_scfg2idx(hbuf->acj, hbuf->jj, hbuf->num_nzAH, hbuf) ) {
        printf("\nERR.SH.scfg2i %2d.%d", mpi_i,hbuf->block); ERR(9999); } }
    for (ii=1;ii<=hbuf->lenAHblk;ii++)
       hbuf->jj[ hbuf->nzofs[ii]-1 ] |= XY_FLAG; /* 2017-04 needed? */
#else
    for (ii=0;ii<hbuf->num_nzAH;ii++) {  /* hbuf->num_nzAH <= NZXMAX*NUM_AHEAD_LINES */
      /* scfg2idx() fails, if b.sign==0 for k=pi symmetries */
      /* ToDo: replace XY_FLAG by nzx[]/ofs[] for speedup (output size needed? n1*(nzx*8B+v0+v1) */
      /* ToDo: replace by hbuf->row_ofs[AH+1]
       *   sparse matrix SpMV + compressed sparse row */
      hbuf->jj[ii] = (hbuf->jj[ii] & XY_FLAG)  /* save XY_FLAG (ToDo remove) */
          |  scfg2idx(hbuf->acj[ii]);  /* 0<=jj<node_n1 */
      /* ToDo: remove XY_FLAG 1bit/nz add nzx[n1_ahead]=1B/n1 + add limit-checks
       * ToDo2015: use blk_scfg2idx */
    }
#endif
  }
  /* ToDo: better do it after _nhv call when needed??? must be done by MPI! */
  if(mpi_n==1 && hbuf->v0){ /* only needed for <nHv> */
    i2=0; r0[0]=sum=0; /* scplx = cplx = 0 ToDo */
    for (ii=0;ii<hbuf->num_nzAH;ii++) {
      sum += hbuf->rr[ii] * M_f2d(hbuf->v0[ hbuf->jj[ii] & ~XY_FLAG]);
      /* store diag-block <i2|H|v0> to r0[], 2013-02-07 set r0[next]=0 */
      if (hbuf->jj[ii] & XY_FLAG) { r0[i2]=sum; i2++; r0[i2]=sum=0; }
    }
  }
#ifdef CONFIG_MPI  /* if mpirun fails, try lamclean+ipcs+ipcrm */
  if (mpi_n>1) { /* nhm_line.mpi should also work for mpi_n=1 */
    int i1, id, block=hbuf->block; /* ToDo: rename block to pt_i */
    void *psrc, *pdst; /* src,dst for better looking code v2.36 */
    int   len1,  len2, rc=0;  MPI_Datatype mpi_dtype;
    /* ToDo: bsend_cfg[B_NUM*NZXMAX*AHEAD] + startidx[block]=MAX/pt,len[]  */
    /* 1. find destination node for each cfg-data (from hbuf->abj[])
          done in each thread + every thread count per dest
     * 2. write sort-back-index to bsort_bck + sort data (new: for all dest)
     * 3. alltoall data transfer (was bad scaling until 2017.1)
          faster if done by thread 0 only (all threads collected)
     * 4. compute data (search cfg-idx + get r*v[idx])
     * 5. alltoall back transfer
     * 6. sum to v[i] */
    /* phase 2: count data per destination node, reorder later */
    for (id=0;id<=mpi_n;id++) bsend_cnt[block][id]=0;
    for (ii=0;ii<hbuf->num_nzAH;ii++) { /* 2017-02 incl. nzr==0 */
      id=    hbuf->abj[ii]; /* destination node */
      bsend_cnt[block][id]++; } /* count for each dest. node */
    bsend_of2[block][0]= bsend_ofs[block][0]= 0; /* ini private ofs */
pthread_barrier_wait(&pt_barrier);  /* 2017-01-18 wait all threads */
    if (block==0) { int it,ofs; /* master thread only collects all ofs +cnt */
      nsend_ofs[0]= 0; /* ini private ofs */
      for (id=0;id<=mpi_n;id++) { /* runs one more, to store overall size! */
        ofs=bsend_ofs[0][id]=nsend_ofs[id];
        for (it=1;it<pt_n;it++) {
          ofs += bsend_cnt[it-1][id]; bsend_ofs[it][id]= ofs; }
        nsend_ofs[id+1] = bsend_ofs[pt_n-1][id] + bsend_cnt[pt_n-1][id];
        nsend_cnt[id]=    nsend_ofs[id+1] - nsend_ofs[id]; }
      nsend_cnt[mpi_i]=0; /* do not send local data, self_len 2017-02-24 */ 
      if (nsend_ofs[mpi_n] > B_NUM*MPI_NZ_BUFSZ1)
        F1("enlarge B_NUM*NZ_BUFSZ1/AH=NZX to %d++",
           nsend_ofs[mpi_n]/(B_NUM*MPI_NZ_BUFSZ1)); 
      ckpt_help(); /* MPI_Bcast(signal) + MPI_Barrier + sleep30 */
      rc=My_MPI_Ialltoall(nsend_cnt,1,MPI_INT,                 /* nhm_line */
                          nrecv_cnt,1,MPI_INT,MPI_COMM_WORLD,&a2a_request);
      if (rc) perror("SH.A2a");
    } /* block==0 only for MPI stuff */
    // PT_Barrier + every thread copies its own data to the global array
    pthread_barrier_wait(&pt_barrier);  /* 2017-01-18 wait all threads */
    /* ... now we sort in groups for each destination */
    for (id=0;id<mpi_n;id++) bsend_of2[block][id]= bsend_ofs[block][id];
    for (ii=0;ii<hbuf->num_nzAH;ii++) { // include rr==0!
      id= hbuf->abj[ii]; /* destination node (array block index) = scfg2node */
      len1 = bsend_of2[block][id]; /* next free entry for destination id */
      bsort_bck[block][ii]     = len1; /* sort-back idx */
      nsend_cfg[len1]   = hbuf->acj[ii];  /* copy sym. spin cfg */
      bsend_of2[block][id]++; /* next free place in list for dest id */
    } /* thread ready for send */
#if DBG_NHM  /* nhm_line ToDo17 replace by dbg_subfunction dbg_hbuf !!!!! */
    if (mpi_n<4 && (vvv&4) && mpi_syncs<8){ usleep((mpi_i+1)*10000); int ix=0;
      char sbuf2[SZccfg];
      printf("\nDBG.nhm_bl %2d.%2d iAH,nAH,nnzAH= %5lu %5d %5d",
           mpi_i, block, hbuf->idxAHblk, hbuf->lenAHblk, hbuf->num_nzAH);
      printf("\nDBG.nhm_bl %2d.%2d ", mpi_i, block);
printf(" cnt=");for(id=0;id< mpi_n;id++)printf(" %5d", bsend_cnt[block][id]);
printf(" ofs=");for(id=0;id<=mpi_n;id++)printf(" %5d", bsend_ofs[block][id]);
      for (ii=0;ii<hbuf->num_nzAH;ii++) {
        if (ii==(int)x_max && hbuf->num_nzAH>7) ii=hbuf->num_nzAH-2;
        printf("\nDBG.nhm_bl %2d.%2d nz_i,ix=%5d %6d bsort=%6d cfg= %s by=%2d",
            mpi_i,block,ii,ix,bsort_bck[block][ii],
            ccfg2(hbuf->acj[ii],sbuf2), hbuf->abj[ii]);
            if (hbuf->jj[ii]&XY_FLAG) ix++;
    } }
#endif
    /* ensure balanced number of MPI calls outside, using hbuf->num_nzAH=0 */
// PT_Barrier + every thread copies its own data to the summ area
    pthread_barrier_wait(&pt_barrier);  /* 2017-01-18 wait all threads */
{   /* MPI communication part for SH starts here, MPI_Alltoallv */
if (block==0) { int ofs; // MPI by thread 0 only (bigger data blocks = speed)
/* 2017.1 not symmetric! splitted elements! disadvantage of ~compression */
      /* compact statistics about off-diagonal-blocks = MPI traffic */
      if (mpi_n<=8 || (vvv&4)) // for big num tasks it eats to much CPU
      {int i,n=(mpi_n>16)?16:mpi_n;for(i=0;i<mpi_n;i++) /* slow? vvv&4 ? */
        mpi_xy_traffic[mpi_i*n/mpi_n*n + i*n/mpi_n] += nsend_cnt[i];}
      /* ToDo: alternatively try to store all nrecv_cnt[n1/AH][mpi_n] */
      My_MPI_Wait_Ia2a(&a2a_request,&mpi_status);
      /* MPI_Alltoall must be finished here, we need nrecv_cnt */
      for (ofs=nrecv_ofs[0]=0,id=1;id<=mpi_n;id++) {
        ofs += nrecv_cnt[id-1]; nrecv_ofs[id] = ofs; }
      len1= nsend_ofs[mpi_n];
      len2= nrecv_ofs[mpi_n];
      if (hbuf->max_nzAH_recv<len2) { /* 2017-02-17 */
          hbuf->max_nzAH_recv=len2; }
      if (B_NUM*MPI_NZ_BUFSZ2<len2)
          F1("enlarge NZ_BUFSZ2 to %d++", len2/B_NUM);
      dbg_mpi("MPI.SH1",nsend_cnt,nsend_ofs,nrecv_cnt,nrecv_ofs,hbuf);
      mpi_syncs++; mpi_traffic += nsend_ofs[mpi_n]-nsend_cnt[mpi_i];
      { int i3,i4; for(i4=0;i4<mpi_n;i4++) if(i4!=mpi_i) { // 2017-01-22
        if (mpi_maxlen<nsend_cnt[i4]) mpi_maxlen=nsend_cnt[i4]; /* max len */
        for(i3=0;i3<6 &&  nsend_cnt[i4]>>10 > 1<<(2*i3);i3++);
        mpi_stats_npkts[i3]++; } } /* 2015-01 mpi statistics */
      len1=sizeof(tbase);
   // mpi_tset(len1,&dtype,nsend_cnt,nsend_ofs,nrecv_cnt,nrecv_ofs);
      if (len1== 4) mpi_dtype=MPI_UNSIGNED; else /* MPI_Datatype */
      if (len1== 8) mpi_dtype=MPI_LONG_LONG; else
      if (len1==16) mpi_dtype=MPI_LONG_DOUBLE; else { mpi_dtype=MPI_BYTE;
        for (ii=0;ii<mpi_n;ii++) nsend_cnt[ii] *= len1;
        for (ii=0;ii<mpi_n;ii++) nsend_ofs[ii] *= len1;
        for (ii=0;ii<mpi_n;ii++) nrecv_cnt[ii] *= len1;
        for (ii=0;ii<mpi_n;ii++) nrecv_ofs[ii] *= len1; }
      rc=My_MPI_Ialltoallv(nsend_cfg, nsend_cnt, nsend_ofs, mpi_dtype,\
                           nrecv_cfg, nrecv_cnt, nrecv_ofs, mpi_dtype,\
         MPI_COMM_WORLD, &a2a_request); if (rc) perror("A2av.SH.cfg");
      /* may hang if memory is to low! ompi-1.4 2M-- free 2015-04 */
      if (len1!=4 && len1!=8 && len1!=16) {
        for (ii=0;ii<mpi_n;ii++) nsend_cnt[ii] /= len1;
        for (ii=0;ii<mpi_n;ii++) nsend_ofs[ii] /= len1;
        for (ii=0;ii<mpi_n;ii++) nrecv_cnt[ii] /= len1;
        for (ii=0;ii<mpi_n;ii++) nrecv_ofs[ii] /= len1; }
      /* local part extracted, to be done parallel to remote data transfer */
} /* block==0 */
      if (l1 && blk_scfg2idx(nsend_cfg+bsend_ofs[block][mpi_i],
                             nrecv_iy +bsend_ofs[block][mpi_i],
                             bsend_cnt[block][mpi_i], hbuf) ) {
        printf("\nERR.SH.scfg2i %2d.%d", mpi_i,block); ERR(9999); }
      if (l1 && hbuf->v0) for (ii=0;ii<bsend_cnt[block][mpi_i];ii++)
        nrecv_vy[bsend_ofs[block][mpi_i]+ii]=
           hbuf->v0[ nrecv_iy[bsend_ofs[block][mpi_i]+ii] ]; /* rnd-access */
if (block==0) My_MPI_Wait_Ia2a(&a2a_request,&mpi_status);
// PT_Barrier + every thread copies its own data to the summ area
      pthread_barrier_wait(&pt_barrier);  /* 2017-01-18 wait all threads */
      len2= (nrecv_ofs[mpi_n]+pt_n-1) / pt_n;
      len1= block*len2; // ofs ... check OVL more threads than matrixsize
      if (len1     >nrecv_ofs[mpi_n]) len1=nrecv_ofs[mpi_n]; 
      if (len1+len2>nrecv_ofs[mpi_n]) len2=nrecv_ofs[mpi_n]-len1;
      /* blk_scfg2idx(tbase *cfg,int *idx,len2); 2017.1 random search SLOW */
      if (l1 && blk_scfg2idx(nrecv_cfg+len1,nsend_iy+len1,len2,hbuf) ) {
        printf("\nERR.SH.scfg2i %2d.%d l1,l2= %d %d",mpi_i,block,len1,len2);
        MPI_Abort(MPI_COMM_WORLD,1); } // 2017-02-01 kago42u15_224x16_@223.15
      /* ToDo: or recv scfg+rr and send back summed rr[i]*v[jj[i]] ? rr/NZX */
      if (l1 && hbuf->v0) for (ii=0;ii<len2;ii++) /* len2 <= MPI_NZ_BUFSZ1 */
        nsend_vy[len1+ii]  = hbuf->v0[ nsend_iy[len1+ii] ]; /* random access */
#if DBG_NHM /* ToDo17 replace by one line sbg_subfunction !!!!!! */
      if((vvv&4) && mpi_n<8 && mpi_syncs<16) { char sbuf2[SZccfg];
        id=0; usleep(1000*mpi_i); /* order */
        printf("\nDBG.scfg2i %2d.%2d ofs,len= %5d %5d", mpi_i,block,len1,len2);
        for (ii=0;ii<len2;ii++) {  /* count blk_hnz (NZX*AHEAD) */
          for (;nrecv_ofs[id+1]<=ii && id<mpi_n;id++);
          if (ii==(int)x_max && len2>(int)x_max+3) ii=len2-2;
          printf("\nDBG.scfg2i %2d.%2d nz_i=%5d iy=%6lu from cfg= %s by=%2d",
            mpi_i, block, ii, (ULong)nsend_iy[len1+ii],
            ccfg2(nrecv_cfg[len1+ii],sbuf2), id);
         // ToDo add hbuf->ix[AH] for easier DBG+handling?
      } }
#endif
// PT_Barrier + every thread copies its own data to the summ area
    rc= pthread_barrier_wait(&pt_barrier) + 0;  /* 2017-01-18 wait all threads */
if (block==0) {
#if DBG_NHM
      if(vvv && mpi_n<8 && mpi_syncs<8)
        printf("\nDBG.MPI_bs %2d. - biAH=%5lu nz= %5d",
           mpi_i, hbuf->idxAHblk, nrecv_ofs[mpi_n]);
#endif
      /* ToDo: let bsend_iy[block] on remote node until geth */
      /* inter-node communication is often the bottleneck */
      /* on Gbit int4/float4 will double the speed vs. long8/float8 */
      /*  float2 during first iterations could be speed up further  */
      /* ToDo17 change nrecv_vy <-> nsend_vy etc. like geth() */
      if (!hbuf->v0)  /* send back ordered data to remote node */
           { len1=sizeof(XY_TYPE); pdst=nrecv_iy; psrc=nsend_iy; }
      else { len1=sizeof(scplx);   pdst=nrecv_vy; psrc=nsend_vy; }
   // mpi_tmod(sizeof(tbase),&dtype,nsend_cnt,nsend_ofs,nrecv_cnt,nrecv_ofs);
      if (len1== 2) mpi_dtype=MPI_SHORT; else /* MPI_Datatype */
      if (len1== 4) mpi_dtype=MPI_UNSIGNED; else /* MPI_Datatype */
      if (len1== 8) mpi_dtype=MPI_LONG_LONG; else
      if (len1==16) mpi_dtype=MPI_LONG_DOUBLE; else {
        for (ii=0;ii<mpi_n;ii++) nsend_cnt[ii] *= len1;
        for (ii=0;ii<mpi_n;ii++) nsend_ofs[ii] *= len1;
        for (ii=0;ii<mpi_n;ii++) nrecv_cnt[ii] *= len1;
        for (ii=0;ii<mpi_n;ii++) nrecv_ofs[ii] *= len1; 
        mpi_dtype=MPI_BYTE;}
      /* send results back to asking threads (block=thread_idx) */
      rc=My_MPI_Ialltoallv(psrc,nrecv_cnt,nrecv_ofs,mpi_dtype,\
                           pdst,nsend_cnt,nsend_ofs,mpi_dtype,\
         MPI_COMM_WORLD, &a2a_request); if(rc)perror("SH.A2av_back");
      My_MPI_Wait_Ia2a(&a2a_request,&mpi_status);
      /* may hang if memory is to low! ompi-1.4 2M-- free 2015-04 */
} /* block==0 */
// PT_Barrier + every thread copies its own data to the summ area
      rc= pthread_barrier_wait(&pt_barrier) + 0;  /* 2017-01-18 wait all threads */
    /* sorting back results from remote nodes */
} /* old MPI_Sendrecv-loop id=0..mpi_n-1 */
//  --- begin --- was out mpi loop until 2015-01-31 
      if (!hbuf->v0)  /* outside the mpi-loop? */
        for (ii=0;ii<hbuf->num_nzAH;ii++) { /* idx_cfg | XY_FLAG */
          hbuf->jj[ii] = (hbuf->jj[ii] & XY_FLAG)
                     | nrecv_iy[ bsort_bck[block][ii] ];
#if DBG_NHM
      if ((!hbuf->v0) && 
       (ii<(int)x_max || ii>hbuf->num_nzAH-4) && mpi_i*pt_n+block<4 && (vvv&4)
    && (hbuf->idxAHblk<2 ||  /* first 2 + last 2 */
        hbuf->idxAHblk>=(b_ofs[hbuf->block+1]
                        -b_ofs[hbuf->block  ])/NUM_AHEAD_LINES-1))
     printf("\nDBG.backst %2d.%2d nz_i=%6d bsort=%6d iy=%6d by=%2d %+4.2f%s",
               mpi_i, block, ii, bsort_bck[block][ii],
                  (int)nrecv_iy[ bsort_bck[block][ii] ], hbuf->abj[ii],
                  D(hbuf->rr[ii]), ((hbuf->jj[ii] & XY_FLAG)?" X++":""));
        // output by 
#endif
      } /* for ii if !hbuf->v0 */
    if (hbuf->v0) /* decompress r2 and add to r0[]=<i2|H|v0> */
    for (r0[0]=0,i1=ii=0;ii<hbuf->num_nzAH;ii++) { /* <i2|H|v0> ??? */
      r0[ i1 ] += hbuf->rr[ii] * nrecv_vy[ bsort_bck[block][ii] ];
      if (hbuf->jj[ii] & XY_FLAG) { i1++; r0[i1]=0; } /* next HLine <= AHEAD_LINES */
    }
//  --- end --- was out mpi loop until 2015-01-31 
  } /* mpi_n>1 */
#endif  /* CONFIG_MPI */
  /* ToDo: better write <nHi> <nHv> separately?
   *    using .rr for two different things confuse the reader
   *  2013-02-07 use nhv[ii] instead of hbuf->rr[ii]
   */
  if (hbuf->v0) /* store <i2|H|v0>, nhv:cplx *r0:scplx */
    for (ii=0;ii<hbuf->lenAHblk;ii++) hbuf->nhv[ii]= r0[ii]; /* a2 */
  hbuf_debug("DBG.cfg2i.e",hbuf,2);
} /* nhm_line() */

/* ------------------------------------------------
 * nhm() put (not symmetrized) off-diag-matrix-element to hbuffer
 * r*(j|j><i|i), Hij=r*|j><i|, i=n, j=m, hbuf->sci,scj ??? 
 *   v2.28: do cfg-to-idx later in nhm_line
 *   v2.35: do minsymcfg later in nhm_line (better caching, +FPGA)
 *  hbuf->ofsAHblk is the cfg index on the node (0..b_len-1) # less bits
 *  cfgj is the cfg modified by operator H, r*cfgj=H(i,j).sci
 * ToDo: store para_index instead of r
 * ------------------------------------------------
 */
static inline void nhm(tbase cfgj, cplx r, thbuf *hbuf) {
  if (r==0) return;    /* 32+8: true=0 (only if parameter==0), speedup? */
#ifdef CONFIG_ANSATZ
  if (!ansatz_test(cfgj)) return;  /* ToDo: buffered call? */
#endif
  if (hbuf->num_nzAH >= NZXMAX*NUM_AHEAD_LINES) hbuf->nzAH_overflow++;
    // ERR(9200); /* 2017-01 moved here  */
  else {
    hbuf->rr[hbuf->num_nzAH]=r;    /* fill H-buffer of this thread */
    hbuf->acj[hbuf->num_nzAH]=cfgj; /* by operator H modified config, H(cfg_i) */
    hbuf->jj[hbuf->num_nzAH]=0; /* v2.30 reset to carry XY_FLAG (set in _nhv) */
    hbuf->num_nzAH++; /* number nz-elements for AH matrixlines */
  }
  // if (hbuf->num_nzAH > NZXMAX*NUM_AHEAD_LINES) ERR(9200);
  return;
} 

/*  called by hamilton_nhv only (2017-01),
 *     thread computes AH-block to {cfg,factor,phase}[AH*nzx]
 *  compute_nzH_blk Operator_H working on hbuf->cfg[nahead],
 *  producing nonzero matrix elements for a piece of rows hbuf->scfg,rr[0..nz]
 *   this can be called to compute i+nahead block, during MPI_IAlltoallv(blk i+0)
 * fill nzH-block by H*[i...i+nahead-1] where i starts with node_i/ptn*pt_i
 *    and loops by i += nahead (ToDo better use AH_block index iAH?)
 * ----------------------------------------------------------------------------- */
int compute_nzH_blk(
    ULong i, ULong nahead, /* start_i=0..node_n1-1, loop i...i+nahead-1 */
       /* i starts with 0 (zero) for all threads !! */
    thbuf *hbuf) { /* input/output+buffer */
  tbase ib1;     /* target and workconfig */
  ULong iahead;
  int ip1,ip2,ii,i1,ia,ib; /* worksites, workindizes */
  int line_start=0; /* index of hbuf.nz[nz_idx] start of new x-row */
  cplx r1;   /* ToDo: old: sum = nhv = <i|H|v0> moved to [2*i+1].rr ??? */
  double rp,  /* temporary exchange parameter for S1SYM || CM(1,1,0) ToDo2015: tU? */
         sun; /* sun=diagonal sum, onlyZ? */
#if CM(0,1,1)
  int j, is;
#endif
  if ((3==3) != 1 || (2==3) != 0) ERR(9999); /* true=1 false=0 needed for ii==3 */
  if (hbuf->num_nzAH>0) ERR(9999); // set outside to zero?
  if (nahead>b_ofs[hbuf->block+1]-b_ofs[hbuf->block]) ERR(9999); /* last blk set outside */
 // hbuf->num_nzAH=0; /* number of stored nz-elements in ahead_nz_buffer outside/inside op? */
  hbuf->nzAH_overflow=0; /* 2017-02-18 */
  // i starts with zero on all threads, nah is AH except for the last block
  // printf ("\nDBG.nzH %2d,%2d i=%lu nah=%lu",mpi_i,hbuf->block,i,nahead);
  /* get H lines ahead to group nonlocal memory access + do mpi blocks */
  /* 0 <= hbuf->block < pt_n,   0 < i <= b_ofs[blk+1]-b_ofs[blk] */
  for (iahead=0;iahead<nahead;iahead++) {
    // if (l1 /* 2015-10-08*/ && ... ) /* 2016-03 bad idea */
    hbuf->sci.scfg=ib1=hbuf->cfg[iahead]; /* sci.scfg ?? */
    hbuf->ofsAHblk=i+iahead; /* ToDo: remove hbuf->ofsAHblk!? but can be used for progress */
    hbuf->idxAHblk=i/NUM_AHEAD_LINES; // ToDo17 better outside this func.
    sun=0;
    hbuf->acj[line_start]=ib1; /* always start with cfg_i (better code) */
    hbuf->jj[line_start]=i+iahead; /* ToDo: replace by num_of_line_elements */
    hbuf->abj[line_start]=mpi_i /* *pt_n+hbuf->block */; /* (abj,jj) == idx(cj) */
    hbuf->rr[line_start]=0; /* set at the end of this loop */
    /* minimum of one entry per line == diag element */
    if (hbuf->num_nzAH<HBLrVS) hbuf->num_nzAH++;
    else                       hbuf->nzAH_overflow++;
#ifdef AddSS	/* only ok for JJ-model !???, AddSS must be a float number */	
    /* add AddSS(SS-ZZ)=sum_(i,j).5(SipSjm+SimSjp) to Ham, max. nu*nd terms? */
    for(ia=0;ia<nn;ia++){    ip1=geti(ib1,ia);
      if(ip1==1 || ip1==2) sun+=.5*AddSS;  /* diagonal part Z^2 */
      for(ib=ia+1;ib<nn;ib++){ ip2=geti(ib1,ib);
        if((ip1^ip2)==3) nhm(Flip2(ib1,ia,ib),AddSS,hbuf);
    } }
#endif
/* ??? should we use symmetry at this point for speed up ??? i-j => pi-pj 
 * with smallest fitting  resulting configuration? */
    for(i1=0;i1<nw;i1++){
      ia=iw[i1][0];ip1=geti(ib1,ia); /* binary_01=up 10=down 00=empty 11=ud */
      ib=iw[i1][1];ip2=geti(ib1,ib); /* binary_01=up 10=down */
      if (ia==ib) continue; /* 2016-01-21 silently ignore bad self-bonds */
      rp=para[ww.wj[i1]];  /* exchange parameter J */
#if defined(CONFIG_S1SYM) && CM(1,0,0)  /* N=8 s=5/2  7*faster */
    /* 2016-01 recheck speedup for cubocta.def s=7/2  9*faster nzx? */
    /*  reducing nzx (nonzero matrix terms by compacting LM-equals)! */
    /* for composed spins take only lowest up(down) (ex: s=5,m=1: uuuudd)
     * and correct by a factor num_up(down) to speed up of (2s)^2/4
     *  ToDo: emulate by S(+/-,idx_composed)
     *  S^+|j,m>=((j-m)*(j+m+1))^(1/2)|j,m+1>
     *  S^-|j,m>=((j+m)*(j-m+1))^(1/2)|j,m-1>
     * Why C does not support break(2) to leave 2 loops?
     * subspins of composed spins must be ordered: uuu...ddd
     * ToDo: test mixed spin-lengths
     *  bonds must be sorted to let this work ??? lowest i-j first
     */
// ToDo: if pre[ia] same LM and same bit, skip (rp=0)
// ToDo:  count rp factor for every next[ia] has same bit[s]
//   if done for tU do if (!rp) continue; = faster skipping
// ToDo: check nzx is minimum for best code, add tU tJ 
// ToDo: simplify,  code is looking unnecessary complex
#if  CM(0,1,1)    /* bugfix 2016-01-04 not final, partly fixed? */
      if ((ip1==1 && ip2==2) || (ip1==2 && ip2==1)) /* tU-JxyJxy */
      /* 2018-01 this looks buggy positioned, ToDo: test tUJ_S1SYM */
#endif 
      if (sym.num_lm!=nn && rp && ip1!=ip2) { /* ud - du flip? ToDo: uu also! */
       int nm[4],i,node;
       if (sym.ls[ia]>1 && sym.ls[ib]>1 && sym.li[ia]==sym.li[ib]) rp=0;
       for (node=0;node<2 && rp;node++){  ii=iw[i1][node]; /* ii={ia,ib} */
         if (sym.ls[ii]>1) {  /* composed spin detected */
           if (sym.li[ii]!=0 && sym.li[ii]!=sym.ls[ii]-1) rp=0; /* JJ only */
           else { /* member 0 or s-1 */
             nm[0]=nm[1]=nm[2]=nm[3]=0; /* later add 3. - .3 flip */
             for (i=0;i<nn;i++) if (sym.lm[i]==sym.lm[ii]) /* ToDo: next! */
               for (;i<nn;i=sym.ln[i]) nm[geti(ib1,i)]++;
             if (sym.li[ii]==0)            rp*=nm[2];  /* S^+|....d> */
             if (sym.li[ii]==sym.ls[ii]-1) rp*=nm[1];  /* S^-|u....> */
           }
         }
       }
     }
#endif /* defined(CONFIG_S1SYM) */
     ii=ip1^ip2;
#if   CM(1,1,0)  /* JJ,tJ (ToDo testcase tUJ? tU is below incl. JH) SiSj-terms */
     /* SiSj - ninj/4 + ham1*ninj/4 */
     /* Ham1,Jadd=-0.25: energy shift -3/4 and 1/4 to -1 and 0 (Ham1=0) */ 
     /* [H += ninj/4] (compensation of -ninj/4), 2spinEV: -3/4, 1/4 */
     /* split .wj to Jxy=.wj and Jz=.wj+.wjz! rp=J*S1Sym */
     /*   [.. .o .x .3 o. oo ox o3 x. xo xx x3 3. 3o 3x 33] 16*2site */
  /* Jz{ 0  0  0  0  0 +1 -1  0  0 -1 +1  0  0  0  0  0} *0.25  ToDo */
  /* reindex to 0..3,4 = [oo,ox,xo,xx,rest] = 3+1-vector (+1 for remove if) */
     /* Jxy= (4+1)*(4+1)-matrix? +1 rest column is for zero case tJ,tU ??? */
     /* ~XY-model (H_JZ=0 faster than h_zpara=0, optimazed away) */
 /* h_zpara * Siz*Sjz + (Ham1-1)*ninj/4  Ham1=E+1/4-shift */
     /* add 1/4 here for all 4 states, add -1/2 for ox,xo only */
     /* (ii==3) is 1 if true or 0 if false */
     CM({},if(ip2 && ip1),{})  /* ni*nj, wjz=-wj vor XY-model */
       sun += H_JZ*h_zpara*(.25+H_Jadd-0.5*(ii==3))*(rp+para[ww.wjz[i1]]);
     if (ii==3 && rp!=0){  /* term for different spin directions: xo, ox */
       /* 2spinEV: -1, 0 = + 0.5 (SipSjm+SimSjp) */
       /* nondiagonal terms (sym: partly diagonal) */ 
       r1=h_xypara*.5*rp;  /* rp = parameter J (extra_wJz=Jz-J) */
       /* signum of imag depends on whether ud or du */
       IFww( r1 *= C(ww.co[i1],ww.si[i1]*((ip2==1)?-1:1)); )
       nhm(Flip2(ib1,ia,ib),r1,hbuf);
     }
#if NOSZSYM /* 2017-12 NoSzSym, see below, terms from user wishlist */
 /* NoSzSym results from  Kxx!=Kyy Bx,By!=0 Dxyz!=0, complex: By,Dx,Dz */
 /* except Kzz,Bz all of-diagonal-terms */
    if ((ip1==1 || ip1==2)
     && (ip2==1 || ip2==2)) { int sgn1=((ip1==ip2)*2-1);
      nhm(Flip2(ib1,ia,ib),      h_xypara*.25*para[ww.wjx[i1]],hbuf);
      nhm(Flip2(ib1,ia,ib),-sgn1*h_xypara*.25*para[ww.wjy[i1]],hbuf);
    }
     /* S^+=Sp=Sx+iSy S^-=Sm=Sx-iSy 
      *  Sx=(Sp+Sm)/2  Sy=-i(Sp-Sm)/2
      *  pauli: sx=[0,1][1,0] sy=[0,-i][i,0] sz=[1,0][0,-1]  spin=[n_up,n_dn]
     * Dzyaloshinskii-Moriya  Dij . (Si x Sj)   = user-wishlist
     *                  = Dx*(SiySjz-SizSjy)
     *                   +Dy*(SizSjx-SixSjz)
     *                   +Dz*(SixSjy-SiySjx)
     *           =Dx(-i(Sip-Sim)Sjz         +iSiz(Sjp-Sjm))/2
     *           +Dy(  Siz(Sjp+Sjm)         - (Sip+Sim)Sjz)/2
     *           +Dz(-i(Sip+Sim)(Sjp-Sjm)/4+i(Sip-Sim)(Sjp+Sjm))/4
     *  = Vector senkrecht zu Si,Sj-Ebene = ij-Richtungsabhaengig!
     *  Kitaev       Kxx*SixSjx + Kyy*SiySjy + Kzz*SizSjz  = user-wishlist
     *    = Kxx/4(Sip+Sim)(Sjp+Sjm)       2 flips but sign depends
     *     -Kyy/4(Sip-Sim)(Sjp-Sjm)
     *     +Kzz   SizSjz
     *  allg?     Sum Si . Gamma_ij . Sj
     *    G = [[Kxx,Dz,-Dy],
     *         [-Dz,Kyy,Dx],
     *         [-Dy,-Dx,Kzz]]
      * SixSjx= 0.25*(Sip+Sim)(Sjp+Sjm)= 0.25(SipSjp+SimSjm+SipSjm+SimSjp)
      * SiySjy=-0.25*(Sip-Sim)(Sjp-Sjm)=-0.25(SipSjp+SimSjm-SipSjm-SimSjp)
      * SixSjy=
      * SixSjx+SiySjy = 0.5 (SipSjm+SimSjp)  # == xy-Heisenberg +-,-+
      * Si x Sj Dzyaloshinskii-Moriya 
      *   h_xypara=0 should be 0 here, use h_xpara, h_ypara
      * Feld h_x, h_y, Jx, Jy
      *  siteij_table[++,+-,-+,--] ip1*2+ip2
      */
#endif
#endif  /* CM(1,1,0) */

#if   CM(0,1,1)		/* t_{ij} cp_i c_j     t=t_{ij} t^*=t_{ji} */
     if( CM( ii , ii>0 && ii<3 , ii ) ) {
       /* valid: .u  ,only tU: ud u3 ToDo: faster with tU only ii>0 ??? */
       is=parity(ib1 & p_mask[ia][ib]); /* fermionic signum without sites */
       for(j=0;j<2;j++)if(ii & (j+1)){ /* up or down */
         r1=is*para[ww.wt[i1]];        /* signum*t_{ij} */
         if( (j==0) ? ip1&2 : ip2&1 ) r1=-r1; /* fermionic sign correction */
         IFww( r1*=C(ww.co[i1],ww.si[i1]*((ip2 & (j+1))?-1:1)); )
         nhm(tbase_xor(tbase_xor(ib1,ia,j),ib,j),r1,hbuf);
       }}
#endif
    } /* for (i1=0;i1<nw;i1++) all dual-bonds */
#if CM(0,0,1)    /* tU */
#if Nv             /* four-fermion-term, inter site repulsion+exchange */
    for(i1=0;i1<Nw;i1++){          /* wv: uu,ud,du,dd,u3,d3,3u,3d,33 */
      ia=iw[i1][0];ip1=geti(ib1,ia);
      ib=iw[i1][1];ip2=geti(ib1,ib);
#if Nv & 4    /* code activated if Nv is 4 for performance reasons */
      /* hole-hole repulsion, hole = empty site = (1-niu*nid) */
      /* this Hamiltonian is used by Fatiha Ouchni, Jul2004 */
      if (ip1==0 && ip2==0) sun+=para[ww.wh[i1]];
#endif /* if Nv & 4 */
#if Ham3
#error "Code not verified. Is the sign correct?"
      if(ip1*ip2==0)if((ip1^ip2)==3) /* .3,3. -JH Oles (jxy==jz only?) */
        nhm(Flip2(ib1,ia,ib),-para[ww.wj[i1]],hbuf);
        /* ensures symmetry */
#endif  /* if Ham3 */  
      /* ToDo: more general would be a 4x4 matrix ww.repulsion with
       *       sun+=para[ww.wv[i1]]*ww.repulsion[ip1][ip2]; (less if's)
       * transform ip1,ip2 to symmetric quantum numbers
       *           nij_sum=ni+nj-2=-2..2, nij_diff=ni-nj  z=nu-nd=-2..2 */
      if((ip1>0)&&(ip2>0)){
        if(ip1&1)if(ip2&1)sun+=para[ww.wv[i1]]+para[ww.wj[i1]];/* uu  U+JH */
        if(ip1&2)if(ip2&2)sun+=para[ww.wv[i1]]+para[ww.wj[i1]];/* dd       */
        if(ip1&1)if(ip2&2)sun+=para[ww.wv[i1]];		/* ud   U    */
        if(ip1&2)if(ip2&1)sun+=para[ww.wv[i1]];		/* du        */
        if((ip1^ip2)==3)				/* ud,du  JH */
         nhm(Flip2(ib1,ia,ib),para[ww.wj[i1]],hbuf);}
    } // for(i1=0;i1<Nw;i1++)
#endif /* if Nv */
#endif /* tU */
#if CM(0,1,1)    /* tU, 2015-09-23 onsite-term added to tJ */
    for(i1=0;i1<nn;i1++){ ip1=geti(ib1,i1);
      sun+= ((ip1&1)+((ip1&2)>>1)) * para[ww.we[i1]];  /* E*(nu+nd) */
      /* 4Ham2: (2nu-1)(2nd-1)  .. .u d. du =  1 -1 -1  1 */
      /* ~Ham2:  U nu nd        .. .u d. du =  0  0  0  1 */
      sun+= (Ham2)*.25*((ip1%3!=0)?-1:1)*para[ww.wu[i1]]; /* U(nu-.5)(nd-.5) */
      sun+=  (1-Ham2)*CM(0,0,1)*(ip1==3)*para[ww.wu[i1]];     /* U nu nd */
    } // nn
#endif /* CM(0,1,1) tU */
 /* NoSzSym results from  Kxx!=Kyy Bx,By!=0 Dxyz!=0, complex: By,Dx,Dz */
 /* except Kzz,Bz all of-diagonal-terms */
  /* S^+=Sp=Sx+iSy S^-=Sm=Sx-iSy    Sp=S^\plus Sm=S^\minus
   *  Sx=(Sp+Sm)/2  Sy=-i(Sp-Sm)/2
   *  pauli: sx=[0,1][1,0] sy=[0,-i][i,0] sz=[1,0][0,-1]  spin=[n_up,n_dn]
   * Zeeman-Term    Sum B.Si = Sum Bz*Siz + Bx*Six + By*Siy  = user-wishlist
   * mit negativen Vorzeichen, dann positive Korellation bei positiven fx,y,z
   *   one-site-term   -( Bz*Siz + Bx*(Sip+Sim)/2 -i By*(Sip-Sim)/2 )
   *             Hf  = -( Bz*Siz + 0.5*(Bx-iBy) Sip + 0.5*(Bx+iBy) Sim )
   */
  /* ToDo: testcase HFIELD +NOSZSYM +CPLX [wf,wfx,wfy] rename wf to wfz */
#if HFIELD
    for(i1=0;i1<nn;i1++){ ip1=geti(ib1,i1);
      sun-=((ip1&1)-(ip1>>1&1))*para[ww.wfz[i1]]/2; /* - (nu-nd)/2 */
# if NOSZSYM /* 2017-12 NoSzSym, see below, terms from user wishlist */
     if (ip1==1 || ip1==2) /* up to down + down to up */
       nhm(Flip1(ib1,i1),                    -0.5*para[ww.wfx[i1]] ,hbuf);
#  if VecType & 4 /* cmplx needed ToDo: check wfy=0 for non-complex */
     if (ip1==1 || ip1==2) /* ip1==1: +i By Sm(up), ip1==2: -i By Sp(down) */
       nhm(Flip1(ib1,i1), (1-2*(ip1-1))*C(0,-0.5)*para[ww.wfy[i1]] ,hbuf);
#  endif /* cmplx */
# endif /* NoSzSYM */
    } /* i1 = nn-loop */
#endif /* HFIELD */

#define S_DEBUG(a) a
#ifdef SSANISO
#ifndef CONFIG_S1SYM
# error "defining SSANISO without CONFIG_S1SYM makes no sence"
#endif
    { /* this code was introduced by Reimar Schmidt, Jan 2004 */
      int k,i2; double oZi; static unsigned int dbg=0;
      S_DEBUG( char sbuf2[SZccfg]; )
      /* oZi - eigenvalue of S^z_i of a composed spin
       *  Di - ww.wd index to float parameter (defined in daten.def)
       *  code revised by Joerg v2.20 */
      /* num_lm is N for a s=1 chain of length N build by NN=2*N spins s=1/2 */
      for(k=0;k<sym.num_lm;k++){  /* k = index of a composed spin-s object */
        for(oZi=0.0,i2=NN,i1=0;i1<nn;i1++)  /* sum S^z(i) over a composed spin */
          if(sym.lm[i1]==k) { /* check if spin i is member of composed spin */
            oZi+=(3-2*geti(ib1,i1))/2.0;
            if (i1<i2) i2=i1; /* get index of 1st member of composed spin */
          }
        sun+=oZi*oZi*para[ww.wd[i2]]; /* add to the diagonal element */
        S_DEBUG( if (dbg++<10) printf("\n dbgD k=%2d oZi=%4.1f d[i]=%.2f %s",\
          k, oZi, para[ww.wd[i2]],ccfg2(ib1,sbuf2));)
      }
    }
#endif
#if CM(1,0,0)
    if (nq) {  /* four-site ring term, see op_rot4() too */
      int i2; tbase ib2;
    /* sample:
     *             1      J12=J23=J34=J41, J13, J24
     *           / | \    Q1234 = four-site term (cyclic exchange)
     *          /  |  \   ferri state: S_total=0, S13=S24=1 EF=-2J12+J13/2+Q1234
     *         4---+---2      J24 term? (Nedko-Solution)
     *          \  |  /
     *           \ | /
     *             3
     */
      for (i1=0;i1<nq;i1++) {
        ib2 = ib1;
        for (i2=0; i2<4; i2++) {
          ip1 = geti(ib1, iq[i1][i2]);             /* take site i */
          ib2 = puti(ib2, iq[i1][(i2+1)%4], ip1);  /* put to site i+1 */
        }
        /* rotation */
        nhm(ib2,para[ww.wq[i1]],hbuf);
         ib2 = ib1;
        for (i2=0; i2<4; i2++) {
          ip1 = geti(ib1, iq[i1][i2]);             /* take site i */
          ib2 = puti(ib2, iq[i1][(i2+3)%4], ip1);  /* put to site i-1 */
        }
        /* inverse rotation */
        nhm(ib2,para[ww.wq[i1]],hbuf);
      }
    }
#endif
#ifdef EDIAG
    sun+=EDIAG;  /* add 1000.0 if energy is -1000.00365 for better results */
#endif
    /* diag <iHi> is the first entry, but sym(H*v) could be diag too */
    hbuf->rr[line_start]= sun; /* set diagonal value <nHn> v2.35 */
    /* nzcnt[i]=ofs[i+1]-ofs[i] is incl. diag element? line_start = line_ofs */
    hbuf->nzofs[iahead]= hbuf->num_nzAH; /* replace XY_FLAG by nz_ofs[AH] */
    line_start= hbuf->num_nzAH; /* store EOL position for next line */
    /* compression is bad for vector computing, ToDo 2015-10-08 */
    hbuf->jj[hbuf->num_nzAH-1] |= XY_FLAG; /* mark EOL, replace by nzofs! */
  } // loop iahead
  if (hbuf->max_nzAH_overflow < hbuf->nzAH_overflow) { /* FATAL */
      hbuf->max_nzAH_overflow=  hbuf->nzAH_overflow;
    /* This is fatal, but we may get more info about correct NZXMAX later
     *  for large computations, think about and correct NZX by hand is cheaper
     *    ToDo: test mpi*pt + overflow min messages? */
    E3("nzAH overflow, plz enlarge NZXMAX by %d/AH= %d mpi=%d",\
      hbuf->nzAH_overflow,\
     (hbuf->nzAH_overflow+NUM_AHEAD_LINES-1)/NUM_AHEAD_LINES, mpi_i);
    if (mpi_n*pt_n>15) F0("AVOIDING vasted CPUh"); // 2018-01 estimate NZX!
  }
  return(0);                /* old: return nHv or nH1(?), new return err */
} /* stored nzofs[AH] compressed sparse matrix elements */

/* ----------------------------------------------------------------
 * m1=mode: 0=<n|H*v0> (return nhn too)
 *          1=<n|H*n>   obsolete! used for 2x2-algo
 * m1=mode: 2=storeh,
 *          4:buf_..=H|n> without_sym+idx ... (called by storeh2)
 *          8:buf_..=H|n> full
 * ToDo: put i into hbuf (reduce arguments)
 * ToDo: split nhv to hbuf=H*|n> (H.elements) and v0*hbuf (v.elements)
 * input: (nahead configs)
 *    l1[0..(node_n1-1)]
 *    hbuf->blk= 0..(pt_n-1) // add mpi*pt_n if needed (b_ofs+b_len)
 *    i  = line index (better, use range to reduce calls)
           0<=i<b_ofs[blk+1]-b_ofs[blk]-1 (0..thread_len)
 *             move to hbuf->ofsAHblk  = 0..(node_n1-1)
 *    hbuf->v0 = NULL(vector.c,spins.c) or v0(xval.c_rh,diag.c) (can be moved to hbuf?)
 *    hbuf = work buffer (something set?)
 *    m1 = mode (see above)
 * modified: hbuf.{}
 *    hbuf.{(rx,nhn if v0 set),n,{acj,jj,rr}[0..n-1]} if v0==NULL
 *     where rr= H[ii,jj] (ii implicite given by counter + jj-flag)
 *    hbuf.{nhv[ahead],nhn[ahead] if v0 set)}
 *     where nhv[ii]= sum_jj H[ii,jj] |jj>    or ii=ix jj=iy
 * must be called by each mpi-process and thread to avoid deadlock (also n1=0)
 * called by:
 *   diag.c:diag2_thread                  < n|H|v0> <n|H|n> m1=0 v1=0?
 *   xval.c:wh                            <v0|H|v0>         m1=0 v1=0?
 *   xval.c:whq                           |H*v0-rh*v0|      m1=0 v1=0?
 *   spins.c:hamilton_geth_block=hamilton i<len[blk] v1+=H*v0 <v0Hv0> m1=0
 *   vector.c:randvec                     i<len[blk] <nHn> v0=0 m1=0
 *   spins.c:storeh_block=storeh          i<len[blk]       v0=0 m1=2
 *   spins.c:storeh2                      i<n1             v0=0 m1=4 (no valid l1)
 *   spins.c:fulldiag         hbuf[nHm..] i<len[blk]       v0=0 m1=8
 * nahead: num H lines to store, nahead=1..NUMAHEAD (1=old version)
 *         0 means no but call MPI block for balance v2.35
 *         the idea of nahead>1 is to reduce hamilton_nhv calls (-pg + gprof)
 * iHi was last element of line Hi, now its first element v2.35
 * ToDo 2015-02: allways fill 2AHEAD buffer and return nahead results from buffer
 * ToDo17: split hamilton_nhv(i,AH,hbuf,m) into
 *  {compute_nzH_blk,scfg_AHblk,scfg2idx_AHblk_useMPI,storeNZ}|loadNZ,addmult_nz_AHblk
 *  prepare for async I/O
 * ---------------------------------------------------------------- */
cplx hamilton_nhv(ULong i, ULong nahead, /* i=0..node_n1-1, loop= i..i+nahead */
       thbuf *hbuf, int m1) { /* input/output+buffer, m1=mode (ToDo obsolete) */
  ULong iahead;
  //  hbuf->sci.o=0;                        /* minsymcfg option */
  // ToDo: {compute_nzH_blk,scfg_AHblk,scfg2idx_AHblk_useMPI,storeNZ}|loadNZ,addmult_nz_AHblk
  hbuf->num_nzAH=0; /* number of stored nz-elements in ahead_nz_buffer */
#if DBG_NHM  /* HBLrVS = (NZXMAX*NUM_AHEAD_LINES) modVS rounded up */
  // bad for bug-search?
  { int k;for (k=0;k<HBLrVS;k++){/* better debug output *hbuf 2015-09 */
     hbuf->asj[k]=0; hbuf->an2[k]=0; hbuf->jj[k]=0; hbuf->acj[k]=TBASE0;} }
#endif
  // printf("L%d mpi=%d n1=%lu nahead=%lu\n",__LINE__,mpi_i,node_n1,nahead);
  /* 2015-10-08 add hbuf->cfg,nzx for simpler code + allow early benchmark */
  /* copy thread specific l1[node_n1]-part to hbuf, or filled by speed_test */
  /* checked: nahead is AH or smaller on last blk, i is 0,1,2,..*AH */
  if (nahead > b_ofs[hbuf->block+1] - b_ofs[hbuf->block] - i ) // last block may be smaller
      nahead = b_ofs[hbuf->block+1] - b_ofs[hbuf->block] - i;
  if (i >= b_ofs[hbuf->block+1] - b_ofs[hbuf->block]) {
     nahead=0; hbuf->lenAHblk=0; } // 2017-02-22 fix bug bad zero blocks
  hbuf->lenAHblk= nahead; /* 2017-02-05 last block may be smaller */
  hbuf->idxAHblk= i/NUM_AHEAD_LINES; // ToDo17 better outside?
#if DBG_NHM
  if ((vvv&4) &&
     (((mpi_i==mpi_n-1 && hbuf->block==pt_n-1) /* last thread or ... */
     ||(mpi_i==0       && hbuf->block==0))     /* ... first thread */
    && (i/NUM_AHEAD_LINES<2 ||  /* first 2 + last 2 */
        i/NUM_AHEAD_LINES>=(b_ofs[hbuf->block+1]
                           -b_ofs[hbuf->block  ])/NUM_AHEAD_LINES-1)))
  printf("\nDBG.nhv    %2d.%2d i,nAH=%6lu %4lu b_ofs[pt..pt+1]=%6lu %6lu",
    mpi_i,hbuf->block,i,nahead,b_ofs[hbuf->block],b_ofs[hbuf->block+1]);
#endif
  if (l1) { /* not on speed_test16 ToDo:loopNAHEAD return next nahead */
    for (iahead=0;iahead<nahead;iahead++)
        { hbuf->cfg[iahead]=getl1(i+iahead+b_ofs[hbuf->block]); }
    hbuf->cfg[nahead]= TBASE0; /* set OVL-error check boundary */
  }
  /* get H lines ahead to group nonlocal memory access */
  /* 0 <= hbuf->block < pt_n, 0 < i <= b_ofs[blk+1]-b_ofs[blk] */
  /* ToDo17 splitto {compute_nzH_blk,scfg_AHblk,scfg2idx_AHblk_useMPI,storeNZ}|loadNZ ... */
  /* ---- async IO + CPU ---- (numAHrounds + iAHround needed, 2-3pipes) */
  //    extra wait_all if checkpoint, keep it simple
  // if(i==0) next=hbuf;
  // if(i==0) compute_nzH_blk(i,nahead,hbuf);
  // if(i==0) compute_dest_nodes+dest_cnt(hbuf); // O(AH*nz) 
  // if(i==0) IA2a_A2av_cnt(hbuf); // O(mpi_n)
  // if(i==0) Sort_local_remote_scfg(hbuf); // O(AH*nz)
  // if(i==0) WA2a_A2av_cnt(hbuf); // wait
  // if(i==0) IA2av_scfg(i+nahead,nahead,hbuf); // O(AH*nz) Sendrecv_cfg
  // hbuf=next; next=hbuf->next; 
  // next->i=i+nahead: hbuf->nahead=nahead; iAHround++;
  //  rename op_H to compute_nzH_and_scfg
  compute_nzH_blk(i,nahead,hbuf); /* hbuf->... = operator_H(hbuf->cfg); AH-block */
  // if(i<n-1) load_nz_i_r(next) // ToDo: load_local_nz_i+load_nz_r (local only?)
  //        or { compute_nzH_blk(next->i,nahead,next); // in:cfg, out:nzr,nzcfg
  //             compute_cfg2scfg(next); // in:nzcfg out:nzscfg(dbg:jj,sgn)
  //             put_num_nzAH_to_table_num_nzAH[numAHrounds] for nz_storage
  //             store_nz_i_r(next); } // instead of local+remote local_only(below!)
  /* m1&4: storeh2 = no l1[] for cfg2idx() available; */
  nhm_line(hbuf, m1); /* ToDo15: split scfg2i2vi, rm m1 */
  // if(i<n-1) compute_dest_nodes+dest_cnt(next); // O(AH*nz) 
  // if(i<n-1) IA2a_A2av_cnt(next); // O(mpi_n)
  // if(i<n-1) Sort_local_remote_scfg(next); // O(AH*nz)
  // if(i<n-1) WA2a_A2av_cnt(next); // wait
  // if(i<n-1) IA2av_scfg(i+nahead,nahead,hbuf->next); // O(AH*nz)
  // if(i<n)   WA2av_scfg(hbuf); // wait
  // if(i<n)   compute_scfg2i2vi(hbuf) // ToDo: next2 
  //           rnd_read_v[nzi]        // ToDo: parallel ALU+MEM=Hyperthreading 
  // if(i<n)   IA2avBack_i_or_vi(hbuf) // ToDo store_i_remote
  // if(i<n)   Wait_i_or_vi(hbuf)
  // if(i<n)   store_nz_i_r // old: remote(bad)+local elements
  //        or compute_v0_add_Hnz_mul_v1 or IStore_Hnz_i
  // ...
  /* vvv&2 makes SH factor 2 slower on V490-sparcIV+ cache? v2.25? */
  if (vvv&2) add_statH(hbuf); /* H-statistics for debugging */
  // if (ah_nz[i/AH]==0) ...puth..., ah_nz[i/AH]=hbuf->num_nzAH;
  if (m1&2) hamilton_puth_line(hbuf); /* write to H-file, ToDo = above */
    /* ToDo: alternativ do sum <v1|H|v0>  outside? */
  hbuf->ofsAHblk=i+nahead;   /* index of configuration 0..b_len[blk]-1, set outside!? */
#if 0 /* old bad stuff ??? */
  for (ii=0; ii<hbuf->num_nzAH; ii++) {
    if (hbuf->jj[ii]==i && hbuf->abj[ii]==mpi_i)
        hbuf->nhn[i] += D(hbuf->rr[ii]);    /* <nhn> */
    if (hbuf->jj[ii]&XY_FLAG) i++; // next matrix line
  }
  if (hbuf->lenAHblk && hbuf->v0) sum= hbuf->rr[i%hbuf->lenAHblk];
#endif
  return(0);                /* old: return nHv or nH1(?), new return err */
} /* hamilton_nhv */


/*
 * void-function for pthreads, read a complete H-block and do v1=H*v0
 * return hbuf->rx=<v0|H|v0> real
 * input: hbuf->{block,v0,v1}
 * working:  ToDo: MPIperLine hbuf.n,hbuf.{jj,abj,rr}[]=H|v0>
 * output: hbuf->rx=<v0|H|v0>, v1+=H*v0
 *  called by hamilton() for diag.c:inc1
 *
 * xx is only correct for the first block ?! 
 * h_get was put into geth_block (v2.26, keywords: h_get, hget, geth, get_h)
 * for better speed, this part should be read in ca. 1000 values from vy, 
 *   than from vx0, than store to vx1 to use the CPU cache optimal
 * ToDo: h_ropen/close h_files before calling this
 * better rename: hamilton_using_stored_Hblks?
 */
void hamilton_geth_block( void *ptr ) { /* thbuf is in prinzip not needed! */
  ULong i,  /* loop variable */
        ihblk, nhblks; /* stored hblks == row-blocks since v2.56 */
  thbuf *hbuf=(thbuf *)ptr;
  const int block=hbuf->block; /* thread_block index 0..pt_n-1 */
  int ii, in; /* loop inside blk = 0...nz-1 */
  /* ToDo: replace by hbuf->idxAHblk = 0..node_n1/AH */
  ULong xx, xy;
  scplx *v0x,*v1x;   cplx r, sum_ry;
  struct thxy *tbh; // ,*tbh2;  /* H storage buffer [B_NUM].{xy,by,rr}[HBLen] */
// ToDo17
  static scplx buf_vy_iy[B_NUM][HBLen]; /* vector elements v0[iy] received */
  const int tblk = mpi_i*pt_n + block;
  if (block>=pt_n) ERR(9999);
  /* tbh points initialy to buffer space for read/write */
  tbh=&(b_hbuf[block]);  /* ->{xy,by,rr}[HBLen] */
  v0x=hbuf->v0 + b_ofs[block]; /* v2.36 */
  v1x=hbuf->v1 + b_ofs[block];
  xx=0;      /* row of matrix xx=0..b_len-1 */
   /* ToDo: b_sn1[] = number cfgs, where H|cfg> is stored */
  sum_ry=0;
  nhblks= min_stored_ah_blks; /* mem+dsk Hblocks stored per thread */
  // nhblks=(min_b_sn1+NUM_AHEAD_LINES-1)/NUM_AHEAD_LINES; ??? better
  ii = 0; /* number hnz element of block, ii<HBLen (int) */
  i  = 0; /* i is only the hnz counter, i<b_len[gblock] */
  if ((vvv&16) && ((tblk&(tblk-1))==0 || tblk+1==mpi_n*pt_n))
    printf("DBG.geth %3d.%-2d stored_ah_blks=%6lu / %-6lu min_stored=%6lu\n",
    mpi_i, block, stored_ah_blks[block], node_ah_blks/pt_n,
              min_stored_ah_blks);
  /* ToDo: better one loop over all blocks + switch mem/dsk/recalc */
  /*   but then mpi-data must switch too, send_scfg_vs_idx */
  for (ihblk=0; ihblk<nhblks; ihblk++) { /* ensure mpi balance */
    // in=b_snz[block]; /* number of stored elements per thread ??? */
    in=ah_nz[block*(node_ah_blks/pt_n)+ihblk]; /* nz per thread_AHblock */
    /* ToDo: in*Hsize/HBLen overflow? safe against -O2? (uint64_t) or 1LL */
    if (in*1LL*Hsize/HBLen>~(size_t)0) ERR(9999); // 2017.04 -m32 uint64_t
    /* need to give (in=len=ah_nz[blk])*8B as memHnext-argument */
    /* ToDo: Hsize/HBLen shoulb be replaced by sizeof(tbh2->mz[0])? */
    tbh=(struct thxy *)memHnext(in*1LL*Hsize/HBLen,block); /* old pipe.c, 2017-01-28 */
    /* ToDo17: replace h_read by my_fread */
    if ((!tbh) && opt_maxfile)
      tbh=h_read(&(b_hbuf[block]),in*1LL*Hsize/HBLen,block); /* v2.36 Jul08 */
    if (!tbh) { static int thberr=0; /* print only once */
      if (!thberr) E2("geth_block.h_read %5lu %5lu",ihblk,nhblks);
      thberr=1; break; } /* MPI deadlock? ToDo17: recompute on fail */
    /* copy random scattered local v0y data to the mpi buffer */
    /* ToDo: unify next loop with MPI part, simpler code */
    if (mpi_n==1) { /* get vector elements v0[iy[ii]], rnd read + seq write */
      for (ii=0l; ii<in; ii++) {
        xy = (ULong)tbh->nz[ ii ].xy; /* incl. XY_FLAG, .by==0 */
#if 0 || DBG_NHM   // slow-debug, err+segfault on bad read, use checksum?
        if ((xy & ~XY_FLAG) >= node_n1)
           printf("\nERR xy-ovl ii=%d xy=%016lx\n",ii,xy);
#endif
        buf_vy_iy[block][ii] = hbuf->v0[xy & ~XY_FLAG]; /* scplx */
      } /* for (i=0..hnz,ii=0..HBLen) */
      if (ii<HBLen) buf_vy_iy[block][ii] = 1e32; /* make bug visible */
    } /* if mpi_n==1   single task */
#ifdef CONFIG_MPI /* now random read becomes remote node random read */
    if (mpi_n>1) { /* should work for mpi_n=1 too, but slower */
      /* sort vector indizes by destination nodes, get back vector elements */
      int len1, len2, id, i1, by=0, rc=0;   ULong yy;
      static int buf_idx[B_NUM][HBLen]; /* sort index for send yy and vr */
      void *psrc, *pdst; /* src,dst for better looking code */
      MPI_Datatype mpi_dtype=MPI_BYTE;
      /* 1. get dest node for each element (already done earlier, by[]) */
      /* 2. count data per detination node, needed for reorder */
      for (id=0;id<=mpi_n;id++) bsend_cnt[block][id]=0;
      for (i1=0l; i1<in; i1++) {
        by = tbh->nz[ i1 ].by; bsend_cnt[block][by]++; } /* count for each node */
      /* generate offsets from counts (last index (n) is HBLen) ... */
      bsend_of2[block][0]= bsend_ofs[block][0]= 0;  /* ini private ofs */
      pthread_barrier_wait(&pt_barrier);  /* 2017-01-18 wait all threads */
if (block==0) { int it,ofs; /* master thread only collects all ofs + cnt */
      nsend_ofs[0]=0;  /* ini private ofs */
      for (id=0;id<mpi_n;id++) { /* runs one more, to store overall size! */ 
        ofs=bsend_ofs[0][id]=nsend_ofs[id];
        for (it=1;it<pt_n;it++) { // gcc492 -unroll warns ofs[it] for BNUM=1
           ofs += bsend_cnt[it-1][id]; bsend_ofs[it][id]= ofs; }
        nsend_ofs[id+1] = bsend_ofs[pt_n-1][id] + bsend_cnt[pt_n-1][id];
        nsend_cnt[id]= nsend_ofs[id+1] - nsend_ofs[id];
      } /* size per destnode is not equal distributed */
      nsend_cnt[mpi_i]=0; /* do not send local data 2017-02-24 */
      if (nsend_ofs[mpi_n] > B_NUM*HBLen*(0+BFACTOR)/100) ERR(9999);
      ckpt_help(); /* MPI_Bcast(signal) + MPI_Barrier + sleep30 */
      /* do this non-blocking before sorting data to _of2, Wait here */
      rc=My_MPI_Ialltoall(nsend_cnt,1,MPI_INT,              /* geth_block */
                          nrecv_cnt,1,MPI_INT,
            MPI_COMM_WORLD,&a2a_request); if(rc)perror("geth.A2a");
}  /* block==0 */
      pthread_barrier_wait(&pt_barrier);  /* 2017-01-18 wait all threads */
      /* ... now we sort in groups for each destination */
      for (id=0;id<mpi_n;id++) bsend_of2[block][id]= bsend_ofs[block][id];
      // if (ihblk<nhblks) /* not for dummy blocks */
      for (i1=0l; i1<in; i1++) {
         xy = (ULong)tbh->nz[ i1 ].xy; /* incl. XY_FLAG */
         by =        tbh->nz[ i1 ].by; /* destination node 0<=b<mpi_n */
         len1 = bsend_of2[block][by]; /* get index to next free element */
         buf_idx[block][i1] = len1;  /* sorting back index, invers stored */
         nsend_iy[len1] = (xy & ~XY_FLAG); /* cfg2idx 0..node_n1-1 */
         bsend_of2[block][by]++; /* new position, ToDo: rename */
      } /* for (i1=0..HBLen), yy data ready to send */
      pthread_barrier_wait(&pt_barrier);  /* 2017-01-18 wait all threads */
      // ToDo: MPI_IAlltoall + compute local data 50% + Wait (as simple ver?)
if (block==0){ int ofs;
      /* v2.53 this is called pt_n times, now one time = faster! */
      My_MPI_Wait_Ia2a(&a2a_request,&mpi_status);
      // printf("\nDBG: %d %d",mpi_n,MPI_MAX_NUM);
      for (nrecv_ofs[0]=ofs=0,id=1;id<=mpi_n;id++) { // use ofs against gcc412-bug
        ofs += nrecv_cnt[id-1]; nrecv_ofs[id] = ofs; }
      dbg_mpi("MPI.it1",nsend_cnt,nsend_ofs,nrecv_cnt,nrecv_ofs,hbuf);
// --- here we are at i100-iteration-section = bottleneck ---
      len1= nsend_ofs[mpi_n]; psrc= &nsend_iy[0]; /* 0..HBLen-1 */
      len2= nrecv_ofs[mpi_n]; pdst= &nrecv_iy[0]; /* .lt. (BFAC+1)*HBLen */
      if (len1>B_NUM*MPI_NZ_BUFSZ1 || len2>B_NUM*MPI_NZ_BUFSZ2) {
        printf("\nERR.geth %2d len1=%d len2=%d /HBLen*100=%d",mpi_i,len1,len2,
               len2*100/HBLen); ERR(9999); } // 2017-02-13 bughunting
  /* using type_iy = int here, on overflow use more tasks!
   * ToDo17 scale_f1(nsend_ofs,nsend_ct,nrecv_ofs,nrecv_cnt,mpi_n,XYsize) */
      len1=sizeof(XY_TYPE);
      if (len1== 2) mpi_dtype=MPI_SHORT; else /* MPI_Datatype */
      if (len1== 4) mpi_dtype=MPI_UNSIGNED; else /* MPI_Datatype */
      if (len1== 8) mpi_dtype=MPI_LONG_LONG; else
      if (len1==16) mpi_dtype=MPI_LONG_DOUBLE; else { mpi_dtype=MPI_BYTE;
        for (ii=0;ii<mpi_n;ii++) nsend_cnt[ii] *= len1;
        for (ii=0;ii<mpi_n;ii++) nsend_ofs[ii] *= len1;
        for (ii=0;ii<mpi_n;ii++) nrecv_cnt[ii] *= len1;
        for (ii=0;ii<mpi_n;ii++) nrecv_ofs[ii] *= len1; }
      rc=My_MPI_Ialltoallv(psrc,nsend_cnt,nsend_ofs,mpi_dtype,\
                           pdst,nrecv_cnt,nrecv_ofs,mpi_dtype,\
         MPI_COMM_WORLD, &a2a_request); if(rc)perror("geth.A2av");
      if (len1!=2 && len1!=4 && len1!=8 && len1!=16) {
        for (ii=0;ii<mpi_n;ii++) nsend_cnt[ii] /= len1;
        for (ii=0;ii<mpi_n;ii++) nsend_ofs[ii] /= len1;
        for (ii=0;ii<mpi_n;ii++) nrecv_cnt[ii] /= len1;
        for (ii=0;ii<mpi_n;ii++) nrecv_ofs[ii] /= len1; }
      My_MPI_Wait_Ia2a(&a2a_request,&mpi_status);
} /* block==0 */
      for (i1=0l; i1< bsend_cnt[block][mpi_i]; i1++) {  /* local data only */
        yy = nsend_iy[bsend_ofs[block][mpi_i]+i1];
        nsend_vy[     bsend_ofs[block][mpi_i]+i1] = hbuf->v0[yy]; }
      pthread_barrier_wait(&pt_barrier);  /* 2017-01-18 wait all threads */
      len2= (nrecv_ofs[mpi_n]+pt_n-1) / pt_n;
      len1= block*len2; // ofs ... check OVL more threads than matrixsize
      if (len1     >nrecv_ofs[mpi_n]) len1=nrecv_ofs[mpi_n]; 
      if (len1+len2>nrecv_ofs[mpi_n]) len2=nrecv_ofs[mpi_n]-len1;
      /* slow random access, saturates for example 4 memory channels  */
      for (i1=0l; i1<len2; i1++) {  // approx HBLen/mpi_n (wasting memory)
        yy = nrecv_iy[len1+i1]; /* index num 0..node_n1-1 */
        nrecv_vy[len1+i1] = hbuf->v0[yy]; } /* for (i1=0..HBLen) */
      pthread_barrier_wait(&pt_barrier);
if (block==0) {
      // send vector elements back to the caller node
      psrc=&nrecv_vy[0];  pdst=&nsend_vy[0];  len1=sizeof(scplx);
      // mpi_tmod(len1,&dtype,nsend_cnt,nsend_ofs,nrecv_cnt,nrecv_ofs);
      if (len1== 2) mpi_dtype=MPI_SHORT; else /* MPI_Datatype */
      if (len1== 4) mpi_dtype=MPI_UNSIGNED; else /* MPI_Datatype */
      if (len1== 8) mpi_dtype=MPI_LONG_LONG; else
      if (len1==16) mpi_dtype=MPI_LONG_DOUBLE; else {
        for (ii=0;ii<mpi_n;ii++) nsend_cnt[ii] *= len1;
        for (ii=0;ii<mpi_n;ii++) nsend_ofs[ii] *= len1;
        for (ii=0;ii<mpi_n;ii++) nrecv_cnt[ii] *= len1;
        for (ii=0;ii<mpi_n;ii++) nrecv_ofs[ii] *= len1;
        mpi_dtype=MPI_BYTE;}
      rc=My_MPI_Ialltoallv(psrc,nrecv_cnt,nrecv_ofs,mpi_dtype,\
                           pdst,nsend_cnt,nsend_ofs,mpi_dtype,\
            MPI_COMM_WORLD, &a2a_request); if(rc)perror("geth.A2av.back");
      My_MPI_Wait_Ia2a(&a2a_request,&mpi_status);
} /* old Sendrecv-loop id=0..mpi_n-1 */
      pthread_barrier_wait(&pt_barrier);
      /* sort back results from remote note */
      /* v2.43 was slower random write, now seq write, ToDo backport */
      for (i1=0l; i1<in; i1++)
        buf_vy_iy[block][ i1 ]= nsend_vy[ buf_idx[block][i1] ];
      if (i1<HBLen) buf_vy_iy[block][i1] = 1e32; /* make bug visible */
    } /* MPI mpi_n>1 */
    // if ((vvv&4) && hbuf->idxAHblk<4) printf("\nMPI_3 %3d %2d %lu %lu", hbuf->idxAHblk, mpi_i*pt_n+block, i, in); OK
#endif  /* MPI */
    for (ii=0l; ii<in; ii++) { /* innermost loop, NFLOP=5*hnz*NumIterations */
      xy=    (ULong) tbh->nz[ ii ].xy; /* idx 0..node_n1=n1/mpi_n [+XY_FLAG] */
#if HRMAX>0
      r = hr[block][ tbh->nz[ ii ].rr ];
#else
      r =     M_f2d( tbh->nz[ ii ].rr );
#endif /* end of h_get */
      /* 2013-04-04 M_f2d added, was buggy for float16 */
      sum_ry += r * M_f2d(buf_vy_iy[block][ii]);  /* <ii|H|v0> += <ii|H|yy> */
      if (!(xy & XY_FLAG)) continue; /* EOL (xx) of sparse matrix H */
      hbuf->rx += D(Conj(M_f2d(*v0x))*sum_ry); /* <vHv> should be real */
      /* 2013-04-04 avoid float16 += float16 ToDo: += (float16)float32 relevant? */
      (*v1x)    =   M_d2f(M_f2d(*v1x)+sum_ry); /* seq. r/w memory, flush cache? */
      xx++; v0x++; v1x++; sum_ry=0; /* next line */
    } /* for (i=0..hnz,ii=0..HBLen) */
  }  /* 0<=ihblk<nhblks */

  /* now handle the nonstored elements of H-matrix */
  if ((vvv&16) && mpi_i<4 && mpi_i>=mpi_n-4) /* v2.33 */
    printf("DBG.geth_block %3d.%-2d ah_blk= %lu/%lu\n",
           mpi_i, block, nhblks, node_ah_blks/pt_n);
  /* recompute unstored matrix elements (slow) */
//  if ((vvv&16) && mpi_i>=mpi_n-2)  // kautz s1_lc27 950*6t crash 2017-10
//      printf("DBG %3d.%-2d alive L%d\n", mpi_i, hbuf->block, __LINE__);
  /* old=unbalanced: 0 <= i < b_len[blk],
   * new=balanced:   min_b_sn1 <= i < max_b_len (zero padded ???) */
  for (ihblk=nhblks; ihblk<node_ah_blks/pt_n; ihblk++) {
    /* if (v0) return hbuf[i].rr = <i|H|v0> (true here), ignore i>=b_len */
    int len= NUM_AHEAD_LINES; /* 2017-02-05 len for readable code */
    i= ihblk*NUM_AHEAD_LINES; /* ULong */
    /* b_ofs[] is set in block_c.b_ini and hilbert_c.speed_tests */
    /* overflow thread_n1 - i if i > thread_n1   2017-10-20 */
    if (         i > b_ofs[hbuf->block+1] - b_ofs[hbuf->block]) len=0;
    else
    if ((ULong)len > b_ofs[hbuf->block+1] - b_ofs[hbuf->block] -i)
               len = b_ofs[hbuf->block+1] - b_ofs[hbuf->block] -i;
    hbuf->lenAHblk= len;
    // problem 2017-10 kautz 950*6t lc27_s1 ...
    //  ... hblk=n-2  len.thread0..4 = AH, len.thread5 < AH,           
    //  ... hblk=n-1  len.thread0..4 < AH, len.thread5 = 0 but AH used
//if ((vvv&16) && mpi_i+2>=mpi_n
//    && ((ihblk&(ihblk-1))==0 || ihblk+2>=node_ah_blks/pt_n))
//  printf("DBG %3d.%-2d alive L%d %d\n", mpi_i, hbuf->block, __LINE__, (int)ihblk);
    hamilton_nhv(i, len, hbuf, 0); /* includes MPI, thread function */
    for (ii=0; ii<len; ii++) {
      // r= hbuf->rr[ii];                      /* <ii|Op|v0> bug 2014-02 */
      r= hbuf->nhv[ii];                      /* <ii|Op|v0> fix 2014-02 */
      /* 2013-04-04 avoid float16 += (float16)r, use float16 (double+=) r */
      hbuf->v1[b_ofs[block]+i+ii] =
        M_d2f(M_f2d(hbuf->v1[b_ofs[block]+i+ii])+r); /* v1+= <ii|H|v0> */
      hbuf->rx += D(Conj(M_f2d(hbuf->v0[b_ofs[block]+i+ii]))*r); /* real <v0|H|v0> */
    }
  } if (hbuf->max_nzAH_overflow) ERR(9999);
//  if (mpi_i>=mpi_n-2)
//      printf("DBG %3d.%-2d alive L%d\n", mpi_i, hbuf->block, __LINE__);
} /* hamilton_geth_block() */

/*  v1=H*v0+v1 (parallel),  return r1=<v0|H*v0>=real
 *   (a-ib)*(c+id) = (ac+bd) + i(ad-bc) 
 * <r-i|Hr+Hi|r+i> = <r|Hr r>+<r|Hi i>+<i Hr i>+<i Hi r>
 * using long double ???
 * ToDo: hamilton_geth_block + hamilton_nhv_block => hamilton_block
 * called by diag.c:lanz1()
 *           diag.c:inc1()   v0+v1 set
 *           diag.c:incv1()
 *
 *   calls hamilton_geth_block(storedH) incl. hamilton_nhv_block(nonstoredH)
 * ToDo: Move pthread-code into MPI-loop
 */
/* v1=H*v0+v1, return <v0|H|v0> */
double hamilton(scplx *v0, scplx *v1, thbuf *hbuf[B_NUM]){
  // ULong i; 
  int block;
  /* thbuf * (hbuf)[B_NUM];  */
  double r2=0.0;  
  for (block=0; block<pt_n; block++) {
    // open on first read, no files written/touched if not needed 2018-11-11
    // if (opt_maxfile!=0) h_ropen(block); /* open I/O streams + buffer v2.26 */
    mopen(block,0); /* readonly */
#if 0 /* obsolete 2014-07 */
    hbuf[block] = alloc_hbuf(); /* better alloc local in thread? but free? */
    if (!hbuf[block]) {
      F2("ERROR hamilton: alloc_hbuf failed, mpi=%d thread=%d", mpi_i, block);
      exit(1);// 2014-07 continue;
    }
#endif
    reset_hbuf(hbuf[block]);   /* 2015-01-31 */
#if 0 && DBG_NHM  // 2017-02 bad to reproduce errors?
    { int i;for (i=0;i<HBLrVS;i++){/* better debug output *hbuf 2015-09 */
        hbuf[block]->asj[i]=0;
        hbuf[block]->an2[i]=0;
        hbuf[block]->jj[i]=0;
        hbuf[block]->acj[i]=TBASE0;} }
#endif
    hbuf[block]->block = block;
    hbuf[block]->v0 = v0;
    hbuf[block]->v1 = v1;
    hbuf[block]->rx = 0.0;
    parallel_start( block, hamilton_geth_block, hbuf[block] );
  }
  for (block=0; block<pt_n; block++) {
    /* if (hbuf[block]) */ parallel_wait(block); /* only for threads */
    if (opt_maxfile!=0) h_rclose(block); /* v2.26 close buffer chain and finaly the file */
    mclose(block);
    // if (!hbuf[block]) continue;
    r2 += hbuf[block]->rx;    /* <vHv> */
/*    free_hbuf(hbuf[block]); obsolete ToDo */
    if (hbuf[block]->max_nzAH_overflow) ERR(9999);
  }
#ifdef CONFIG_MPI
  { double add=r2, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    r2=sum; }
#endif
  return r2;
}

/* void-function for pthreads
 * ToDo: return compute time to calculate load balance
 *  input: blk=0..pt_n-1 (its cheaper to add mpi_i*pt_n as dividing by pt/n)
 * must be called for all blocks per mpi-process avoiding deadlocks
 */
void storeh_block( void *ptr ){ /* block means thread here */
  ULong i, iout, ihblk;
  double tm1, tm2; /* time in seconds 2013-04-23 */
  thbuf *hbuf=(thbuf *)ptr;
  const int block=hbuf->block,            /* node block    0..pt_n-1 */
      c_blk=mpi_i*pt_n+hbuf->block; /* cluster block 0..pt_n*mpi_n-1 */
  tm1=fstime();  /* 2013-04-23 itime (long) replaced by fstime (double) */
  iout=1<<6;  /* output alive msg sparingly, N=40: 2^29 8outs (old=1<<12) */
  /* one MPI transaction per loop, max(b_len[c_blk]) for balance */
  stored_ah_blks[hbuf->block]=0; /* reset AH-block counter (mem+dsk) */
  /* 2017-10-s1_lc27: nz(AHblk=0) differ factor 4, but t100% ok (mpi inside)
   *   AHblk_mpi_transfer/local_access grows with AHblock-index, t100%++
   *   ToDo: at log2(iAHblk) we could MPIMAX/MIN + better nzx debug output */
  for(ihblk=0; ihblk<node_ah_blks/pt_n/*max_b_len/AH*/; ihblk++){
    int len= NUM_AHEAD_LINES; /* 2017-02-05 added for better radable code */
    i= ihblk*NUM_AHEAD_LINES; /* generate start cfg index from AH-blk */
    if ((ULong)len > b_ofs[hbuf->block+1] - b_ofs[hbuf->block] -i)
               len = b_ofs[hbuf->block+1] - b_ofs[hbuf->block] -i;
    hbuf->lenAHblk= len; /* length of config block, mostly nAH, rest smaller */
    /* could change this to log2(ihblk) and use MPI for better approx nzx */
    if ((!c_blk) && iout && i>=iout) { /* give alive signals */
      /* ToDo: sync threads and output min/max? 2013 */
      tm2=fstime()-tm1; iout<<=2; /* factor 4, iout may overflow to 0 */
      if (tm2>2) /* old: 10s .. 40s */
        printf(" SH %6.2f%% hnz/i= %6.2f nzxmax= %3d nz/s= %d*%8.2e"
               " t100%%= %6.2fm\n",
                  i*100./b_ofs[1], b_hnz[0]*1./i,
                  (int)statH[0].nzxmax,       /* 2012-01-23 speed tuning */
                  pt_n*mpi_n, b_hnz[0]*1./tm2, /* 2014-06 speed nz/s */
                  tm2/60.*b_ofs[1]*1./i);
    }
    /* ToDo: return number of lines filled into the buffer? */
    /* ToDo: send/recv Hij elements to/from other nodes (cached) */
    /* storeH to disk
     * ToDo17: split hamilton_nhv(i,AH,hbuf,m) into
     *         {compute_nzH_blk,cfg2scfg,scfg2idx_useMPI,storeNZ}|loadNZ,
     *         addmultiply_nz_AHblk
     *         prepare for async I/O
     */
    hamilton_nhv( i, len, hbuf, 2);
  } /* loop i<max_b_len */
#if 0
  if ((vvv&4) && (c_blk<4 || c_blk>=mpi_n*pt_n-4)) 
    printf("\nvvv&4 SH_block.end %3d ah_blks= %9lu snz= %9lu hnz= %9lu",
             c_blk, stored_ah_blks[block], b_snz[block], b_hnz[block]);
#endif
  if (err_fatal && (mpi_i&(mpi_i-1))==0) /* 2015-02 log output */
    printf("\nERR%d: %4d.%2d blk=%lu= %6.2f%% h_nz=%.1e (SH_block.end)",\
       err_fatal,mpi_i,block, stored_ah_blks[block],
      stored_ah_blks[block]*pt_n*100./node_ah_blks,
        1.*b_hnz[block]); /* 2016-01-23 add b_sn1 */
  hbuf->rtime=fstime()-tm1;
  if (hbuf->max_nzAH_overflow) ERR(9999);
}

/* store H elements on disk, reading disk is faster than computing!? Aug07 */
void storeh(thbuf *hbuf[B_NUM]){
  int block;
  double tm1;               /* fstime() = MPI_Wtime() runtime in seconds */
  /* thbuf * (hbuf)[B_NUM]; ToDo: obsolete */
  double h_ns; /* sum stored h_nz */
  double     rtime_max,       rtime_min; /* balance + timing measurement */
  int  block_rtime_min, block_rtime_max; /*  thread index for min/max timings */
  rtime_max=rtime_min=0.0;
  block_rtime_min=block_rtime_max=0; /* avoid gcc warning */
  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  if (!mpi_i) {
    printf("\n --- starting hilbert matrix generation (storeh) ---");
    /* (buffersize-overhead)/numelements=size_per_element */
    printf("\n NZXMAX= %d  AHEAD_LINES= %d hnzSize=HSize/HBLen: %.0fK/%.0fK=%.2f",
      NZXMAX, NUM_AHEAD_LINES, Hsize/1024., HBLen/1024., Hsize*1./HBLen);
    /* meggie 2015-03 sq40 8+32 gcc492+SSE3+3dNOW */
    printf("\n vvv&2 (=%d) statH (speed-~5%%), vvv&(64+2) (=%d) prtH",
      vvv&2,vvv&64);
  }
  statH_ini();	/* initialize H-statistics Oct99 */
  ini_thxy(n1); /* see h_file.c, initialize H-storage, print size needed */
  min_stored_ah_blks=  node_ah_blks; /* ini to max possible thread blocks */
  max_stored_ah_blks=             0; /* ini to min possible thread blocks */
  min_Mstored_ah_blks= node_ah_blks; /* ini to max possible thread blocks */
  /* 2015-04-09 moved to h_file.c ini_thxy(): 
   *  if (!mpi_i) printf(" sizeof(HR_TYPE)=%d\n",(int)sizeof(HR_TYPE)); */
  tm1=fstime();
  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  if (checkpoint<4) { // old 3
    for(block=0;block<pt_n;block++) {
      /* 2018-11-11 h_wopen is called by first write to avoid empty files */
      // if (opt_maxfile!=0) h_wopen(block); /* open I/O streams + ini buffer */
      mopen(block,1); /* memory write only */
      /* big local vars (sbase), relatively bad for parallelisation */
      // hbuf[block]=alloc_hbuf(); /* ToDo: alloc data area within the thread */
      reset_hbuf(hbuf[block]); /* 2015-01-31 */
#if DBG_NHM  /* but may change buggy code paths = bad (ToDo) */
      { int i;for (i=0;i<HBLrVS;i++){/* better debug output *hbuf 2015-09 */
          hbuf[block]->asj[i]=0;
          hbuf[block]->an2[i]=0;
          hbuf[block]->jj[i]=0;
          hbuf[block]->acj[i]=TBASE0;} }
#endif
      hbuf[block]->block=block; /* overall block number */
      /* start compute threads, hbuf->v0=v1=NULL */
      parallel_start( block, storeh_block, hbuf[block] );
    }
    
    for(block=0;block<pt_n;block++){
      parallel_wait(block); /* stop compute threads and let them wait */
      /* extra write for last block, not needed since 2017-02-16 */
      if (opt_maxfile!=0) h_wclose(block);
      mclose(block);      /* flush H-buffers to disk */
      if (min_stored_ah_blks > stored_ah_blks[block]) /* get minimum M+D */
          min_stored_ah_blks = stored_ah_blks[block];
      if (max_stored_ah_blks < stored_ah_blks[block]) /* get maximum M+D */
          max_stored_ah_blks = stored_ah_blks[block];
      if (min_Mstored_ah_blks > memH[block].wrt) /* get minimum M */
          min_Mstored_ah_blks = memH[block].wrt;
      if (block==0) {
        rtime_max=rtime_min=hbuf[0]->rtime; /* balance + timing measurement */
        block_rtime_min=block_rtime_max=0;
      } else {
        if (rtime_min>hbuf[block]->rtime) {
            rtime_min=hbuf[block]->rtime; block_rtime_min=block+mpi_i*B_NUM;}
        if (rtime_max<hbuf[block]->rtime) {
            rtime_max=hbuf[block]->rtime; block_rtime_max=block+mpi_i*B_NUM;}
      } // free_hbuf(hbuf[block]);
    } // block-loop
    MPI_Barrier(MPI_COMM_WORLD); /* block to get walltime, ToDo, better get maxtime? */
    tm1=fstime()-tm1;
    /* ToDo: output time in a form where it can be removed for comparisions
     * example: diff <(sed 's/\([0-9]{2}\)/#1/g' out1) <(sed '' out2)
     * ToDo: output trace?
     */
    h_checkpoint(1); /* write hr_saveNNN.dat for STOREH!=0 */
    checkpoint=5; if(!mpi_i)fputs(" checkpoint5.w", stdout); // SH=45 old=3
    if(!mpi_i){FILE *f;f=fopen(FCHKPT,"wt");fputs("5\n",f);fclose(f);}
  } else h_restore(1); /* checkpoint>=4 read hr_saveNNN.dat */
  /* storeh complete, print statistics */
  
#ifdef CONFIG_MPI /* ToDo: MAX(i) = -MIN(-i) could be used to reduce MPI */
  { ULong  t1[6], t2[6], t3[6]; /* ULLong may be a problem for MPI */
    t1[0]=h_error;                t1[1]=min_stored_ah_blks; t1[2]=hdsk_nzmax;
    t1[3]=hbuf[0]->max_nzAH_recv; t1[4]=min_Mstored_ah_blks;
                                  t1[5]=max_stored_ah_blks;
    MPI_Allreduce(t1,t3,6,MPI_UNSIGNED_LONG, MPI_MAX, MPI_COMM_WORLD);
    h_error=t3[0]; hbuf[0]->max_nzAH_recv=(int)t3[3];
    t1[0]=hdsk_nzmax; t1[1]= min_stored_ah_blks;
    MPI_Allreduce(t1,t2,6,MPI_UNSIGNED_LONG, MPI_MIN, MPI_COMM_WORLD);
    if (t3[2]>t2[0] && !mpi_i) /* max > min */
      printf("\nWarning: hdsk_nzmax reduced by mpi from max=%lu to min=%lu",
        t3[2],t2[0]);
    if (t3[1]>t2[1] && !mpi_i) /* max > min */
    printf("\nWarning: stored_ah_blks reduced by mpi from max=%lu to min=%lu",
        t3[1],t2[1]);
    if (!mpi_i) printf("\n optimization of size via BFACTOR:"
      "\n max_nzAH_recv= %lu= %lu*AH vs. NZX=%d>=(2*nu*bonds/site+1)",
      t3[3], (t3[3]+(pt_n*NUM_AHEAD_LINES)-1)/(pt_n*NUM_AHEAD_LINES), NZXMAX);
    hdsk_nzmax=t2[0];
    min_stored_ah_blks=  t2[1];
    min_Mstored_ah_blks= t2[4];
    max_stored_ah_blks=  t3[5]; }
#endif  // Apr09

  if (h_error){  /* error during write is fatal, ignore all things written */
    /* not a good solution, if 90% SH fits to mem, nothing is used = slow */
    if (!mpi_i) E0("disk/mem full? try to use it partly, may fail (ToDo: change!)"); 
    h_error=0;  /* check 2014-08-07 4 threads 72% stored diff 1e-8 !? */
/* ToDo: use num_blks_stored_mem, num_blks_stored_dsk, num_blks_not_stored */
  }
  
  /* max/min/sum hnz_blocks per thread (outpu balance!) */
  /* hnz = nonzero of matrix, snz = stored nonzero of matrix */
  /* sn1 = stored blocks in n1 units (bas cfg), new stored_blks = sn1/AH */
  bhnz_sum=bhnz_max=0; /* unit= HBlocks of size HBLen */
  bhnz_min=((b_hnz[0]+HBLen-1)/HBLen)*mpi_n*pt_n;
  /* ToDo: measure balance, hnz_min*pt_n/h_nz%..hnz_max... */
  for (block=0;block<pt_n;block++) {
    /* ToDo: using HBLen as unit is confusing versus AHEAD
           because HBLen should be (max of all ahead blocks of) AHEAD*nzx */
    /* ToDo17: now no HBLen-blocks, but sum-nz per thread in Units of HBLen
         replace by 1K or 1M  */
    ULong nblks=(b_hnz[block]+HBLen-1)/HBLen;
    if  (nblks!=(b_hnz[block]+HBLen-1)/HBLen)
           F0("ULong overflow, enlarge HBLen");
    /* encode info about blk-src size*(numtasks) + idx_blk_num */
    /* min/max hnz (stored?) per thread for balance ? */
    if (bhnz_max<nblks*mpi_n*pt_n)
        bhnz_max=nblks*mpi_n*pt_n+mpi_i*pt_n+block;
    if (bhnz_min>nblks*mpi_n*pt_n)
        bhnz_min=nblks*mpi_n*pt_n+mpi_i*pt_n+block;
    /* now stored elements only */
    nblks=(b_snz[block]+HBLen-1)/HBLen;  /* stored elements ToDo17: change to blks */
    bhnz_sum+=nblks;
  }
  for (h_nz=0,block=0;block<pt_n;block++) h_nz+=(double)b_hnz[block];
  for (h_ns=0,block=0;block<pt_n;block++) h_ns+=(double)b_snz[block];
#ifdef CONFIG_MPI
  { double r1[2],r2[2]; int i0; ULong t1[9],t2[9]; /* no ULLong in old MPI */
    /* stored + overall hnz + non stored n1 */
    t1[0]=0; t1[1]=bhnz_max; t1[2]=mpi_maxlen;
    MPI_Allreduce(t1, t2,      3, MPI_UNSIGNED_LONG, MPI_MAX, MPI_COMM_WORLD);
    bhnz_max=t2[1]; mpi_maxlen=(int)t1[2]; t1[0]=bhnz_min;
    MPI_Allreduce(t1,&bhnz_min,1, MPI_UNSIGNED_LONG, MPI_MIN, MPI_COMM_WORLD);
    t1[0]=bhnz_sum;
    MPI_Allreduce(t1,&bhnz_sum,1, MPI_UNSIGNED_LONG, MPI_SUM, MPI_COMM_WORLD);
    r1[0]=h_nz; r1[1]=h_ns;
    MPI_Allreduce(r1, r2,      2, MPI_DOUBLE,        MPI_SUM, MPI_COMM_WORLD);
    h_nz=r2[0]; h_ns=r2[1]; t1[0]=mpi_traffic; t1[1]=mpi_syncs;
    /* t1[2]=mpi_z_pkts; */
    for (i0=0;i0<7;i0++) t1[2+i0]= mpi_stats_npkts[i0];
    MPI_Allreduce(t1,t2,       3, MPI_UNSIGNED_LONG, MPI_SUM, MPI_COMM_WORLD);
    mpi_traffic=t2[0]; mpi_syncs=t2[1];
    for (i0=0;i0<7;i0++) mpi_stats_npkts[i0]= t1[i0+2];
    /* mpi_z_pkts=t2[2]; */
  }
#endif  // Jan09
  /* see comments in doc/speed_estim.html */ 
  if (!mpi_i) {
    /* parallel computing balance: (compare to b_hnz = length of htmp*) */
    printf("\nPRF: SH.t_min=  %5.2f blk= %d", rtime_min/60., block_rtime_min);
    printf("\nPRF: SH.t_max=  %5.2f blk= %d", rtime_max/60., block_rtime_max);
    printf("\nPRF: SH.t[min]= %5.2f", tm1/60.);
    /* time in seconds, no div 0, t/nz is per task, speed is agreagated */
    printf( "\nPRF: SH_speed.nz/s= %8.2e  t[ns]/nz= %7.1f (per core) %3.1f*%d",
     h_nz*1./((tm1)?tm1:1)/(mpi_n*pt_n), tm1*1.0e9/h_nz*mpi_n*pt_n,
     tm1*1.0e9/h_nz*mpi_n*pt_n/(1+sym.numall), 1+sym.numall);
    printf( "\nPRF: SH_speed.nz/s= %8.2e  t[ns]/nz= %7.1f (%d cores)",
     h_nz*1./((tm1)?tm1:1), tm1*1.0e9/h_nz, mpi_n*pt_n);
      /* ToDo: per task! ? */
    if (opt_perf_test) printf("\nPRF: speed is overestimated (ToDo)\n");
  } /* mpi_i==0 */
#ifdef CONFIG_MPI
  if (vvv && mpi_n!=1 && !mpi_i) { int i;
    /* t=bytes/BW+pkts*latency, maxlen in data objects per mpi-task */
    printf("\nMPI_traffic  pkts/task=2*n1/ahead*mpi_n= %8g ?= %8g (counted)",
   2.*(((n1+mpi_n-1)/mpi_n+NUM_AHEAD_LINES-1)/NUM_AHEAD_LINES)*mpi_n*mpi_n,
    mpi_syncs*2.);
    /* for lot of zero_packets we should higher AH or remove sendrcvs */
    /* max_pcksize=NZ_BUFSZ1=NZXMAX*NUM_AHEAD_LINES 
                     mpi_maxlen*sizeof(tbase) */
    printf("\nMPI_traffic  bytes/nz= %d  maxpktlen= %d*tbase",
            (int)(sizeof(tbase)+Lsize), mpi_maxlen);
    /* ToDo: max + mean packet size, ahead dependent? */
    for(i=0;i<6;i++)
    printf("\nMPI_pkt ..size/nz=%6dK cnt=%8g",1<<(2*i),mpi_stats_npkts[i]*1.);
    printf("\nMPI_pkt size../nz=%6dK cnt=%8g",1<<10,mpi_stats_npkts[6]*1.);
  }
  mpi_traffic=0;  mpi_syncs=0;  mpi_maxlen=0;
  { int i;for(i=0;i<=6;i++) mpi_stats_npkts[i]=0; } /* reset cnt */
#endif /* CONFIG_MPI */
  if (!mpi_i) {
    ULong memHofs=0l; int i; for (i=0;i<pt_n;i++) memHofs+= memH[i].ofs;
    printf("\n n1 = hilbert dimension,  hnz = number of nonzero matrix elements");
    /* ToDo: better output counted hnz: n1 hnz/n1 ..., 99.9999% as 99%!? */
    printf("\n n1= %.3e      nzx_raw=hnz/n1= %.3f  ", 1.*n1, h_nz/n1);
    // printf("\n n1_stored= %5d*%2d*%.3e",mpi_n, pt_n, 1.*min_b_sn1);
    if (h_nz==0) printf("\nWARNING: h_nz=0, nothing to store, empty matrix"); 
    else printf("\n hnz_stored= %3d%%   hnz_mem= %3d%%   "
                "nz_blks= %d*%d*%.0fKB(mem)",
            (int)((100*h_ns)/h_nz), 
            (int)(memHofs*100.*mpi_n/(Hsize)/bhnz_min+.5),
             mpi_n, pt_n, memH[0].ofs/1024. /* hnz-blks */);
    if (node_ah_blks/pt_n==0)
       printf("\nWARNING: ah_blks=0, nothing to store, empty matrix"); 
    else printf("\n stored_nz_blks= %3.0f%%   mem= %3.0f%%   "
                "nz_blks= %d*%d* %lu..%lu of %lu (mem+dsk)",
                  (min_stored_ah_blks*100.*pt_n/node_ah_blks), 
                 (min_Mstored_ah_blks*100.*pt_n/node_ah_blks),
             mpi_n, pt_n, min_stored_ah_blks, 
                          max_stored_ah_blks, node_ah_blks/pt_n /* hnz-blks */);
    /* balance=max_efficience=max(tm1/(n*tn))=sumhnz/(n*blkhnzmax)
     * ToDo: better hnz instead of blocks (case 1block/thread) !?
     *  hnz_blks=n1*nzx/HBLen (better n1/AH) so fix HBLen=AH*nzx (ToDo17)
     */
    printf("\n hnz_blks/thread= %9.2f"
           " balance= %.0f..%.0f max_eff=%.1f%% min,maxblk=%lu,%lu",
             bhnz_sum*1./(pt_n*mpi_n),
             /* min,max has its blocknum encoded, decode it */
             (bhnz_min/(mpi_n*pt_n))*1.,
             (bhnz_max/(mpi_n*pt_n))*1.,
             (bhnz_sum*100./(pt_n*mpi_n))/(bhnz_max/(pt_n*mpi_n)),
             (bhnz_min%(mpi_n*pt_n)),
             (bhnz_max%(mpi_n*pt_n)) );
    /* 2017-02-17 added block_summed_nonzero max_of_threads */
    if (node_ah_blks%pt_n) // on programming ah_blks assumed pt_n-multiple
      printf("\nERROR: node_ah_blks%%pt_n=%lu", node_ah_blks%pt_n);
    printf("\n hnz_blks/node=%6lu blk=AH=%d (blks*AH>=n1) hnz_blk_sz=%.2e",
           node_ah_blks, NUM_AHEAD_LINES, 1.*Hsize);
  }
  if (mpi_i==0 && mpi_n>1 && 2*(h_nz/node_n1)<NZXMAX)
    printf("\nWarning 2hnz/n1<NZXMAX: waste of memory, see v&2");
/* --- 2017.1 obsolete !? ---
  if (mpi_i==0 && mpi_n>1 && MaxSBase>1)
    printf("\nWarning MaxSBase>1: waste of MPI bandwidth (factor %.1f),"
           "\n please reduce MaxSBase in spins.h to 1",
           sizeof(thbuf)*1./(sizeof(thbuf)
              -(MaxSBase-1)*(sizeof(tbase)+sizeof(tphase))));
*/
  if ((!mpi_i) && h_ns<h_nz) {
    printf(
     "\nWarning: H stored incompletely, need maxfile=%.2e for best speed.",
            1.*Hsize*h_nz/HBLen/(mpi_n*pt_n)+Hsize);
  }
  if (vvv&2) { 
    statH_out(); /* ToDo: speed_test using rnd cfgs, error handling */
    /* ToDo: set errornumber in hbuf !? */
    MPI_Barrier(MPI_COMM_WORLD); /* debug */
  }
  return;
}

/* --------------------- Jan13 2000 -------------------------------
 * new way for storing l1 and H, is it faster than the old way?
 * it should generate the symmetric states itself by
 * - starting with symmetric config
 * - using iterativ H, and symconfig and sort to generate new sym-configs
 * - until H generates no new configs (m1=0 run1 for n1, m1=1 run2 for sh)
 * algo&8:
 * ToDo: make MPI ready! sort in using ising-ergs + hashing!
 *       or mean-(diag|ising)erg of 0th and 1st excitations
 *       to have best blocking and small number hnz in of diagonal blocks
 * use kind of MC? 
 *    1st step generate mpi_n*pt*100 states and part it to mpi_n*pt blocks
 *    generate all states between start_block_i and start_block_i+1
 * mpi_n = 1 !!!
 */
void storeh2(ULong m1, thbuf *hbuf){
  ULong tm1,i1,i2,i3,i4,i5,i6,i0; int j1,j2,block;
  struct sbase b1,b2; tbase ib; double h_nz, h_ns;
  char sbuf2[SZccfg];
  /* 2014-07-17 old local: thbuf *hbuf; */
  /* hbuf=alloc_hbuf();  */ /* very big */
  reset_hbuf(hbuf);  /* 2015-01-31 */
  hbuf->lenAHblk=1; /* bad, but code is not ready */
  if (vvv) printf("\n storeh2 (warning, buggy code, needs review)");
  // b1.o=0; 
  statH_ini();	/* Oct99 */
  /* das speichern des letzten hnz der Zeile macht probleme! da sich n1 aendert! */
  ini_thxy(n1); /* compute bytes needed for storing H elements, ini files */
  tm1=itime();
  i1=0;		/* pointer to last unchecked state */
  i2=1;		/* pointer to next free entry */ 
  if(m1==0){		/* starting state == Neelstate (lowest ising erg) */
    ib=TBASE0;i0=0l;
    for(j1=0;j1<2*nd;j1+=2) ib=puti(ib,j1%nn+j1/nn,2); /* ib=Neel */
    /* MC steps to lower Ising energy if possible, only JJ-model, ToDo */
    /* ToDo: more simple: erg=nHn + nhv to get next n for next better min? */
    {
      int *k1, *k2;   /* pair correlations == ising erg */
      tbase ib2, ib3=ib;      /* test state */
      int z1, z2, j3;
      for (j3=2;j3>0;j3--) { /* down counter */
        k1=get_dcorr(ib ,3, NULL);
        for (j1=0   ;j1<nn-1;j1++)
        for (j2=j1+1;j2<nn  ;j2++)
          if(j1!=j2 && (z1=geti(ib,j1))!=(z2=geti(ib,j2))) {
          /* now j1 and j2 are different neighbours */
          ib2=puti(puti(ib,j2,z1),j1,z2); /* j1 <=> j2 */
          k2=get_dcorr(ib2, 1, NULL); /* get NumIsingExcitations, ToDo: ising, not dcorr */
          /* prefer more movable pairs, ToDo: check all corrs */
          if (k1[1*CM(2,4,5)]< k2[1*CM(2,4,5)] ||  /* smallest j1 */
             (k1[1*CM(2,4,5)]==k2[1*CM(2,4,5)] &&
              k1[2*CM(2,4,5)]< k2[2*CM(2,4,5)])) { 
                 j3=2; ib=ib2; free(k1); k1=k2; j1=j2=nn;
                 minsymcfg_dflt(ib,hbuf); b1=hbuf->scj;
                 if (b1.sign!=0) ib3=ib; /* store sign!=0 */
              }
          else free(k2);
        }
        free(k1);
      }
      ib=ib3;
    }
    printf("\n start config= %s",ccfg2(ib,sbuf2));
    minsymcfg_dflt(ib,hbuf); b1=hbuf->scj; ib=b1.scfg;
    printf("\n         scfg= %s  sign=%+g",
            ccfg2(b1.scfg,sbuf2), D(b1.sign));
    if( b1.sign==0
#ifdef CONFIG_ANSATZ
     || 0==ansatz_test(hbuf->acj[j1])		/* variational ansatz */
#endif
    ) {
    printf("\n bad start konfig (sign==0 || proj==0), try another");
    for(i0=0;i0<10000;i0++){ ib=ll1(i0,nn,nu,nd); 
      minsymcfg_dflt(ib,hbuf); b1=hbuf->scj; ib=b1.scfg;
      if( b1.sign!=0
#ifdef CONFIG_ANSATZ
      && 0!=ansatz_test(hbuf->acj[j1])		/* variational ansatz */
#endif
      )break; }
    }
    putl1(0,ib,l1);	/* first state = starting state, n1=1 ??? */
    if (l2b) l2b[0]=0;	/* sorting-idx */
    printf("\n start config= %s",ccfg2(ib,sbuf2));
  } else i2=m1;   /* start with picked state */
  h_wopen(0);mopen(0,1); /* write only */
  if (mpi_i==0 && mpi_n>1) E0("missing MPI code in storeh2()");
  MPI_Barrier(MPI_COMM_WORLD);
  for (;;) {
    hbuf->num_nzAH=0;
    hbuf->lenAHblk=1; // 2017-02-05
    /* ToDo: i1 must match 0..(b_len[blk]-1) v2.27 */

    /* ToDo: set lenAHblk=1 ? */
    hamilton_nhv( i1, 1, hbuf, 4);    /* store nHm only in buf_ */
    if (err_fatal) { E1(" err_fatal: i1=%lu",i1); break;}
    if (i1<100 && vvv&64) { printf("\n H|i1> = H|%s> i1=%5lu i2=%5lu",
                       ccfg2(getl1(i1),sbuf2),i1,i2);
                       get_dcorr(getl1(i1),2,NULL);
    }
    minsymcfg_dflt(getl1(i1),hbuf); b2=hbuf->scj; /* need for b2.norm2 */
    if(Norm(b2.sign)<1e-6) E0(" should not happen!");

    /* set scfg (compress extracted in v2.29, getbase replaced) */
    for (j1=0;j1<hbuf->num_nzAH;j1++) if (hbuf->rr[j1]!=0) /* test all H|cfg> */
    {					/* every cfg with <cfg|H|i1>!=0 */
      minsymcfg_dflt(hbuf->acj[j1],hbuf); b1=hbuf->scj;/* dont forget is==0!!! ??? */
      /* representator, better use w2w ??? */
      hbuf->rr[j1]*=b1.sign;
      hbuf->acj[j1] =b1.scfg;
      if(b1.norm2!=0 && b2.norm2!=b1.norm2)
         hbuf->rr[j1]*=sqrt((double)b2.norm2/(double)b1.norm2);
    }		/* debugging friendly src-code ??? */

    /* compress */
    for (j1=0;j1<hbuf->num_nzAH;j1++) if (hbuf->rr[j1]!=0) /* test all H|cfg> */
    {					/* every cfg with <cfg|H|i1>!=0 */
      /* search hbuf->acj[j1] in rest of hbuf, compression */
      for(j2=j1+1;j2<hbuf->num_nzAH;j2++)
        if (hbuf->rr[j2]==0
         || hbuf->acj[j2]==hbuf->acj[j1]) {
        hbuf->rr[j1] += hbuf->rr[j2];
        hbuf->num_nzAH--;
        hbuf->rr[j2] = hbuf->rr[hbuf->num_nzAH];
        hbuf->acj[j2] = hbuf->acj[hbuf->num_nzAH]; /* compress */
        j2--;
      }
    }		/* debugging friendly src-code ??? */

    /* remove zeros */
    for (j1=0;j1<hbuf->num_nzAH;j1++) if (hbuf->rr[j1]==0) /* test all H|cfg> */
    {					/* every cfg with <cfg|H|i1>!=0 */
        hbuf->num_nzAH--;
        hbuf->rr[j1] = hbuf->rr[hbuf->num_nzAH];
        hbuf->acj[j1] = hbuf->acj[hbuf->num_nzAH]; /* compress */
        j1--;
    }		/* debugging friendly src-code ??? */

/* parallel ???????, braucht immense Zeit */
    i4=i2;
    for(j1=0;j1<hbuf->num_nzAH;j1++)if(hbuf->rr[j1]!=0){	/* allready stored? */
      if(hbuf->acj[j1]==getl1(i1)){ hbuf->jj[j1]=i1;continue; }
      i3=i2;
      if(l2b)
      {	/* suche im sortierten hilfsindex - schneller?! */
        for(i6=1;i6<=i4;i6<<=1);	/* i4 - hintere bits */
        for(i5=0;i6;i6>>=1){		/* interval-Halbierung */
          if (i5+i6>=i4) continue;      /* i+i/2 > max */
          if (getl1(l2b[i5+i6])<=hbuf->acj[j1]) i5+=i6;
        }
        if(getl1(l2b[i5])==hbuf->acj[j1]){
              hbuf->jj[j1]=l2b[i5];continue; } /* found, min_j? */
        if(i4>=n1){ hbuf->rr[j1]=0;continue; } /* 0-Projection, count? */
        /* insert new element, max_j? -> max(maxj-minj) v-size? */
        for(i6=i4;i6>i5;i6--)l2b[i6]=l2b[i6-1];
        if(i5==0 && hbuf->acj[j1]<=getl1(l2b[i5])) l2b[0]=i4; else
        l2b[i5+1]=i4;
      } else {	/* sequential searching is very slow! */
        for(i3=0;i3<i2;i3++)if(hbuf->acj[j1]==getl1(i3))break;
        if(i3<i2){ hbuf->jj[j1]=i3;continue; }	/* already stored */
      }
      if(i4>=n1){ hbuf->rr[j1]=0;continue; }
      putl1(i4,hbuf->acj[j1],l1); hbuf->jj[j1]=i4;
      i4++;
      if(i4==n1) E1("n1-ovl, n1=%.0f", 1.*n1); /* report once */
    }
    if(i4==1 && hbuf->num_nzAH==0){ ERR(500);break; }
    if(i4>=n1){ E0("n1-ovl");break; }
    hbuf->ofsAHblk=i1;
    /* vvv&2 makes SH factor 2 slower on V490-sparcIV+ cache? v2.25? */
    if (vvv&2) add_statH(hbuf); /* H-statistics for debugging */
    hamilton_puth_line(hbuf);	/* write Hamiltonian!!! run2 n1=correct */
    i2=i4;i1++;if(i1>=i2)break;

  } // storeh2 needs review 2017-02!
  if (hbuf->max_nzAH_overflow) ERR(9999);
  if (b_bnz[0]!=0 && !h_error){   /* flush buffer to disk-file */
    int i;
    for (i=b_bnz[0];i<HBLen;i++){
      b_hbuf[0].nz[i].xy=0;
      b_hbuf[0].nz[i].by=0;
      b_hbuf[0].nz[i].rr=0;
    }
    if (my_fwrite((void *)&b_hbuf[0],Hsize,1,0)!=1) {
      hdsk_nzmax=b_snz[0]-hmem_nzmax; /* ToDo 2014: set to mpi_min */
      b_bn1[0]=0; b_bnz[0]=0; h_error=1;
    } else { /* reset counter buffered_nonzero-entries */
      stored_ah_blks[0]++; b_snz[0] += b_bnz[0]; b_bnz[0]=0;
    }
  } /* if HBLen full */
  if (opt_maxfile!=0) h_wclose(0);
  mclose(0); 
  tm1=itime()-tm1; h_checkpoint(1); n1=i1;
  if (vvv) printf(" SH2 nk=%lu n1=%lu\n",i2, i1);
  if (vvv) printf("PRF: SH2.t=%s\n",sclock((long)tm1));
  if (vvv&2) statH_out(); /* storeh2 ToDo: error handling hbuf->err=2 */
  /* hnz = number of stored elements, snz = number of stored lines */
  for(h_nz=0,block=0;block<pt_n;block++) h_nz+=(double)b_hnz[block];
  for(h_ns=0,block=0;block<pt_n;block++) h_ns+=(double)b_snz[block];
  if (vvv && !mpi_i)
    printf("\n h_nz = num of matrix elements, h_ns = stored elements");
  printf("\n h_nz %.3e h_ns %.3e %3.0f%% hxy %d",
             h_nz,     h_ns,(h_ns*100.)/h_nz,(int)Hsize/HBLen);
  if(h_ns<h_nz)
  printf("\nWarning: %.2e B for maximum speed needed.", 1.*Hsize*h_nz/HBLen);
  free_hbuf(hbuf);
  return;
}

/* ToDo: move to spins_dbg.c # -Os */
int outbasis(FILE *f, thbuf *hbuf){
  ULong i; char sbuf2[SZccfg];
  static struct sbase b1; int i1,i2,i3,i4,i6; // b1.o=0;
  if (vvv&64) { 
    if (!mpi_i) fprintf(f,"\n vvv&64: outbasis.x_max=%lu B=[", x_max);
    if (node_ofs<x_max) usleep(10000*mpi_i); /* sort parallel output */
    if (node_ofs<x_max)
    for (i=0l;i<node_n1 /*&& node_ofs+i<x_max*/;i++){
      if (node_ofs+i==x_max && x_max+1<n1) { // 2018-08-23
        i=node_n1-1; fprintf(f,"\n ... skip output"); }
      if (mpi_i!=0 || i!= 0l)fprintf(f,",");
      fprintf(f,"\n%4lu %s ",(ULong)node_ofs+i,ccfg2(getl1(i),sbuf2));
      minsymcfg_dflt(getl1(i),hbuf); b1=hbuf->scj;
      fprintf(f,"*q(%4g)=(",(double)b1.norm2);
      /* this is ising like energy for all diff. correlations */
      /* ToDo: replace by get_dcorr */
      for(i4=0,i1=1;i1<sym.num_dcorr;i1++){ i6=0;
         for(i2=   0;i2<nn;i2++)
         for(i3=i2+1;i3<nn;i3++)
         if(sym.corr[i2][i3]==i1) {
          if(geti(b1.scfg,i2)==1 && geti(b1.scfg,i3)==1); else
          if(geti(b1.scfg,i2)==2 && geti(b1.scfg,i3)==2); else
          i6++; /* ud - symmetric */
         }
         i4+=i6;
         fprintf(f," %2d",i6); /* fprintf(f,"+%2d+%2d",i4,i6); 2012-05 */
      }
      fprintf(f," sum= %d)  ",i4);
    } /* for i=0...node_n1 */
    MPI_Barrier(MPI_COMM_WORLD);
    if (mpi_i+1==mpi_n) { fprintf(f,"]\n"); fflush(f); }
    /* sleep(1); */ /* sort parallel output */
  } /* if (vvv&64) */
  return(0);
}

double sq( double x){ return x*x; }

/* calculate the geometrical distance^2 between site i1 and i2 */
double distance( int i1, int i2){
  double d1,d2;
  int i,j,k;
  d1=sq(pos[i1][0]-pos[i2][0])
    +sq(pos[i1][1]-pos[i2][1])
    +sq(pos[i1][2]-pos[i2][2]);
  for(i=-1; i<2; i++)
  for(j=-1; j<2; j++)
  for(k=-1; k<2; k++) { /* brute force, test surrounding boxes */
    d2=sq(pos[i1][0]-pos[i2][0]+i*edg[0][0]+j*edg[1][0]+k*edg[2][0])
      +sq(pos[i1][1]-pos[i2][1]+i*edg[0][1]+j*edg[1][1]+k*edg[2][1])
      +sq(pos[i1][2]-pos[i2][2]+i*edg[0][2]+j*edg[1][2]+k*edg[2][2]);
    if (d2<d1) d1=d2;
  }
  return d1;
}

/*
 ***********************************************************************
 * output a table(j1)/matrix(j1,j2) of Ising weigths (picture of state?)
 * using w2w-matrix (output via vvv&8)
 */
#define Imax (2*NN)    /* maximum number uu-pairs per bond type */
extern int w2w[NN][NN];

struct s_ising_weight {
  int blk; /* thread number */
  ULong  states[Imax*Imax];  /* number of states */
  double weight[Imax*Imax];  /* weight matrix */
  int   d1a,d1b, d2a,d2b;     /* matrix dimension (maxIsingErg=max_ud_per_bond) */
  scplx *v0;
} ising_ptr[B_NUM];

void ising_weight_thread( void *ptr ) {
  /* map an site pair nisnjs to a Ising index to count up, see get_dcorr */
  const char map[16]= /* fits to L1 cache line !? */
    {  2, 3, 3, 4 ,  3, 0, 1, 3 ,  3, 1, 0, 3 ,  4, 3, 3, 2 };
   /* .. .u .d .3   u. uu ud u3   d. du dd d3   3. 3u 3d 33 => 0..4 = 3bit */
   /* index: 0=uu+dd  1=ud+du  [2=..+33  3=.u+u.+.d+d.  [4=.3+3.]] */
  int   *buf=NULL;  /* buffer for counted pairs (IsingErg) */
  struct s_ising_weight *pp=(struct s_ising_weight *)ptr;
  int blk=pp->blk;
  ULong i1;       /* state counter */
  tbase b1;       /* cfg[i] */
  int j1, j2;     /* num excitations */
  int i2, i3, i4, ib1, ib2; /* work variables */
  buf=(int *)malloc(CM(2,4,5)*(numpara+2)*sizeof(int));
  if (!buf) { E0("ising_weight: malloc failed"); return; }
  /* upper border of IsingErgs=Neel is nn, add a overflow index nn */
  for (j1=0;j1<Imax*Imax;j1++) {
    pp->states[j1]=0;
    pp->weight[j1]=0;
  }
  pp->d1b=pp->d2b=0;
  pp->d1a=pp->d2a=Imax;

  for(i1=0l; i1<b_ofs[blk+1]-b_ofs[blk]; i1++){  /* l1[] sort for norm2[]??? */
    b1=getl1(i1+b_ofs[blk]);
    // l2.num=0; l2.o=0; minsymcfg_dflt(ib1,hbuf); l2=hbuf->scj; /* get norm2 = slow */
    /* reset IsingErg table */
    for (i4=0;i4<numpara*CM(2,4,5);i4++) { buf[i4]=0; }
    /* get Ising energies per bond */
    for (i2=   0;i2<nn-1;i2++)    /* N*(N-1) site pairs */
    for (i3=i2+1;i3<nn  ;i3++) {
           i4=w2w[i2][i3];    /* access N*N pair array */
           if (i4<0 || i4>=numpara || i4>2+1) continue;
           ib1=geti(b1,i2);  /* 0=. 1=u=o 2=d=x 3=u+d=Q */
           ib2=geti(b1,i3);
           if (CM(2,4,5)*i4+map[4*ib1+ib2]>CM(2,4,5)*(numpara)) ERR(9999);
           buf[CM(2,4,5)*i4+map[4*ib1+ib2]]++;
    }
    j1=buf[CM(2,4, 5)]/2;
    if (numpara<3) j2=0; else
    j2=buf[CM(4,8,10)]/2; /* segfault if numpara<3 ? */
    if (j1>=Imax) j1=Imax-1; /* overflow */
    if (j2>=Imax) j2=Imax-1; /* overflow */
    if (j1>pp->d1b) pp->d1b=j1; /* max matrix idx, minimum output matrix */
    if (j2>pp->d2b) pp->d2b=j2;
    if (j1<pp->d1a) pp->d1a=j1; /* min matrix idx, minimum output matrix */
    if (j2<pp->d2a) pp->d2a=j2;
    pp->states[j2*Imax+j1]++;
    pp->weight[j2*Imax+j1]+=Norm(pp->v0[b_ofs[blk]+i1]); // *l2.norm2;
  }
  free(buf);
}

int out_ising_weight(scplx *v0) {
  long  tm1=itime();
  int j1, j2, blk;
  double sum;
  ULong i1;
  if (!mpi_i)
    printf("\nout_ising_weight: numpara= %d", numpara);
  MPI_Barrier(MPI_COMM_WORLD);
  if (mpi_n>1 && !mpi_i)
    printf("\nWarning: mpi_n>1 can produce wrong results, skip (ToDo)");
  if (mpi_n>1) return 0;

  for(blk=0; blk<pt_n; blk++) {
    ising_ptr[blk].blk=blk;
    ising_ptr[blk].v0=v0;
    parallel_start(blk, ising_weight_thread, &ising_ptr[blk]);
  }
  for(blk=0; blk<pt_n; blk++) {
    parallel_wait(blk);
  }
  if (!mpi_i)
    printf("\nout_ising_weight: parallel work done.");
  MPI_Barrier(MPI_COMM_WORLD);
  /* reduce results from threads */
  for(blk=1; blk<pt_n; blk++) {
    for (j1=0;j1<Imax*Imax;j1++) {
      ising_ptr[0].states[j1] += ising_ptr[blk].states[j1];
      ising_ptr[0].weight[j1] += ising_ptr[blk].weight[j1];
    }
    if (ising_ptr[0].d1b<ising_ptr[blk].d1b)
        ising_ptr[0].d1b=ising_ptr[blk].d1b; /* max matrix idx, min output */
    if (ising_ptr[0].d2b<ising_ptr[blk].d2b)
        ising_ptr[0].d2b=ising_ptr[blk].d2b;
    if (ising_ptr[0].d1a>ising_ptr[blk].d1a)
        ising_ptr[0].d1a=ising_ptr[blk].d1a; /* min matrix idx, min output */
    if (ising_ptr[0].d2a>ising_ptr[blk].d2a)
        ising_ptr[0].d2a=ising_ptr[blk].d2a;
  }
  if (!mpi_i)
    printf("\nout_ising_weight: threads summed.");
  MPI_Barrier(MPI_COMM_WORLD);

#ifdef CONFIG_MPI
  {
    ULong  tstates[Imax*Imax];
    double tweight[Imax*Imax];  /* weight matrix */
    int   td1a,td1b, td2a,td2b;     /* matrix dimension (maxIsingErg=max_ud_per_bond) */
    MPI_Allreduce(&ising_ptr[0].states, &tstates, Imax*Imax, MPI_UNSIGNED_LONG, MPI_SUM, MPI_COMM_WORLD);
    MPI_Allreduce(&ising_ptr[0].weight, &tweight, Imax*Imax, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    for (j1=0;j1<Imax*Imax;j1++) {
      ising_ptr[0].states[j1]=tstates[j1];
      ising_ptr[0].weight[j1]=tweight[j1];
    }
    MPI_Allreduce(&ising_ptr[0].d1a, &td1a, 1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
    MPI_Allreduce(&ising_ptr[0].d2a, &td2a, 1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
    MPI_Allreduce(&ising_ptr[0].d1b, &td1b, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
    MPI_Allreduce(&ising_ptr[0].d2b, &td2b, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
    ising_ptr[0].d1a=td1a;
    ising_ptr[0].d2a=td2a;
    ising_ptr[0].d1b=td1b;
    ising_ptr[0].d2b=td2b;     
  }
  if (!mpi_i)
    printf("\nout_ising_weight: processes summed.");
  MPI_Barrier(MPI_COMM_WORLD);
#endif

  if (!mpi_i) { /* ToDo: mass center, mean deviation, momentums? */
    double center1=0, center2=0; /* mass center of Ising weight */
    double dev1=0, dev2=0,       /* deviation of Ising weight */
           cstate1=0, cstate2=0, /* center of states */
           r1, r2;
    for (j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      center1+=j1*ising_ptr[0].weight[j2*Imax+j1];
      center2+=j2*ising_ptr[0].weight[j2*Imax+j1];
      cstate1+=j1*ising_ptr[0].states[j2*Imax+j1];
      cstate2+=j2*ising_ptr[0].states[j2*Imax+j1];
    }
    for (j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      r1=(center1-j1)*ising_ptr[0].weight[j2*Imax+j1]; dev1+=r1*r1;
      r2=(center2-j2)*ising_ptr[0].weight[j2*Imax+j1]; dev2+=r2*r2;
    }
    printf("\nPRF: iw.t= %.2f min (Ising weight)"
           "\n entropy_center1,2= %8.4f %8.4f "
           "\n mass_center1,2=    %8.4f %8.4f  deviation1,2= %8.4f %8.4f"
           "\n weight matrix (%d-%d,%d-%d):",
            (itime()-tm1)/60., cstate1/n1, cstate2/n1,
             center1, center2, sqrt(dev1), sqrt(dev2),
             ising_ptr[0].d1a, ising_ptr[0].d1b,
             ising_ptr[0].d2a, ising_ptr[0].d2b);

    for (j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      if (j1==ising_ptr[0].d1a) printf("\n ");
      if (ising_ptr[0].states[j2*Imax+j1])
            printf(" %11.9f", ising_ptr[0].weight[j2*Imax+j1]);
      else  printf(" %11g",   ising_ptr[0].weight[j2*Imax+j1]);
    }
    printf("\n#");
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      sum=0;
      for (j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
        sum+= ising_ptr[0].weight[j2*Imax+j1];
      printf(" %11.9f", sum);
    }

    printf("\n states matrix (%d-%d,%d-%d):",
      ising_ptr[0].d1a, ising_ptr[0].d1b,
      ising_ptr[0].d2a, ising_ptr[0].d2b);
    for (j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      if (j1==ising_ptr[0].d1a) printf("\n ");
      printf(" %11lu", ising_ptr[0].states[j2*Imax+j1]);
    }
    printf("\n#");
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      for (i1=0,j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
        i1+= ising_ptr[0].states[j2*Imax+j1];
      printf(" %11lu", i1);
    }
  }

  return 0;
}

/*  ************ check of the eigenvector and eigenvalue ************ */
/* Inp: v0 (ToDo: use static if v0 != v1)
 * Out: v1 (used as work array) 
 *  print H*v0, rh=<v0*H*v0>, rq=norm(H*v0-h*v0) 
 *  print out expectation values as XX=... , use gawk to extract results
 */
int chk3(scplx *v0,scplx *v1,int iv,thbuf *hbuf[B_NUM]){
    int i1,i2,i3,j1,j2,j3,deg; long l1; double d1,d2,rr;
    static double sisj[NN][NN],zizj[NN][NN];
    deg=0; /* degenerate ? */

    for(i1=0;i1<nn;i1++)
    for(i2=0;i2<nn;i2++)zizj[i1][i2]=sisj[i1][i2]=0.0;

#if !defined(Dimer)	/* better: only for generators <T>!!! Nov98 */

/* write a general diff_pair_corr(fkt{-NR},array[][],v0)
 *  and call with sisj,zizj or fkt=1,2,3... */
    l1=itime(); /* measure zizj-time */

#if 1  /* 1: compute spin-correlations and magnetization for EV, 0: skip it */
    /* 3*<ZiZj>=<SiSj> nur, wenn <SS>=0 (singlet), auch bei Entartung (?) */
    /* could be problematicaly od degenerate states! not proofed */
    for(i3=0;i3<sym.num_dcorr;i3++){
      int sites[2]; long l2=itime();
      sites[0]=i1=sym.dcorr[i3][0];
      sites[1]=i2=sym.dcorr[i3][1];
      // old version does non-diag-MPI communication (extremly slower)
      // new variant does no interprozess communication JS-2011-10-24 
      // zizj[i1][i2]= wop(&op_zizj,NULL,2,sites,v0,v0,hbuf); /* ZiZj */
      zizj[i1][i2]= w_diag_op(&op_zizj,NULL,2,sites,v0,v0,hbuf); /* ZiZj */
      for(j3=j1=0;j1<nn;j1++)for(j2=0;j2<nn;j2++)
      if(sym.corr[j1][j2]==i3){
        zizj[j1][j2]=zizj[i1][i2]; /* copy same bonds */
        j3++;              /* count num of same bonds */
      }
      if (!mpi_i) {
        printf("\n i-j %2d %2d ZiZj %9.6f",i1,i2,zizj[i1][i2]); /* kurz+knapp */
        printf("  num= %4d  r^2= %7.3f", j3, distance(i1,i2));
        /* ToDo: no or only once? if fast, easier for diff-check */
        if (n1>>16) printf(" ij.t= %5.1f m", (itime()-l2)/60.); /* for diff-check */
      }
    } /* for(i3=0;i3<sym.num_dcorr;i3++) */
    if (!mpi_i){
      printf("\n %dzij.t= %s", sym.num_dcorr, sclock(itime()-l1));

      for(i3=-1,i1=0;i1<nn;i1++)
      if(sym.corr[i1][i1]>i3){ i3=sym.corr[i1][i1];
        printf("\n i= %2d ZiZj=",i1);	/* 3ZiZj using sh/awk (simpler!) */
        for(i2=0;i2<nn;i2++) printf(" %9.6f",zizj[i1][i2]);
      }
      printf("\n Warning: SiSj=3ZiZj, ZiZj only correct for EA=1, sum over Sz too!");/* Jun99 */
    }

    for(i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++){
      sisj[i1][i2]=3*zizj[i1][i2]; }		/* 3*ZiZj isotropic */
    if (cfg_sisj&1)	/* set in daten.i */
    { 
      int sites[2]; long l2;
      l1=itime();
      if ((vvv&1) && !mpi_i) printf(" t+=%s",sclock(~0l));
      if (!mpi_i) printf("\n cfg_sisj&1 is set, sisj will be calculated ");
      /* could be problematicaly on degenerate states! not proofed */
      for(i3=0;i3<sym.num_dcorr;i3++){
        l2=itime();
        sites[0]=i1=sym.dcorr[i3][0];
        sites[1]=i2=sym.dcorr[i3][1];
        sisj[i1][i2]=wop(&op_sisj,NULL,2,sites,v0,v0,hbuf);  /* SiSj */
        if (!mpi_i)
          printf("\n i-j %2d %2d SiSj %9.6f",i1,i2,sisj[i1][i2]);
        for(j3=j1=0;j1<nn;j1++)for(j2=0;j2<nn;j2++)
        if(sym.corr[j1][j2]==i3){
          sisj[j1][j2]=sisj[i1][i2]; /* copy same bonds */
          j3++;              /* count num of same bonds */
        }
        d1=sisj[i1][i2]-3*zizj[i1][i2];
        if (!mpi_i) {
          printf("  Sij-3Zij= %9.6f %6.0e num= %4d", d1, d1, j3);
          if (n1>>16) printf(" ij.t= %5.1f m", (itime()-l2)/60.);
          /* print time is bad for diff-check */
        }
      } /* for(i3=0;i3<sym.num_dcorr;i3++) */
      if (!mpi_i) {
        printf("\n %dsij.t= %s", sym.num_dcorr, sclock(itime()-l1));
        for(i3=-1,i1=0;i1<nn;i1++)
        if(sym.corr[i1][i1]>i3){ i3=sym.corr[i1][i1];
          printf("\n i= %2d SiSj=",i1);				/* SiSj */
          for(i2=0;i2<nn;i2++) printf(" %9.6f",sisj[i1][i2]); }
      }
# if NOSZSYM  /* Jan18: SixSjx, SiySjy see Kitaev */
      if (!mpi_i) printf("\n cfg_sisj&1 is set, XiXj will be calculated ");
      for(i3=0;i3<sym.num_dcorr;i3++){
        l2=itime();
        sites[0]=i1= sym.dcorr[i3][0];
        sites[1]=i2= sym.dcorr[i3][1];
        d1= wop(&op_xixj,NULL,2,sites,v0,v0,hbuf);  /* SxiSxj */
        d2= wop(&op_yiyj,NULL,2,sites,v0,v0,hbuf);  /* SyiSyj */
        if (!mpi_i)
          printf("\n i-j %2d %2d XiXj %9.6f YiYj %9.6f",i1,i2,d1,d2);
        if (n1>>16) printf(" ij.t= %5.1f m", (itime()-l2)/60.);
        /* print time is bad for diff-check */
      } /* for(i3=0;i3<sym.num_dcorr;i3++) */
# endif /* NOSZSYM */
    }  /* (cfg_sisj&1) set in daten.i */
    if ((!(cfg_sisj&1)) && !mpi_i)
      printf("\n cfg_sisj&1 is not set, sisj will not be calculated ");

    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++) d1+=sisj[i1][i2]; 
    d2=fabs(d1); d2=fabs(-1.0+sqrt(1+4*d2));    /* 2S */
    if (!mpi_i) {
      printf("\n SS= %5.3f 2S= %5.3f % 5.0e", d1, d2, d2-(int)(d2+.5));
      if (cfg_sisj&1) printf(" (as sum sisj)");
      else            printf(" (as sum 3zizj)");
    }

    if ((vvv&1) && !mpi_i) printf(" t+=%s",sclock(~0l));

    for(i3=-1,i1=0;i1<nn;i1++) if(sym.corr[i1][i1]>i3){	/* oct98 */
      if (!mpi_i) printf("\n i= %2d Zi=",i1);		/* Zi */
      i3=sym.corr[i1][i1];
      // d1=wop(&op_zi,NULL,1,&i1,v0,v0,hbuf); /* Zi 2017-01-27-- */
      d1=w_diag_op(&op_zi,NULL,1,&i1,v0,v0,hbuf); /* Zi 2017-01-28++ */
      if (!mpi_i)
        printf(" %9.6f Zi-(nu-nd)/(2N)= %9.6f",d1,d1-(nu-nd)/(2.*nn));
    }
    if ((vvv&1) && !mpi_i) printf(" t+=%s",sclock(~0l));

# if NOSZSYM  /* 2018-01-09 */
    /* ToDo: Xi(real) Yi(cplx)  */
    if (cfg_sisj&1)
    for(i3=-1,i1=0;i1<nn;i1++) if(sym.corr[i1][i1]>i3){	/* oct98 */
      if (!mpi_i) printf("\n i= %2d Xi=",i1);		/* Xi=Sxi Jan18 */
      i3=sym.corr[i1][i1];
      d1=wop(&op_xi,NULL,1,&i1,v0,v0,hbuf); /* Xi nondiag = slow, wud==0 ! */
      if (!mpi_i)
        printf(" %9.6f Xi-(nu-nd)/(2N)= %9.6f",d1,d1-(nu-nd)/(2.*nn));
    }
    if ((cfg_sisj&1) && (vvv&1) && !mpi_i) printf(" t+=%s",sclock(~0l));
#if VecType & 4
    if (cfg_sisj&1)
    for(i3=-1,i1=0;i1<nn;i1++) if(sym.corr[i1][i1]>i3){	/* oct98 */
      if (!mpi_i) printf("\n i= %2d Yi=",i1);		/* Yi=Syi Jan18 */
      i3=sym.corr[i1][i1];
      d1=wop(&op_yi,NULL,1,&i1,v0,v0,hbuf); /* Yi nondiag = slow, wud==0 ! */
      if (!mpi_i)
        printf(" %9.6f Yi-(nu-nd)/(2N)= %9.6f",d1,d1-(nu-nd)/(2.*nn));
    }
    if ((cfg_sisj&1) && (vvv&1) && !mpi_i) printf(" t+=%s",sclock(~0l));
#endif
    if ((!(cfg_sisj&1)) && !mpi_i)
      printf("\n cfg_sisj&1 is not set, xixj will not be calculated ");
# endif /* NOSZSYM */

    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++) d1+=fabs(zizj[i1][i2]); 
    d2= d1= fabs(d1/(nn*nn)); /* ZMag=1/N^2 Sum |ZiZj| */
    if (!mpi_i) printf("\n ZMag= %10.8f",d1);

    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++)
      d1+=fabs(sisj[i1][i2]-zizj[i1][i2]); 
    d1=fabs(d1/(nn*nn));
    if (!mpi_i) printf(  " XYMag= %10.8f",d1);		/* XYMag=1/N^2 Sum |XYiXYj| */

    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++) /* SMag=1/N^2 Sum |SiSj| */
      d1+=fabs(sisj[i1][i2]);
    d1=fabs(d1/(nn*nn));
    if (!mpi_i) printf(  " SMag= %10.8f SMag-3ZMag= % 10.8f",d1,d1-3*d2);

    /* this type of magnetization is for test if SMag(E) in same order? */
    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++) d1+=zizj[i1][i2]*zizj[i1][i2]; 
    d2= d1= sqrt(d1/(nn*nn));
    if (!mpi_i) printf("\n Z2Mag= %10.8f",d1); /* Z2Mag=(1/N^2 Sum ZiZj^2)^.5 */

    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++) d1+=sisj[i1][i2]*sisj[i1][i2]; 
    d1=sqrt(d1/(nn*nn));
    if (!mpi_i) printf(  " S2Mag= %10.8f ",d1);  /* S2Mag=(1/N^2 Sum SiSj^2)^.5 */

#endif /* spin-correlations */

#if 1 /* check again energy and its error: 0=yes 1=no */
    if (!mpi_i)
    printf("\nChange L%d in spins.c to"
           " check the energy again (slow!)\n",__LINE__-8);
#else /* nonparallel and slow version, makes no sence */
    r=wh(v0);	 if (!mpi_i) printf(" E%d/N= %12.8f",iv,D(r)/nn);
    		 if (!mpi_i) printf(  " E%d= %12.8f",iv,D(r));
    r=whq(v0,r); if (!mpi_i) printf(" %5.0e" ,D(r));   /* (H*v-<vHv>*v)^2 */
#endif
    /* if((fabs(r)>1.e-5) && (v1!=NULL)) return deg; */ 

    if(sym.numsym<60 || (double)n1<1e6) /* takes time, change ??? */
    /* wss is slow, if we have ss from sisj we dont need it again */
    if ((!mpi_i) && !(cfg_sisj&2))
      printf("\n cfg_sisj&2 is not set, 2S will not be calculated ");
    if(cfg_sisj&2) {
      if (!mpi_i) printf("\n cfg_sisj&2 is set, 2S will be calculated,");
      l1=itime();
      rr = wss(v0,v0,hbuf);
      if (!mpi_i) printf(" 2S= %5.3f", rr);		/* 2S */
      if (!mpi_i) printf(" 2S.t=%s", sclock(itime()-l1));
    }
#if  CM(0,0,1)
    rr = D( wjj(v0,v0,hbuf) );
    if (!mpi_i) printf(" 2J= %5.3f", rr);		/* 2J */
#endif

    MPI_Barrier(MPI_COMM_WORLD);
#if defined(SITE_DIST)
    /* --- staggered Magnetization --- */
    for(rr=i2=0;i2<nn;i2++)rr+=((site_dist(0,i2)&1)?-1:1)*sisj[0][i2];
    if (!mpi_i) printf(" StaggMag= %9.6f", rr/nn);
#endif
# ifdef Spin1chain
    stringorder2(stdout,v0);
# endif    
#else /* defined(Dimer) */
				/* SiSj */
    stringorder(stdout,v0);
    for(i1=0;i1<nn;i1++)sisj[0][i1]=3*zizj[0][i1];/* sowieso schon Dez99 */
    for(i1=0;i1<nn;i1++)sisj[2][i1]=3*zizj[2][i1];
    /* ##### C_x=Sum_(i!=j,(i,j aus X)){|<SiSj>|} ##### */
    d1=.0;							/* C_tot */
    for(i1=0;i1<nn;i1++)if(i1!=0)  d1+=fabs(sisj[0][i1]);
    for(i1=0;i1<nn;i1++)if(i1!=2)  d1+=fabs(sisj[2][i1]);
    if (!mpi_i) printf(" C_tot= %8.6f",(double)d1/(2*(nn-1)));
    d1=.0;							/* C_d */
    for(i1=0;i1<nn;i1++)if(i1%4<2)if(i1!=0) d1+=fabs(sisj[0][i1]);	
    if (!mpi_i) printf(  "C_d= %8.6f",(double)d1/(nn/2-1));
    d1=.0;							/* C_p */
    for(i1=0;i1<nn;i1++)if(i1%4>1)if(i1!=2) d1+=fabs(sisj[2][i1]);	
    if (!mpi_i) printf(  "C_p= %8.6f",(double)d1/(nn/2-1));
    d1=.0;							/* C_pd */
    for(i1=0;i1<nn;i1++)if(i1%4<2)if(i1!=2) d1+=fabs(sisj[2][i1]);	
    if (!mpi_i) printf(  "C_pd %8.6f",(double)d1/(nn/2));
#endif /* Dimer */
    /* samples for 'exotic' corelation terms: (please double check results) */
#if 0 /* 1: compute site-rotation-terms, 0: disabled (not computed) */
    void op_rot4(void *para,int ns,int *iw,thbuf *hbuf); /* xval.c */
    /* example for H. Mosadegh, Sep09: Heisenberg model only! untested */
    if (nn>4) {
      int iw[4]={0,1,2,3};
      rr = wop(&op_rot4,NULL,4,iw,v0,v0,hbuf);
      if (!mpi_i) printf("\n <s_rot(0,1,2,3)> i= 0123 % 9.6f", rr);
    }
    if (nn>6) {
      int iw[6]={0,1,2,3,4,5};
      rr = wop(&op_rot4,NULL,6,iw,v0,v0,hbuf);
      if (!mpi_i) printf("\n <s_rot(0,1,2-5)> i= 012345 % 9.6f", rr);
    }
#endif
#if 0 /* 1: compute 4-site-terms, 0: disabled (not computed) */
    void op_diff_sisj_sksl(void *para,int ns,int *iw,thbuf *hbuf); /* xval.c */
    void op_mult_sisj_sksl(void *para,int ns,int *iw,thbuf *hbuf); /* xval.c */
    /* example for F. Ouchni, makes sence for tU-chains N>4 */
    if (nn>4) {
      int iw[4]={2,4,0,2}; rr= wop(&op_diff_sisj_sksl,NULL,4,iw,v0,v0,hbuf);
      if (!mpi_i) printf("\n <s(i)s(i+2)-s(i-2)s(i)> i= 2 % 9.6f", rr);
    }
    /* example for J. Richter, makes sence for J-chains N>=4 */
    if (nn> 4) {
      int iw[4]={2,4,0,2}; rr= wop(&op_mult_sisj_sksl,NULL,4,iw,v0,v0,hbuf);
      if (!mpi_i) printf("\n <s(i)s(i+2)*s(i-2)s(i)> i= 2 % 9.6f", rr);
    }
    if (nn>=4) {
      int iw[4]={0,1,2,3}; rr= wop(&op_mult_sisj_sksl,NULL,4,iw,v0,v0,hbuf);
      if (!mpi_i) printf("\n <s(0)s(1)*s(2)s(3)> % 9.6f", rr);
    }
    if (nn>=4) {
      int iw[4]={0,1,0,1}; rr= wop(&op_mult_sisj_sksl,NULL,4,iw,v0,v0,hbuf);
      if (!mpi_i) printf("\n <s(0)s(1)*s(0)s(1)> % 9.6f", rr);
    }
    if (nn>=4) {
      int iw[4]={0,1,1,2}; rr= wop(&op_mult_sisj_sksl,NULL,4,iw,v0,v0,hbuf);
      if (!mpi_i) printf("\n <s(0)s(1)*s(1)s(2)> % 9.6f", rr);
    }
    if (nn>=4) {
      int iw[4]={1,2,0,1}; rr= wop(&op_mult_sisj_sksl,NULL,4,iw,v0,v0,hbuf);
      if (!mpi_i) printf("\n <s(1)s(2)*s(0)s(1)> % 9.6f", rr);
    }
#endif
#if  CM(0,1,1)
    if (!mpi_i) printf("\n niu.=");  /* occupation value <v0|n_s|v1> */
    for(i1=0;i1<nn;i1++) {
      rr= wop(&op_nis,(void *)"1",1,&i1,v0,v0,hbuf);
      if (!mpi_i) printf(" %8.6f", rr); 			/* niu */
    }
    if (!mpi_i) printf("\n niud=");  /* double occupation value (nu*nd)(i) */
    for(i1=0;i1<nn;i1++) {
      rr= wop(&op_nis,(void *)"3",1,&i1,v0,v0,hbuf);
      if (!mpi_i) printf(" %8.6f", rr); /* niud */
    }
    if ((!mpi_i) && !(cfg_sisj&1))
      printf("\n cfg_sisj&1 is not set, slow cicj will not be calculated ");
    for(i3=0;i3<sym.num_dcorr;i3++){
      double uu, dd, ud, du;
      int iw[2];
      iw[0]=i1=sym.dcorr[i3][0];
      iw[1]=i2=sym.dcorr[i3][1];
      rr =  wop(&op_ninj,NULL,2,iw,v0,v0,hbuf); /* ninj */
      if (!mpi_i) printf("\n i-j %2d %2d nn % 9.6f", i1, i2, rr); /* ninj */
      /* for (sym_ud) nuij is wrong, better take (nuij+ndij)/2 */
      uu=wop(&op_nisnjs,(void *)"1111",2,iw,v0,v0,hbuf);
      dd=wop(&op_nisnjs,(void *)"2222",2,iw,v0,v0,hbuf);
      ud=wop(&op_nisnjs,(void *)"1122",2,iw,v0,v0,hbuf);
      du=wop(&op_nisnjs,(void *)"2211",2,iw,v0,v0,hbuf);
      rr=wop(&op_nisnjs,(void *)"3030",2,iw,v0,v0,hbuf);
      if (!mpi_i) printf(  " uu % 9.6f",uu);
      if (!mpi_i) printf(  " dd % 9.6f",dd);
      if (!mpi_i) printf(  " ud % 9.6f",ud);
      if (!mpi_i) printf(  " du % 9.6f",du);
      if (!mpi_i) printf(  " hh % 9.6f",rr);
      /* we need symmetrized operators if we use spin-inversion */
      if (sym.wud) if (!mpi_i) printf(  " uudd % 9.6f", 0.5*(uu+dd));
      if (sym.wud) if (!mpi_i) printf(  " uddu % 9.6f", 0.5*(ud+du));
      /* hopping between site iw[0] and iw[1]    *
       *    sum_s c(+,i,s)c(-,j,s)+h.c.          */
      if (cfg_sisj&1) { /* slow */
        rr = wop(&op_cicj,NULL,2,iw,v0,v0,hbuf);
        if (!mpi_i) printf(  " cc % 9.6f", rr);
      }
    }
    if (sym.wud && !mpi_i)
      printf("\nWarning: use uudd and uddu instead of uu, dd, ud and du");
#if 0 && CM(0,1,1)
/*  Definition by F. Ouchni (do you know about citations?)
 *  "bond-order wave" (BOW) (product of four fermionic operators)
 *  reference:PRB vol_67 page_205109
 *     BOW(r) = (-1)^r ( (\sum_{l} <B_{l,l+1} B_{l+r,l+r+1}>/N) - B^2)
 *          B = \sum_{l} <B_{l,l+1}>/N
 *  B_{l,l+1} = \sum_{spin} (C^{\dag}_{l s} C_{l+1 s} + h.c)
 *  sample: 
 *  correct? only if using symmetry? see m_tU.c
 *   m_tU 10 >daten.def  # v.0.22
 *      param=10,0,1,0,0,1,0.5,0.3333333,0.25,0.1 nud=2,2
 *   E0= 0.4921456 (ok)
 *        sum_k= -2 0 -1000 : <B(l,l+1)> l= 0  -0.642523 (see m_tU.c)
 *               <B(l,l+1) B(l+r,l+r+1)> r= 1   0.299550 OK
 *               BOW(r)                  r= 1   0.113287 OK (F.Ouchni)
 *               
 */
    /* only for chains */
    {
      int iw[4]={0,1,1,2}, r=1;
      double b2=0, b4;
      if (nn>1) printf("\n <B(l,l+1)>              l= 0 % 9.6f",
        b2= wop(&op_cicj,NULL,2,iw,v0,v0,hbuf) );
      if (nn>2)
      for (r=1; r<=nn/2; r++) {
        iw[2]=iw[0]+r;
        iw[3]=iw[1]+r;
        b4= wop(&op_cicjckcl,NULL,4,iw,v0,v0,hbuf);
        if (mpi_i) continue;
        printf("\n <B(l,l+1) B(l+r,l+r+1)> r= %d % 9.6f",r,b4);
        printf("\n BOW(r)                  r= %d % 9.6f",r,(1-((r&1)<<1))*(b4-b2*b2));
      }
    }
#endif

#else /* now only JJ */
    if ((!mpi_i) && !(cfg_sisj&4))
      printf("\n cfg_sisj&4 is not set, mrule will not be calculated ");
    if (  cfg_sisj&4 ) {
      if (!mpi_i) printf("\n cfg_sisj&4 is set, mrule will be calculated ");
      for (i1=0;i1<sym.subl;i1++){
        for(i3=i2=0;i2<nn;i2++) if (sym.sl[i1][i2]+1>i3) i3=sym.sl[i1][i2]+1;
        mrule(v0,sym.sl[i1],i3); /* Marshall+output */
      }
    }
#endif /* JJ */

    MPI_Barrier(MPI_COMM_WORLD);
    /* calculate structure factor S(q)/N^2 v2.27 */
    for (i1=0;i1<sym.subl;i1++){
      int *sl, nsl;
      sl=sym.sl[i1];
      /* get the maximum SL index */
      for(nsl=i2=0;i2<nn;i2++) if (sl[i2]+1>nsl) nsl=sl[i2]+1;
      d1=0;
      if (nsl==2) {
        /* sublattices A,B => S(q=(pi,pi)), A-A: cos=1, A-B: cos=-1 */
        for (i2=0;i2<nn;i2++) for (i3=0;i3<nn;i3++)
          d1+=sisj[i2][i3]*((sl[i2]==sl[i3])?1:-1);
        if (!mpi_i) printf("\n SL%d S(pi)/N^2= % .8f",i1,d1/(nn*nn));
      }
      if (nsl==3) {
        /* sublattices A,B,C => S(q=(4pi/3,0)), A-A: cos=1, A-B: cos=-0.5 */
        for (i2=0;i2<nn;i2++) for (i3=0;i3<nn;i3++)
          d1+=sisj[i2][i3]*((sl[i2]==sl[i3])?1:-0.5);
        if (!mpi_i) printf("\n SL%d S(q3)/N^2= % .8f",i1,d1/(nn*nn));
      }
      if ((!mpi_i) && 2<=nsl && nsl<=3) {
        printf(" SL= ");
        for (i2=0;i2<nn;i2++) printf("%c",sl[i2]+'A');
      }
    }
    MPI_Barrier(MPI_COMM_WORLD); /* sync output */

#if CM(1,0,0)
#if defined(N1) && defined(N2)
#if N1>1 /* ToDo: adapt to symmetries */
    printf("spins.c L%d: old version, please adapt to SBase\n", __LINE__);
   /* local sublattice magnetization Msi^2=<1/N*sum_{j=1}^N{ti*tj*Siz*Sjz}>
    * sublattice magnetization Ms^2=1/N*sum_{i=1}^N{Msi^2}
    * ti=-1 for i odd (A), ti=+1 for i=even (B)      */
    printf(" Z= %8.5f",    D(wz(v0)));
    wtatb(&r1,&r2,v0);
    printf(" TA= %8.5f",   D(r1));
    printf(" TB= %8.5f",   D(r2));
    printf(" Dx= %8.5f",   D(wdx(v0)));
    printf(" Dy= %8.5f",   D(wdy(v0)));
    printf( " C= %8.5f",    D(wc(v0)));
#endif /* N1>1 */
#endif /* defined betts */
#endif

    if (vvv&1) out_ising_weight(v0);
    MPI_Barrier(MPI_COMM_WORLD);

    /* output up to x_max x coefficients and its basecfg */
    if (vvv)             x_out(stdout,v0,0,x_p,hbuf[0]); /* max first */
    if (vvv && x_max<n1) x_out(stdout,v0,1,x_p,hbuf[0]); /* min first */
    MPI_Barrier(MPI_COMM_WORLD);
    /* ToDo: bug! hangs for mpi + ea=2  2017-01-23 only small systems n1<1e6 */
//    if (mpi_n>1) return(deg); /* 2017-01-23 avoids crash */
    if ((vvv&1) && !(n1>>20)) x_sort(stdout,v0,v1); /* slow v0=input, v1=work */
// return(deg); /* ToDo2016 ??? */
    if ((vvv&1) && !mpi_i) printf(" t+=%s",sclock(~0l));
    MPI_Barrier(MPI_COMM_WORLD);
    return(deg);
} /* chk3() */
/* ====================== end of chk3 ======================= */    
    
    
/* sorting indizes using shellsort, R.Sedgewick: "Algorithmen in C" */
/* input: a,n;  output: ia; Nov98 JS +error in 1st Auflage */
/* Mar99-JS: ULong n statt int */
void shellsort(const double a[],int ia[],const ULong n){
  int v; ULong i,j,h;
  if(vvv) printf(" shellsort");
  for(i=0;i<n;i++) ia[i]=(int)i;	/* set indizes, (int) ??? Mar99 */
  for(h=1;h<=n/9;h=3*h+1);		/* using int or ULong ??? confusing */
  for(;h>0;h/=3)
   for(i=h-1;i<n;i++){
     v=ia[i]; j=i;
     while(j>=h && a[ia[j-h]]>a[v]){
       ia[j]=ia[j-h]; j-=h; }
     ia[j]=v;
   }
}

/* ================== USING LAPACK (linear algebra package) ================
  ToDo2016: runtime switch between cplx/real code (like lapack)
   also OpenBLAS: Basic Linear Algebra  (see utils/lapack_test.c n1^2-OVL32)
 DSYEV - eigenvalues and eigenvectors of a real symmetric matrix A (LAPACK)
 ZHEEV - eigenvalues and eigenvectors of a complex Hermitian matrix A
 Makefile: add DFLAGS= -DLAPACK and system specific options LIB=...
 Please test BLAS/LAPACK before using spinpack, see utils/lapack_test.c
   grep zheevd /usr/lib64/lib* # scalapack,lapack,openblas,mkl,acml
 time[s] = 4e-8 * n1^3 (wf: GS160-alpha-731MHz dsyev nev=1 double ) 
 time[s] = 8e-8 * n1^3 (wf: GS160-alpha-731MHz zheev nev=1 complex)
 example: n1=4000 4s-5min, use N(sz=0) slightly above n1_max, reduce n1 by Sz++
   LC40_5+35  k=-2 1 n1=16450
   LC40_4+36  k=-2 1 n1= 2280 nev=1
   LC18_9+9   k=-2 1 n1= 2700 nev=0 oblas0.2.14.zheev   16comp2=0.055m -14.53171376
   LC20_7+13  k=-2 1 n1= 3876 nev=0 oblas0.2.14.zheev   16comp2=0.171m -12.16910877 +9.90211303
     lapack_test -Z      3876       oblas0.2.14.zheev   16comp2=0.172m
   LC20_7+13  k=-2 1 n1= 3876 nev=1 oblas0.2.14.zheevdV 16comp2=0.327m -12.16910877 +9.90211303
     lapack_test -Z -a-1 3876 -d -V oblas0.2.14.zheevdV 16comp2=0.330m 2016-05
     lapack_test -Z      3876 -d -V oblas0.2.14.zheevdV 16comp2=1.047m
   LC20_7+13  k=-2 0 n1= 3876 nev=0 oblas0.2.14.dhyev   16comp2=0.070m -11.72911456 +10.0000000
     lapack_test         3876       oblas0.2.14.dhyev   16comp2=0.072m 2016-05
   LC20_7+13  k=-2 0 n1= 3876 nev=1 oblas0.2.14.dhyevdV 16comp2=0.146m -11.72911456 +10.0000000
     lapack_test -a-1    3876 -d -V oblas0.2.14.dhyevdV 16comp2=0.161m 2016-05
     lapack_test         3876 -d -V oblas0.2.14.dhyevdV 16comp2=0.160m
   LC20_8+12  k=-2 1 n1= 6288
   LC20_9+11  k=-2 1 n1= 8398 nev=1 oblas0.2.14.zheev.16comp2=1.652m
             lapack_test 8398 -Z -d -V 
   LC20_10+10 k=-2 1 n1= 9225 oblas0.2.14.zheev.16comp2
   LC24_12+12 k=-2 1 n1=112632 malloc_xm failed
   4x4-square-cplx sym=q_0_0_0+ n1=365 e0=-11.22848321
   jacobi=21s  lapack=1s  (alpha731MHz, real)   431e-9s*n1^3
   jacobi=     lapack=1s  (alpha731MHz, complex)
   i686-2.7GHz n1=1789 jacobi nev=1 14.1m 53MB 4x4_q00  147e-9s*n1^3 19It,rots=10.7*n^2
   i686-2.7GHz n1=1789 jacobi nev=0 5.38m 28MB 4x4_q00   55e-9s*n1^3 19It,rots=10.7*n^2
   i686-2.7GHz n1=914  jacobi nev=1 1.52m 17MB 4x4_q000 120e-9s*n1^3 17It,rots=9.85*n^2
   i686-2.7GHz n1=914  jacobi nev=0 0.53m 11MB 4x4_q000  40e-9s*n1^3 17It,rots=9.85*n^2
   LC40j2nd4   n1=1240   jaco 640MB-flt  omp1= 98%(sy=2%)  1.2m/5.7e6rots=0.17m/it(7e5rots=.49n**2)
   LC40j2nd4   n1=1240   jaco 640MB-flt  omp2=124%(sy=8%)  5.2m/5.7e6rots=0.7m/it(7e5rots=.49n**2)
   LC40j2nd4   n1=1240   jaco 640MB-flt  omp8=140%(sy=14%)    /5.7e6rots=3.7m/it(7e5rots=.49n**2)
   LC40j2nd4   n1=1240  cjaco 640MB-cflt omp8=150%(sy=14%) 25m/5.7e6rots=3.2m/it(7e5rots=.49n**2)
   LC40j2nd5   n1=8311  cjaco 640MB-cflt omp8=190%(sy=13%)
   LC40j2nd6   n1=48564 cjaco 19GB-cflt it=1:1.82m omp8=328%(sy=10%)
   LC32j2nd7   n1=52819 cjaco 11GB-flt
   LC24j2z0    n1=28968 cjaco 3.3GB-flt 12Zregions 41m++/It nrots=2e6/9m
           skip 0,0-rots: 2e6/1.97m
            pcjago nrots=16k/0.82m(omp1) 16k/1.1m(omp2) 16k/1.4m(omp8)
   LC18j2z0    n1=765   cjaco= 35s nrots=2701576 -15.17070224
                       pcjaco=100s nrots=1748922 -15.17070224 omp2=105s
                       localized?  nrots=1878070 -15.17070224 9m
   ToDo: pthread?
   i686-2.6GHz n1=1617 zheev nev=1 139s  44MB # lapack-3.0 SuSE-9.2
   i686-2.6GHz n1=1617 zheev nev=0  24s  44MB
   i686-2.6GHz n1=1617 dsyev nev=1  44s  23MB # lapack-3.0 SuSE-9.2
   i686-2.6GHz n1=1617 dsyev nev=0   8s  23MB
   i686-2.6GHz n1=3583 dsyev nev=0  85s 103MB
   i686-2.6GHz n1=4704 dsyev nev=0 189s 176MB v2.22
   AOpt-2.2GHz n1=5481 dsyev nev=0 281s 240MB v2.23 4m..5m14s (259M/231M) acml=5m11s
                 10584 dsyev nev=0  ??  880MB v2.23 (vsz reduced)
                 16852 zheev nev=0 314m       v2.24 loki-Opteron-2.2GHz
                 18018 zheev nev=0 13h=48600s v2.22 8e-9s*n1^3 marvel
                 25000 zheevd     1e5s (ca.)
  max_n1=(2^32/cplx?)**(1/2)=2^16=65536
# LC18 J=2,0.125 q=-2,0,0 n1=622 of 48620 76s   0.27m (2011-12-11)
# eigenvals= -15.17070224 -13.22659499 -11.74608234 -11.56012456 -10.96556301
# q=-2,1,x n1=1336 18min jacobi2
# eigenvals= -14.19170172 -12.39204849 -11.79140730 -10.96800408 -10.95840566
# cjacobi: 3.25m (factor 5 faster than [[Re,-Im],[Im,Re]]-trick)
# eigenvals= -14.19170172 -12.39204849 -11.79140730 -10.96800408 -10.95840566
# real: 0.25m (cplx=0.57m ompHT=0.43m)
# eigenvals= -15.17070224 -13.22659499 -11.74608234 -11.56012456 -10.96556301
 !!! Jacobi is better in parallel than QR-algorithm
 * * * WARNING * * *
 SuSE-9.2 + lapack-3.0-920 = OK (n1>5400 wrong Es, spur ok)
  valgrind: Warning: set address range perms: large range 240330888, a 0, v 1
   5481^2*8 = 240330888

 BLAS/LAPACK is a Fortran-Library. This C interface may possibly not work.
 ToDo: extract DSYEV and ZHEEV and convert to C routines
       would allow to use long (8byte) instead of int (4byte) for bufsize
    DSYEV (..,..,A=RealMatrix,LDA,...W=Eigenvalues)
      DSYTRD = tridiag (real sym) = DLATRD+DSYR2K + DSYTD2 (slow=dlarfg+dsymv+daxpy+dsyr2)
        dgemv+dlarfg+dsymv+dscal+daxpy
        dgemv=a*A*x+b*y or a*A^T*x+b*y
      DSTERF = EW or DORGTR+DSTEQR = EW+EV
    ZHEEV (..,..,A=CplxMatrix,LDA,...W=Eigenvalues)
      ZHETRD = tridiag (real sym)
        T = Q**H * A * Q , opt. blocked loops: NB= 32 (ideal), 1 (worst)
        nb-blocked code: ZLATRD + ZHER2K + rest unblocked code ZHETD2
      DSTERF(N,*D,*E,INFO) = EW or ZUNGTR + ZSTEQR = EW+EV
*  if UPLO = 'U':              
*    (  d   e   v2  v3  v4 )   
*    (      d   e   v3  v4 )   
*    (          d   e   v4 )   
*    (              d   e  )   
*    (                  d  )   
*/
/* PMMAAA=dsyev_   see https://en.wikipedia.org/wiki/LAPACK
 * P:      d=double=64bit,s=single=32bit,c=cplx=2x32bit,z=cplxd=2x64bit 
 *  MM:    sy=symmetric, he=cplx_hermitian, ge=unsymNxN
 *    AAA: ev_=simple evx=expert evd=divide-and-conquere(size+N**2) evr=MRRR */
#ifdef LAPACK
#ifdef __cplusplus
extern "C"    /* C++ shall not create special object names */
#endif
void dsyev_( /* reel symmetric eigenvalues + optional(slow) eigenvectors */
        const char *jobz,   /* (in) "N"=NoVectors, "V"=Vectors */
        const char *uplo,   /* (in) "U"=UpperTriangleA, "L"=LowerTA */
        const int *n,       /* (in) order of matrix */
        double *a,          /* a[n][lda] (in/out) */  
        const int *lda,     /* (inp) leading dimension >=max(1,n) */
        double *w,          /* w[n] (out) eigenvalues */
        double *work,       /* work[lwork] (workspace/out) */
        const int *lwork,   /* (in)   >= max(1,3*n-1) better (nb+2)*n */
        int *info);         /* (out) ==0 on success */
#ifdef __cplusplus
extern "C"    /* C++ shall not create special object names */
#endif
void dsyevd_(   /* 2016-04-26 +EV about 4-5 times faster, lapack_test.c */
        const char *jobz,   /* (in) "N"=NoVectors, "V"=Vectors */
        const char *uplo,   /* (in) "U"=UpperTriangleA, "L"=LowerTA */
        const int *n,       /* (in) order of matrix */
        double *a,          /* a[n][lda] (in/out) */
        const int *lda,     /* (inp) leading dimension >=max(1,n) */
        double *w,          /* w[n] (out) eigenvalues */
        double *work,       /* work[lwork] (workspace/out) */
        const int *lwork,   /* (in) dsyev=3*n-1, min=1
                                    dsyevd=1+6n+2n^2 for V, 2n+1 for N */
        int *iwork,         /* array[liwork], [1] returns optimal LIWORK */
        const int *liwork,  /* size = 3+5n */
        int *info);         /* (out) ==0 on success */
#ifdef __cplusplus
extern "C"    /* C++ shall not create special object names */
#endif
void zheev_( /* complex hermitian */
        const char *jobz,   /* (in) "N"=NoVectors, "V"=Vectors */
        const char *uplo,   /* (in) "U"=UpperTriangleA, "L"=LowerTA */
        const int *n,       /* (in) order of matrix */
        double *a,          /* a[n][lda][2] (in/out), diagonal part I=0 */
        const int *lda,     /* (inp) leading dimension >=max(1,n) */
        double *w,          /* w[n] (out) eigenvalues */
        double *work,       /* work[lwork][2] (workspace/out) */
        const int *lwork,   /* (in)   >= max(1,2*n-1) better (nb+2) */
        double *rwork,      /* rwork[max(1,3N-2)] workspace 0..? 1..? */
        int *info);         /* (out) ==0 on success */
#ifdef __cplusplus
extern "C"    /* C++ shall not create special object names */
#endif
void zheevd_(
        const char *jobz,   /* (in) "N"=NoVectors, "V"=Vectors */
        const char *uplo,   /* (in) "U"=UpperTriangleA, "L"=LowerTA */
        const int *n,       /* (in) order of matrix */
        double *a,          /* a[n][lda][2] (in/out), diagonal part I=0 */
        const int *lda,     /* (inp) leading dimension >=max(1,n) */
        double *w,          /* w[n] (out) eigenvalues */
        double *work,       /* work[lwork][2] (workspace/out) */
        const int *lwork,   /* (in)  zheev=2*n-1, zheevd=2n+n^2  */
        double *rwork,      /* rwork[max(1,3N-2)] workspace */
        const int *lrwork,  /* rwork-dim V: 1+5n+2n^2 */
        int *iwork,         /* array[liwork], [1] returns optimal LIWORK */
        const int *liwork,  /* size = 3+5n */
        int *info);         /* (out) ==0 on success */
/* nb = blocksize for ZHETRD returned by ILAENV = 32 */
#endif  /* ifdef LAPACK */

/* ??? use hbuf without ev! a4
 * for complex (hermitian) matrizes of type: A+iB
 *   simply diagonalyze [[ A,-B], [ B, A]], but all eigenvals twice.
 *   EVs: (u,v) -> u+iv
 * old driver for routine jacobi() for symmetric reel matrizes
 * jacobi sparc-Ultra1-143MHz(boson): time/s = 1.5e-6 dim ^ 3.2
 * ToDo: output matrix_asc.pgm for vizualization (or using gnuplot?)
 * --------------------------  Matrizen ----------------------------------- */
/* older real version replaced by complex version 2011-12-11 */
/* declaration for cjacobi from diag.c, ToDo: better sdbl[2] than scplx! */
void  cjacobi(const unsigned long n, scplx *A, double *d, scplx *U);
//void pcjacobi(const unsigned long n, scplx *A, double *d, scplx *U);

// ToDo: input global ULong n1 as argument
// 2016-11-11 add ew as argument replacing global variable ew + return fail
//      rc=0 OK, else fail         
int fulldiag(scplx *ev[], int nev, thbuf *hbuf, double *ew){ 
  // single thread hbuf (2016: means we need to be single thread here?)
  int    rc=0, *ix=NULL, blk, ii; ULong k,i,j,iv=1;
  double *xd=NULL, xr,xi,trace1,trace2,h12,i12,max_abs_H=0.0; 
  double tm1, tm2; /* compute time */
#if !defined(LAPACK)  /* ToDo16: use same double[2] as cplx emulation! */
  scplx  *xm=NULL, *xv=NULL;  /* memory O(n1**2), ToDo: dbl[iv] */
#else
  double *xm=NULL, *xv=NULL;
  int lwork; /* LAPACK */
  int liwork=-1;  /* array[liwork], [1] returns optimal LIWORK */
  int *iwork=NULL;     /* [3+5*n] */
  int lrwork=-1;
  double *rwork=NULL;  /* defined more locally */
  int aaa= (nev) ? 1 : 0; /* fastest algo aaa: 0=ev 1=evd (lapack_test.c) */
  char cev = (nev) ? 'V' : 'N'; /* optionally eigenvectors or not */
#endif
  // thbuf *hbuf;
  minfo_t m_xm = {NULL,0,0,TMPDIR "xm.map",NULL}; /* input matrix */
  minfo_t m_xv = {NULL,0,0,TMPDIR "xv.map",NULL}; /* output/work matrix */
  minfo_t m_xd = {NULL,0,0,TMPDIR "xd.map",NULL}; /* energies */
  minfo_t m_ix = {NULL,0,0,TMPDIR "ix.map",NULL}; /* sort index */
  xi=0;
  if (!mpi_i) printf("\n fulldiag cplx=%d nev=%d n1=%lu", (VecType&4),nev,n1);
  if (mpi_n>1 && !mpi_i) E0("fulldiag dont work with MPI (ToDo: ScaLapack)");
  if (mpi_n>1) return -1; // ToDo: replace by parallel code later
#ifdef LAPACK
  if (!mpi_i) printf(" using BLAS/LAPACK ...\n");
#else
  if (!mpi_i) printf(" using Jacobi-method ...\n");
#endif
  /* standart jacobi and LAPACK expects simple vectors, convert block form */
  /* add 1 for the case n1=1, where Lapack fails */
  trace1=trace2=0; h12=i12=0;
#define S_double sizeof(double)
  /* int overflow ? */
  // xv needed for LAPACK, xm+xv needed for Jacobi
#ifndef LAPACK
  if (!mpi_i) {
    xm=  (scplx *)mymalloc(n1*n1*sizeof(scplx),&m_xm); /* Matrix_H-in =  */
    if (!xm) E1("malloc xm    failed, size= %g",n1*n1*1.*sizeof(scplx));
    if (nev) {
      xv=  (scplx *)mymalloc(n1*n1*sizeof(scplx),&m_xv); /* eigenvects-out */
      if (!xv) E1("malloc xv failed, n1^2*scplx= %g",n1*1.*n1*sizeof(scplx));
    }
  }
#else /* LAPACK */
  iv=((VecType&4)>>2)+1; /* 1=real, 2=emulate_{r,i} by r[2] */
  // for LAPACK we alloc xm(work/out-array) later
  if (!mpi_i) {
    xm=(double *)mymalloc(n1*n1*iv*S_double,&m_xm); /* eigenvects-out */
    if ((!xm)) E1("malloc xm    failed, size= %g",n1*iv*n1*1.*S_double);
  }
#endif
  if (!mpi_i) {
    xd=(double *)mymalloc(      iv*n1*S_double,&m_xd); /* eigenvalues-out */
    ix=   (int *)mymalloc(         n1*S_int,&m_ix);    /* sort-idx-out */
    if ((!xd) || (!ix)) F0("malloc xd+ix failed");
    for(i=0;i<n1;i++) ix[i]=(int)i; /* initialize sort-index (LAPACK) */
  }
  // hbuf=alloc_hbuf(); // all tasks?
  reset_hbuf(hbuf); /* 2015-01-31 lenAHblk=AH */
  /* 2016-11-10 avoid re-use of old values on failed malloc */
  for (j=0;j<NEW;j++) ew[j]=0.0;

  // hbuf->lenAHblk=1; /* ToDo: check! v2.32 */
  if (mpi_i>0) F0("missing MPI code");
  if (xm==NULL || xd==NULL || ix==NULL) rc=-1;  /* return error */
  if (xm!=NULL && xd!=NULL && ix!=NULL) {
    int xflag=0;
    cplx hh;    /* temporary store for array element double/complex */
    double ahh; /* temporary fabs(hh) (slow) */
    if (vvv && !mpi_i) printf(" store matrix H ...");
    /* --- generate a simple matrix, store all elements to xm[] --- */
    tm1=fstime();
    if (xm) for(i=0;i<iv*n1*n1;i++) { xm[i]=0.0; }
    if (xv) for(i=0;i<iv*n1*n1;i++) { xv[i]=0.0; }
    for (blk=0;blk<pt_n;blk++) for(i=0;i<b_ofs[blk+1]-b_ofs[blk];/* i++ */){
      int len= NUM_AHEAD_LINES; /* 2017-02-05 added for better radable code */
      if ((ULong)len > b_ofs[hbuf->block+1] - b_ofs[hbuf->block] -i)
                 len = b_ofs[hbuf->block+1] - b_ofs[hbuf->block] -i;
      hbuf->lenAHblk= len;
      hbuf->num_nzAH=0; /* reset H-buffer, ToDo: lenAHblk=1 ??? */
      hbuf->block=blk;
      /* get hbuf->el.jj+r fulldiag, single thread only */
      hamilton_nhv( i, len, hbuf, 8); // all tasks?
      /* printf("\n n=%d ",hbuf->num_nzAH[0]); */
      for(ii=0;ii<hbuf->num_nzAH;ii++) {
        j =hbuf->jj[ii]; xflag=((j&XY_FLAG)?1:0); j&=~XY_FLAG;
        hh=hbuf->rr[ii];   /* complex matrix element */
        /* if (i+b_ofs[blk]==j) trace1+=Re(hh); */ /* just as a test */
        /* lapack zheev computes max(abs(a[i,j])) for scaling */
        if ((vvv&32)&&!mpi_i)
          printf("\n H blk= %2d xy= %3ld %3ld r= %6g ",
                       blk, i+b_ofs[blk], j, D(hh));
        if (xm) {
#ifdef LAPACK
          /* */      xm[((i+b_ofs[blk])*n1+j)*iv  ] += Re(hh);
          if (iv==2) xm[((i+b_ofs[blk])*n1+j)*iv+1] += Im(hh);
#else /* jacobi */
          /* 2013-04-04 avoid flt+=dbl, use fltA = fltA + dblB */
          xm[              (i+b_ofs[blk])*n1+j] =
           M_d2f( M_f2d(xm[(i+b_ofs[blk])*n1+j]) + hh);
#endif
        } // xm
        i12+= Im(hh)*Im(hh); /* test for a reel matrix (0 if real) */
        /* do lapack.zlanhe needed in lapack.zheev 2013-05-24 */
        ahh=Cabs(hh); /* sqrt(r*r+i*i)==hypot(r,i) */
        if (max_abs_H < ahh) max_abs_H = ahh;
        if (xflag) i++; /* ToDo: v2.32 ??? check correctness  */
      } // for hbuf->num_nzAH
    } // for blk=0..pt_n-1 for ofs[]
    tm2=fstime()-tm1;
    if (vvv && !mpi_i) {
      printf("\nPRF: storeH.t[m]= %.3f",tm2/60.);
      printf("\n max_abs_H= %e", max_abs_H);
      if (max_abs_H<1e-30 || max_abs_H>1e30) printf(" WARNING!");
    }
#if 0
    { int i1,i2;
      for (i1=0;i1<n1;i1++) for (i2=i1+1;i2<n1;i2++)
        if (fabs(xm[i1*n1+i2]-xm[i2*n1+i1])>1e-12)
          printf("\n %3d %3d % e",i1,i2,xm[i1*n1+i2]-xm[i2*n1+i1]); }
#endif
    /* Epsilon       = 5.96046E-08 */
    /* Safe minimum  = 1.17549E-38 */
    /* max_abs_H < sqrt(safeminimum/eps=1e-30) sigma=sqrt()/max_abs_H */
    /* max_abs_H > sqrt(eps/safeminimum=1e+30) sigma=/max_abs_H/sqrt() */
    /* eps   = relative machine precision 1e-12? */
    /* sfmin = safe minimum, such that 1/sfmin does not overflow 1e-10? */
    if (iv==2 && i12==0 && !mpi_i)
      printf("\nWARNING: You are wasting memory"
             " and time using complex mode for real matrix H!");
    /* 1024^2*dbl=8M, 20k^2*dbl=3GB max_size=1024 max_scale=32 */
    if (vvv && !mpi_i)
      printf("\n use vvv&256 to output matrix.pgm (n1<=32*1024)\n");
    if ((vvv&256)&& n1<=32*1024 && !mpi_i) { /* compress n1>=1024 JS2011 */
      FILE *f1; unsigned int bb, ib;
      int    *sorti=NULL; 
      double *diagH=NULL, maxnondiag=0.0;
      sorti=           (int    *)malloc(n1*S_int); /* JS 2011-12-12 */
      if (sorti) diagH=(double *)malloc(n1*S_double);
      if (!sorti || !diagH) E0("malloc sorti,diagH");
#define Hfull(ii,jj) xm[(ii*n1+jj)*iv]
      printf(" use vvv&768 for sorted matrix.pgm\n");
      if (diagH) for (i=0;i<n1;i++) {
        /* ToDo: we need Jzz but use diagH */
        sorti[i]=i; diagH[i]=Re(Hfull(i,i));
        /* 2011-11-15 JS better sorted? using sum off-diag */
        for (j=0;j<n1;j++) diagH[i]+=/* -fabs */(Re(Hfull(i,j))*1e-9); 
      }
      if (diagH && vvv&512) shellsort(diagH, sorti, n1);
      /* 2011-11-15 JS better sorted? but roughly marked */
      if (diagH) for (i=0;i<n1;i++) diagH[i]=Re(Hfull(i,i));
      /* output of relevant Hii ranges by gray bg, 2011-11-15 JS updated */
      if (diagH && vvv&512)
      for (ib=1,i=0;i<n1;i++) {
        if (i>0) {ib=0; if (fabs(diagH[sorti[i  ]]
                                -diagH[sorti[i-1]])>1e-6) ib=1;}
        if (ib) printf("\n %6lu  ii= %6d  Hii= %15.10f",
                        i, sorti[i], diagH[sorti[i]]);
      }  if (!mpi_i) printf("\n");
      f1=fopen("matrix.pgm","w+"); /* PGMBIN 1=black 0=white HL */
      if (diagH && f1) {
        /* 2011-11-14 JS changed from PBMBIN to PGMBIN + scale=1..maxint */
        unsigned int n1sc=n1, scale=1, is, js;
        /* screen=1024 300dpi*A4=2700px? */
        if (n1>1023) { scale=n1/1024+1; n1sc=n1/scale; }
        fprintf(f1,"P5\n# SpinPack matrix H d=%lu\n%u %u 255\n",
             n1, n1sc, n1sc);
        for(i=0;i<n1;i++)
          for(j=i+1;j<n1;j++)
            if (maxnondiag<Cabs(Hfull(i,j)))
                maxnondiag=Cabs(Hfull(i,j));
        printf(" maxnondiag= %.0e", maxnondiag);
        /* round() math.h -lm _XOPEN_SOURCE >= 600 || _ISOC99_SOURCE; ToDo */
        /*     special cases: round(0.5)=1, round(-0.5)=-1 */
        /* dbl rint(dbl): BSD4.3 _BSD_SOURCE || _SVID_SOURCE
             || _XOPEN_SOURCE >= 500 || _ISOC99_SOURCE; or cc -std=c99 */
        /*  ceil(0.5) is 1.0, and ceil(-0.5) is 0.0 (next bigger integral) */
        /* floor(0.5) is 0.0, and floor(-0.5) is -1.0 (lower integral) */
        /* round(x) replaced by ceil(x-0.5), floor(x+0.5)  2013-04-24 */
        /* need nearest, J1=0.01 J2=1 should show J2 matrix */
        for (i=0;i<n1sc;i++) { /* 2016-12 removed usage ib bit 0+1+3 */
          ib=0; /* ib = background color, showing same sorted-diag range */
          if (i>0 && vvv&512)
            if ( fabs(diagH[sorti[ i   *scale]]
                     -diagH[sorti[(i-1)*scale]])>1e-3) ib=4;
          if (n1sc<40) printf("\n%c ",((ib)?'+':'-')); /* 2016-12 add TXT */
          for (j=0;j<n1sc;j++) {
            if (j>0 && vvv&512)
              if ( fabs(diagH[sorti[ j   *scale]]
                       -diagH[sorti[(j-1)*scale]])>1e-3) ib^=4;
            bb=0; /* intensity of pixel propto nonzero-part, 255=white */
            for (is=0;is<scale;is++) /* scale is .ge. 1 */
            for (js=0;js<scale;js++)
              if (Re(Hfull(sorti[i*scale+is],sorti[j*scale+js]))) bb++;
            fprintf(f1,"%c", (255-bb*64/(scale*scale)) ^ ((ib&4)*4) );
            if (n1sc<40) printf(" %c",((bb)?'1'+8*bb/(scale*scale):'.'));
          }
        }
        fclose(f1);
      } /* open(f1) */
      if (sorti) free(sorti); /* JS 2011-12-12 */
      if (diagH) free(diagH);
    } /* vvv&256 == output matrix.pgm */
    if (xm) {
      for(i=  0;i<n1;i++) { trace1+=Re(xm[(i*n1+i)*iv]); } /* matrix trace */
      for(i=  0;i<n1;i++)            /* as a check sum upper left elements */
      for(j=i+1;j<n1;j++) {  h12+=Re(xm[(i*n1+j)*iv]); }
    }
    tm1=fstime();  /* start time fulldiag */
#ifdef LAPACK
    ii=(int)n1;              /* buffer for n1 */
 /* lwork=(nb+1)*n1; */  /* zheev-optimum = (32+1)*n1 working space */
 /* lwork=(nb+2)*n1; */  /* dsyev-optimum = (32+2)*n1, 2016-04-13 */  
    lwork=(int)((32+2)*n1); /* buffer for lwork >= max(1,2*n-1) better (nb+1)*n */
    /* 2013-05-24 nrot=3*n1 is less than recommended 33*n1
       and gives sometimes poor results (see bugs.tex) */
    /* for cplx (nb=32+1)*n1 (n1=489 3*n1=0.024m) */
    /* ToDo: problem if nrot is 4 byte and n1^2 is larger than 2e9 */
    if (iv==2) { /* cplx aaa=1(aldo zheevd) */
      lwork=(32+1)*n1;   /* zheev-optimum = (32+1)*n1 working space */
      lrwork= 3*n1-2;  /* added 2016-04-13 for rwork, unused for DSYEV/ZHEEV */
      if (aaa==1) lwork= (nev) ? 0+2*n1+n1*n1 +30*n1 : 1+n1; /* +30n1 faster */
      if (aaa==1) lrwork=(nev) ? 1+5*n1+2*n1*n1 : n1; /* optimum 2016-04-26 */
    } else {
      lwork=(32+2)*n1;   /* dsyev-optimum = (32+2)*n1, 2016-04-13 */
      if (aaa==1) lwork= (nev) ? 1+6*n1+2*n1*n1 +26*n1: 32+2*n1; /* dsyevd 2016-04 */
    }  /* dsyevd-optimum "V" = 2016-04-25 */
    if (aaa==1 && !mpi_i){
      liwork= (nev) ? 3+5*n1 : 1;  /* zheevd, dsyevd */
      iwork= (int *)malloc(liwork*sizeof(int));
      if (!iwork){printf(" malloc *iwork failed ... fallback ~AAd\n");aaa=0;}
    }
    /* use  ev as matrix  and  xm as workbuffer */
    if (!mpi_i) {
      myfree((void *)xv,&m_xv);
      xv=(double *)mymalloc(lwork*iv*S_double,&m_xv); /* work-space */
      if (!xv) E0("lapack.malloc workspace xv failed");
    }
    if (!mpi_i) {
      int info=-1;
      if (iv==1 && aaa) { int lwork2=-1, liwork2=-1; /* output int array!? */
        dsyevd_(&cev,"U",&ii,xm,&ii,xd,xv,&lwork2,iwork,&liwork2,&info);
//    dsyevd_(&jobz,"U",&n1,a,&lda,w,work,&lwork2,iwork,&liwork2,&info);
        printf(" optimal workspace  work= %9.0f <= %9d\n",D(xv[0]),lwork);
        printf(" optimal workspace iwork= %9d <= %9d\n",iwork[0], liwork);
        printf("\n LAPACK dsyevd"); /* 2016-04 dsyevd is faster + parallel */
        dsyevd_(&cev,"U",&ii,xm,&ii,xd,xv,&lwork,iwork,&liwork,&info);
      }
      if (iv==1 && !aaa) { int lwork2=-1; /* output int via cplx array!? YES! result=(32+2)*n1 */
        dsyev_(&cev,"U",&ii,xm,&ii,xd,xv,&lwork2,&info); /* dsyevd_() */
        printf(" optimal workspace  work= %9.0f <= %9d\n",D(xv[0]),lwork);
        printf("\n LAPACK dsyev"); /* 2016-04 dsyevd is faster for nev=0 */
        dsyev_(&cev,"U",&ii,xm,&ii,xd,xv,&lwork,&info); /* dsyev_() */
      }
      /* --- complex hermitian matrix, iv=2 --- */
      minfo_t m_rw = {NULL,0,0,TMPDIR "rw.map",NULL}; /* work-buffer */
      if (iv==2) {
        rwork=(double *)mymalloc(lrwork*sizeof(double),&m_rw);
        if (!rwork) { E0(" malloc *rwork failed - skip zheevd");aaa=0; }
      }
      /* valgrind shows that zhemv_ of blas-3.0 accesses WORK+len=xm+lwork...
       * so we malloc more elements than lwork=LWORK counts    */
      if (iv==2 && aaa) { int lwork2=-1; /* output int via cplx array!? YES! result=(32+1)*n1 */
        int lrwork2=-1;
        int liwork2=-1; /* detect optimum workspace size */
        zheevd_(&cev,"U",&ii,xm,&ii,xd,xv,&lwork2,
                rwork,&lrwork2,iwork,&liwork2,&info); /* zheevd_() */
        printf(" optimal workspace  work= %9.0f <= %9d\n", D(xv[0]),lwork);
        printf(" optimal workspace rwork= %9.0f <= %9d\n",rwork[0],lrwork);    
        printf(" optimal workspace iwork= %9d <= %9d\n",iwork[0],liwork);      
        printf(" LAPACK zheevd");
        zheevd_(&cev,"U",&ii,xm,&ii,xd,xv,&lwork,
                 rwork,&lrwork,iwork,&liwork,&info); /* zheevd_() */
      }
      if (iv==2 && !aaa) { int lwork2=-1; /* output int via cplx array!? YES! result=(32+1)*n1 */
        zheev_(&cev,"U",&ii,xm,&ii,xd,xv,&lwork2,rwork,&info);
        printf(" optimal workspace  work= %9.0f <= %9d\n", D(xv[0]),lwork);
        printf(" LAPACK zheev");
        zheev_(&cev,"U",&ii,xm,&ii,xd,xv,&lwork,rwork,&info);
      }
      printf("_info=%d (0=ok)\n",info);
      /* free(rwork); rwork=NULL; */
      if (rwork) myfree(rwork,&m_rw); /* crashes using mpt + openblasp 2016-04 */
    } /* !mpi_i */
  /* if not freed, zheev causes segfault after about 7 runs on Tru64-cxx */
    if (iwork) { free(iwork); iwork=NULL; }
#else // not LAPACK (slower)
    if (!mpi_i) {
      printf(" JACOBI_cplx"); /* JS 2011-12-10 speedup ... for cplx */
      cjacobi(n1, xm, xd, ((nev)?xv:NULL));
      shellsort(xd,ix,n1);	/* sort xd[ix[i]], 2016-04 bad iv*n1 */
    }
#endif
    tm2=fstime()-tm1; /* pure lapack time measured without output */
    if (!mpi_i) printf(" eigenvals=");
    for (j=0;j<n1;j++){ trace2+=xd[ix[j]]; printf(" %+12.8f",xd[ix[j]]); }
    printf("\n  traceH= %e sumE= %e traceH-sumE= %e\n sum_h12= %e",
           trace1, trace2, trace2-trace1, h12);
    if (vvv&4) printf("\n number eigenvals eigenvects (sorted):");
    /* convert simple vectors to block form for calling <O>-functions */
    for (j=0;j<NEW && j<n1;j++) ew[j]=xd[ix[j]]; /* 2012-05-05 */
    if (nev)
    for (j=0;j<n1;j++) { /* all eigenvects */
      //  if (vvv&1) printf("\n%3ld %3d: %+12.8f",j,ix[j],xd[ix[j]]); /* eigenvalue */
      if (vvv&1) printf("\n%3ld: %+12.8f",j,xd[ix[j]]); /* eigenvalue */
      if (xv && xm) /* 2011-10-28 in case of low mem, no eigenvects */
      for(k=0;k<b_ofs[pt_n];k++){
#ifdef LAPACK   /* this is hopefully right, normalization? */
        xr=           xm[(k+n1*ix[j])*iv  ]; /* no ix[j] needed, sorted */
        if (iv>1) xi=-xm[(k+n1*ix[j])*iv+1];
        /* the minus sign was found empirically (N=6=4+2, k=-2 1) */
#else /* transposed ev-matrix! */
        xr= Re(xv[k*n1+ix[j]]); /* buggy until 2016-04 from 2011-12-- ? */
        xi= Im(xv[k*n1+ix[j]]);
        xr= Re(xv[k+n1*ix[j]]); /* fixed 2016-04 */
        xi= Im(xv[k+n1*ix[j]]);
#endif
        if (vvv&4) { /* output all coefficients of the eigenstate */
          printf(" ");
          if(iv>1)  printf("(");
                                 outmpl(stdout,xr);
          if(iv>1){ printf(","); outmpl(stdout,xi); printf(")"); }
        }
        /* copy ev in blk-structure */
        if (!A_ptr(ev)) continue;
        ev[0][k] = M_d2f(C(xr,xi)); /* ignore xi if not complex */
      } // k=0...b_ofs[pt_n](n1) ???
      if (!A_ptr(ev)) continue; /* 2016-04-29 added to avoid wop-output */
      /* call <ev|O|ev>-functions here, SiSj-example: */
      if(cfg_sisj&5)
      { /* new block necessary to allow definition of int sites */
        int i1,i3,sites[2];   /* must follow the left bracket */
        /* could be problematicaly on degenerate states! not proofed */
        for(i3=0;i3<sym.num_dcorr;i3++){
          sites[0]=sym.dcorr[i3][0];
          sites[1]=sym.dcorr[i3][1];
          /* Warning: wop want to have *hbuf[B_NUM] but gets *hbuf[1] */
          printf(" Z%dZ%d= %9.6f", sites[0], sites[1],
                  wop(&op_zizj,NULL,2,sites,ev[0],ev[0],&hbuf));
          if(cfg_sisj&1)
          printf(" S%dS%d= %9.6f", sites[0], sites[1],
                  wop(&op_sisj,NULL,2,sites,ev[0],ev[0],&hbuf));
        }
        for(i3=-1,i1=0;i1<nn;i1++) if(sym.corr[i1][i1]>i3){ /* oct09 Zi */
          i3=sym.corr[i1][i1];
          printf(" Z%d= %9.6f",i1,wop(&op_zi,NULL,1,&i1,ev[0],ev[0],&hbuf));
        }
      } // <ev|O|ev>-functions
      /* normvec(ev[0]);        LAPACK not normalized? */
      if(cfg_sisj&2) wss(ev[0],ev[0],&hbuf); /* output S^2 */
    } // j=0..n1-1, all eigenvects
    if (vvv&4 ) printf("\n");
    if (A_ptr(ev))
    for (j=0;j<(ULong)nev;j++){    /* copy eigenvects */
      for(k=0;k<b_ofs[pt_n];k++){ 
#ifdef LAPACK   /* this is hopefully right */
        xr=           xm[(k+n1*ix[j])*iv  ];
        if (iv>1) xi=-xm[(k+n1*ix[j])*iv+1];
#else
        xr= Re(xv[k*n1+ix[j]]); /* bad row/column ToDo 2016-04 */
        xi= Im(xv[k*n1+ix[j]]);
        xr= Re(xv[k+n1*ix[j]]); /* fixed 2016-04 */
        xi= Im(xv[k+n1*ix[j]]);
#endif
        ev[j][k] = M_d2f(C(xr,xi));
      }
    } // for $nev eigenvects
    if (!mpi_i) printf("\nPRF: fulldiag.t[m]= %.3f", tm2/60.);
  } // if(xv!=NULL && xd!=NULL && ix!=NULL) malloc=OK
  if (hbuf->max_nzAH_overflow) ERR(9999);
  /* 2016-05 move PRF to the end, no more scrolling for PRF-tests */
  if (!mpi_i) printf("\n"); /* 2016-04 free-msg in next line! */
  if (ix) myfree((void *)ix,&m_ix);
  if (xm) myfree((void *)xm,&m_xm);
  if (xv) myfree((void *)xv,&m_xv);
  if (xd) myfree((void *)xd,&m_xd);
//    free_hbuf(hbuf);
  return rc;
}

/* ================== read/write parameters  ====================== */

char *strsub(char *sd,const char *ss,int i1,int i2){/* copy ss[i1..i1+i3-1] */
 int i3;for(i3=0;i3<i2;i3++)sd[i3]=ss[i1+i3];sd[i2]=0;return(sd);}

char *fk(const char *ss,int i2){ /* get i2-tes element from komma-list */
 static char s1[512];int i1,i3,i4,i5;char s2;
 s1[0]=0;
 i1=0;  /* pos letztes komma +1 */
 i3=0;  /* klammerebene, why? used? cplx input? */
 i4=0;  /* komma-nr ToDo18: allow space separated lists too */
 for(i5=0;i5<(int)strlen(ss)+1;i5++){
  s2=ss[i5];if(s2==0)s2=',';
  if(s2=='(') i3++;else
  if(s2==')') i3--;else
  if((s2==',') && (i3==0)){ 
    i4++; if(i4<i2) i1=i5+1; else break;  } }
 if((i4>=i2) && (i2>0))strsub(s1,ss,i1,i5-i1);
 if((s1[0]=='(')&&(s1[strlen(s1)-1]==')'))strsub(s1,s1,1,(int)strlen(s1)-2);
 return s1; }


cplx atoc(char *s){	/* (r,w) => r*exp(I*w*Pi) */
  double x=0.0, y=0.0; int i=0;
  while(s[i]!=0 && s[i]!=',')i++;
  x=atof(s); if(s[i]==',') y=atof(&s[i+1]);
#if VecType & 4     
  return polar(x,y*M_PI);}
#else
  if(y!=0)E0("imag(r)!=0"); return C(x,0);}
#endif     

void outpara(FILE *in){
  int i;
  fprintf(in," param=");
  for(i=1;i<numpara;i++){
    fprintf(in," %.9g",(double)para[i]);
    if(i<numpara-1)fprintf(in,",");
  } 
  fprintf(in,"\n# p1= %g\n# p2= %g  ... can also be used\n",
          (double)para[1], (double)para[2]);
  fflush(in);
  return;
}

/* set complex phase of bond (ToDo: TBC only? not tested since 2000?) */
/* ToDo: fix bad global var usage */
int set_ww(){
  int i;
  for(i=0;i<nw;i++){
    ww.co[i]=cos(M_PI*para[ww.w[i]]*ww.ex[i]);
    ww.si[i]=sin(M_PI*para[ww.w[i]]*ww.ex[i]);
  }
  return 0;
}

/* check_ud separated from scan_ud, if no ud option is in i-file */
int check_ud(){ /* ToDo18: add arguments nn,nu,nd */
#if CM(1,0,0)  /* 2015-03 add output pre state, 2016-03 fix nu>nn */
  if(nu+nd!=nn || nu<0 || nd<0){
    if (!mpi_i) printf("\nwarning: nu+nd!=N     nu=%d nd=%d N=%d",nu,nd,nn);
    if (nu<-nn) nu=1; /* 2016-03 fix -40+140==100 */
    if (nd<-nn) nd=1;
    if (nu>nn) nu=nn-1;
    if (nd>nn) nd=nn-1;
    if (nu<0 && nd>=0) { nu=nn+nu; nd=nn-nu; } /* want have high Sz ? */
    if (nd<0 && nu>=0) { nd=nn+nd; nu=nn-nd; } /* want have high Sz ? */
    if (nu+nd!=nn && abs(nu-nd)<=nn) {nu=(nn-abs(nu-nd))/2;nd=nn-nu;} /* low Sz */
    if (nu+nd!=nn && nu<nn) nd=nn-nu; /* easy to solve */
    if (nu+nd!=nn && nd<nn) nu=nn-nd; /* easy to solve */
    if (nu+nd!=nn) { nu=nn/2; nd=nn-nu; } /* Sz=0 or minimum else */
    if (!mpi_i) printf("\nwarning: nu+nd!=N set nu=%d nd=%d",nu,nd); }
#endif  
#if CM(0,1,0)
  if(nu+nd>nn || nu<0 || nd<0){
    if (!mpi_i) printf("\nwarning: nu+nd> N nu=%d nd=%d N=%d",nu,nd,nn);
    if (nu<-nn) nu=1; /* 2016-03 fix -40+140==100 */
    if (nd<-nn) nd=1;
    if (nu<0) { nu=nn+nu; } /* want have high nu ? */
    if (nd<0) { nd=nn+nd; } /* want have high nd ? */
    while (nd+nu>nn && nu>=0 && nd>=0) { nu--; nd--; }; /* easy to solve */
    if (nd+nu>nn) { if (nd>nu) nd=nn-nu; else nu=nn-nd; }
    if (!mpi_i) printf("\nwarning: nu+nd> N set nu=%d nd=%d",nu,nd); }
#endif  
#if CM(0,0,1)
  if(nu>nn || nd>nn || nu<0 || nd<0){
    if (!mpi_i) printf("\nwarning: nu,nd> N nu=%d nd=%d N=%d",nu,nd,nn);
    if (nu<-nn) nu=1; /* 2016-03 fix -40+140==100 */
    if (nd<-nn) nd=1;
    if (nu<0) { nu=nn-nu; } /* want have high nu ? */
    if (nd<0) { nd=nn-nd; } /* want have high nd ? */
    if (nu>nn) nu=nn; /* easy to solve */
    if (nd>nn) nd=nn;
    if (!mpi_i) printf("\nwarning: nu,nd> N set nu=%d nd=%d",nu,nd); }
#endif
  return 0;
}

int scan_ud(char *s1,int sum){  /* ToDo18: replace by scan_inta() */
  if (sum) {  /* add values, used for loops */
    nu+=atoi(fk(s1,1));  /* fk() separate klammer/komma-list */
    nd+=atoi(fk(s1,2));
  } else {
    nu=atoi(fk(s1,1));
    nd=atoi(fk(s1,2));
  }
  // check_ud(); /* ToDo18 use arguments nn,nu,nd, not global vars */
  return 0;
}

/* scan komma separated float parameter list from string s1
 *    and store in parameter array para[]
 *   sum: 0=set("param="), 1=add("param+=")
 *   1..numpara-1
 *   ToDo18: ~global_var_numpara, check num kommas >= np-1 + warn else
 */
int scan_para(char *s1, int sum){  /* ToDo18: replace by scan_dbla() */
  int i;    /* fk() separate klammer/komma-list cplx input also? */
  if (sum)  for(i=1;i<numpara;i++){para[i]+=atof(fk(s1,i));}
  else      for(i=1;i<numpara;i++){para[i]= atof(fk(s1,i));}
  set_ww(); /* set complex phase bonds, TBC only? */
  return 0;
}

/* =========================== useful functions (handling strings) ====== */

void testsym(){
}

/*
 * search end of line
 */
char *scan_eol(char *s){
  char *s2;
  s2=strchr(s,'#');if(s2) return s2; /* comment-line */
  s2=strchr(s,'%');if(s2) return s2; /* comment-line */
  s2=s+strlen(s);
  return s2;
}
/*
 * search "vnam=" in s, get position of next char, ret 0=not found
 *  ToDo: search expression "p[0-9]*=" or p%d=%f ?
 */
char *scan_var(char *s,const char *vnam){
  char *s1,*s2;
  if ((s1=strstr(s,vnam))==NULL) return NULL;
  s1+=strlen(vnam);
  s2=scan_eol(s); if(s2) if(s2<s1) return NULL; /* comment */
  if (s1[0]!='='){ if (vvv) printf(" %s=???",vnam); return NULL; }
  s1++; while(s1[0]==' ' || s1[0]=='\t') s1++;
  return s1;
}

/* ToDo18: replace by scan_inta() */
int scan_int(char *s,const char *vnam,int *var){
  char *s1;
  if( (s1=scan_var(s,vnam))==NULL ) return -1;
  if(var){ *var=atoi(s1); if(vvv && !mpi_i) printf(" %s=%d",vnam,*var);}
  return 0;
}

int scan_inta(char *s,const char *vnam,int *var,int len){
  char *s1,*s2; int i=0;
  if( (s1=scan_var(s,vnam))==NULL ) return -1;
  s2=scan_eol(s); len=(int)(s2-s1);
  if(var){
   if (vvv && !mpi_i) printf(" %s=",vnam);
   for(i=0;i<len;i++){
    var[i]=(int)strtol(s1,&s2,10);
    if(s1==s2 || s2[0]==0) break;
    if (vvv && !mpi_i) printf(" %d",var[i]);
    s1=s2;
   }
   if (vvv && !mpi_i) printf("\n");
  }
  return i;
}

/* search "vnam=" in s, set value to var, ret 0=found,1=not found,2=err
   last modify 1998-139, ToDo18: missing scan_dbla() replaces scan_param */
int scan_dbl(char *s,const char *vnam,double *var){
  char *s1;
  if( (s1=scan_var(s,vnam))==NULL ) return -1;
  if(var){ *var=strtod(s1,NULL);
    if (vvv && !mpi_i) printf(" %s=%g",vnam,*var);
  }
  return 0;}

int p_mask_ini(int nn){
  int i,j;
  /* --- initialize bitmasks for faster parity() --- */
  for(i=0;i<nn;i++)for(j=0;j<nn;j++)
    if(i<j) p_mask[i][j]=(~(~(tbase1)0 << j << j)) & (~(tbase1)0 << i << i << 2);
    else    p_mask[i][j]=(~(~(tbase1)0 << i << i)) & (~(tbase1)0 << j << j << 2);
  return 0;
}

void ns_thread( void *ptr); /* see hilbert.c */
void wopij_block( void *ptr ); /* see xval.c */
void wop_block( void *ptr ); /* see xval.c */
void diag2_thread( void *ptr ); /* see diag.c */

/* ToDo: write a test_task which only outputs "(mpi.thread): I am here" */
void test_pthread( void *ptr ){
  main_pts *pp=(main_pts *)ptr;
  if (mpi_i*pt_n+pp->num<4 || mpi_i*pt_n+pp->num>mpi_n*pt_n-1-4)
    printf(" node.thread %d.%d is running as block %d\n",
           mpi_i, pp->num, mpi_i*pt_n+pp->num);
  else
  if (mpi_i*pt_n+pp->num==4)
    printf(" skip output of threads %d.%d..%d.%d\n",
      mpi_i, pp->num, (mpi_n*pt_n-1-4)/pt_n, (mpi_n*pt_n-1-4)%pt_n);
}

void pt_cpuset(int blk) { /* 2017-01 called within thread or unthreaded task */
#if defined(CFG_CPUSET)
  /* ToDo: use CPUSET scheme of memspeed or mpi_stress !! 2014-06 */
  // this processor bounding is useful for the case, that the jobsystem
  //   does not do this automatically
  // ToDo16: --pin_ncpu=16 --pin_ofs=0 --pin_dist=1 --pin_size=1 ???
  //  ToDo: warning if numcores overruns CPUSET size 2013-05
  // ToDo 2016: comp2 getaffinity dflt= 0x0000ffff + use this mask and
  //             take one of each bit for tasks,
  //             allows dynamic taskset multibits from outside
  // 2016-12-22 replace use of hard coded mask CFG_CPUSET by getaffinity()         
  //            example use cores 4-8: taskset 0x0FF0 ./spin -t8
  // ToDo16: use KMP_AFFINITY=[<modifier>,...]<type>[,<permute>][,<offset>]
  // Problem18: mpi + openblas(NO_AFFINITY=off) some tasks pinned to core 0
  //         causing 2++ tasks pinned to core0 and other core unused = slow
  //    collect(gather) all SETs to task0 (not to all nodes because mem=n^2)
  //    also sorted output, but hostname? 8chars as double? group per node?
  // ToDo18: MPI_Gather(send_from_node,rcv_from_nodes_root_only,...)
  //          Reduce XOR + AND=0(per node) + OR() but MPI thread 0 only!
  //         collect affinity threads first + pt_barrier
  cpu_set_t cpuset;
  int ii;
  /* 2016-12-22 added more info, usually 1024, bigger than numcores */
  /* DBG_AF: cpuset[bits]=4096  CPU_SETSIZE=4096  supermuc11.fat=40c*2ht */
  /*  4096 bits means 512 Byte/task */
  /*  1024 bits = 128B/task * 5800tasks = 725kB + 0x003f on kautz */
  if (mpi_i==0 && blk==0) // 2018.01 add CFG_CPUSET 0=no 1=get 2=set
    printf("DBG_AF: cpuset[bits]=%u  CPU_SETSIZE=%u CFG_CPUSET=%d\n",
      (unsigned)(8*sizeof(cpu_set_t)), CPU_SETSIZE, CFG_CPUSET);
  /* 2016-12 comp2: 0x0000ffff on 16-core machine without any affinity set */
  /* ToDo2017 what about pthread_getaffinity_np ??? */
  /* int sched_setaffinity(pid_t pid, unsigned int cpusetsize,
   *                       cpu_set_t *mask);  + <sched.h>
   *   pid=0 is local thread, or use pid_t gettid(void); 
   * int pthread_getaffinity_np(pthread_t thread, size_t cpusetsize,
   *                       cpu_set_t *cpuset); + pthread.h + _GNU_SOURCE
   *   to be set on motherthread??? (not neccess.)
   *   but ntpl/pthread.h must be used? sched is better!? use pthread_self()?
   */
  /* log(mpi) delay? for sorted output? */
  usleep(mpi_i*20000/mpi_n + blk*4000/pt_n);
#if CFG_CPUSET > 1   /* 2017-01-30 added 1=show_only 2=set+show */
  if (sched_getaffinity(0, sizeof(cpu_set_t), &cpuset)) {
    if (mpi_i<8) printf("ERR.getaffinity %5d.%2d failed\n", mpi_i, blk); }
  /* output 4 first and 2 last (to see bad effects like t1 set by t0) */
  if (mpi_i*pt_n+blk<3 || mpi_i*pt_n+blk >= mpi_n*pt_n-1) {
    ULong cpuset2=0L; int nbits=0, lbit=-1, hbit=-1;
    for(ii=0;ii<128 && ii<CPU_SETSIZE;ii++) // supermuc.fat=80vcores
      if(CPU_ISSET(ii,&cpuset)) {
        hbit=ii; if (lbit<0) lbit=ii;
        if (ii<8*(int)sizeof(cpuset2)) cpuset2|=1UL<<ii;
        nbits++; }/* 64bit-mask */
    printf("DBG.getaffinity %5d.%2d cpuset=%016lx nbits=%d %2d...%d\n",
         mpi_i, blk, cpuset2, nbits, lbit, hbit); /* bad for 32/64++ cores */
  }
  /* reduce mask to 1 bit */
  { int  ncores=0, in=0;
    for (ii=0; ii<CPU_SETSIZE && ncores<B_NUM; ii++)
                                     if (CPU_ISSET(ii,&cpuset)) ncores++;
    for (ii=0; ii<CPU_SETSIZE; ii++) if (CPU_ISSET(ii,&cpuset)) {
        if ((mpi_i*pt_n+blk)%ncores==in) break;  // mycore = 0,1,2,...
        in++;
    }
  }
  CPU_ZERO(&cpuset); CPU_SET(ii,&cpuset);
  /* output 4 first and 4 last */
  if (mpi_i*pt_n+blk < 3 || mpi_i*pt_n+blk >= mpi_n*pt_n-1) {
    printf("DBG.setaffinity %5d.%2d to core %2d,  node=%s\n",
           mpi_i, blk, ii, nodename);
  }
  // no sched_setaffinity on OSF1
  if (sched_setaffinity(0, sizeof(cpu_set_t), &cpuset)) {
    /* failes if ii is above real core numbers (2016-12 comp2 size=1024) */
    printf("ERR.setaffinity %5d.%2d %2d failed\n",
            mpi_i, blk, ii); usleep(10000); perror("setaffinity");
  }
  usleep(40000);
#endif /* CFG_CPUSET > 1 */
  /* 2016-12-22 added getaffinity */
  if (sched_getaffinity(0, sizeof(cpu_set_t), &cpuset)) {
    if (mpi_i<8) printf("ERR.getaffinity %5d.%2d failed\n", mpi_i, blk); }
  /* output max. 5...6...30 */
  if (mpi_i*pt_n+blk < 4 || mpi_i*pt_n+blk >= mpi_n*pt_n-2
   || mpi_i*pt_n+blk == mpi_n*pt_n/2){
    ULong cpuset2=0L; int nbits=0, lbit=-1, hbit=-1;
    for(ii=0;ii<128 && ii<CPU_SETSIZE;ii++) // supermuc.fat=80vcores
      if(CPU_ISSET(ii,&cpuset)) {
        hbit=ii; if (lbit<0) lbit=ii;
        if (ii<8*(int)sizeof(cpuset2)) cpuset2|=1UL<<ii;
        nbits++; }/* 64bit-mask */
    printf("DBG.getaffinity %5d.%2d cpuset=%016lx nbits=%d %2d...%2d\n",
         mpi_i, blk, cpuset2, nbits, lbit, hbit); /* bad for 32/64++ cores */
  }
  fflush(stdout); /* 2017-02-25 do not cache output of other threads */
  usleep(40000); /* why? wait for other outputs, ToDo: like ordered by OMP */
  // if (pt_n>1) pthread_barrier_wait(&pt_barrier); // 2017-09-12 but hangs with t1 @ kautz
//  if (!blk) MPI_Barrier(MPI_COMM_WORLD); // 2017-09-12 but fails on thread
#endif /* CFG_CPUSET defined */
}

/* void-function for pthreads, waiting for work, send via main_pts[] */
void main_pthread( void *ptr ){
#ifdef CONFIG_PTHREAD
  /* main_pts = struct {} */
  /* why volatile was outcommended until 2016-12, think its needed */
  /* 2016-12-22 possible icpc may create endless-loop without volatile? */
  volatile main_pts *pp=( volatile main_pts *)ptr;
  pt_cpuset(pp->num); /* 2017-01 show + set affinity of thread/task */
  /* test_pthread(ptr); */
#ifdef _OPENMP
  if (pp->num<8 && !mpi_i)
    printf("DBG_PT: enter worker loop %d\n", mpi_i*pt_n+pp->num); 
  /* endless wait for new work from master thread */
  for(;;) {
    while (pp->task==0)  /* volatile! */
      usleep(1000);  /* 10ms testwise */
    if (pp->task<0) break;
    pp->worker( (void *)pp->data ); /* worker thread */
#pragma omp critical
{
    pp->task=0;
}
#pragma omp flush (pp)  // ??? ToDo pointer
  }  /* for(;;) */
#else /* ~_OPENMP */
  for(;;) {  /* endless loop for worker (child-) threads */
    pthread_mutex_lock(&pt_mutex1);
    // printf("DBG: thread %d starts waiting\n",pp->num);
    while (pp->task==0)  /* volatile! */
      pthread_cond_wait(&pt_cond1,&pt_mutex1); /* unlock+wait_cond+lock */
      /* ToDo: better have one(two) condition variable(s) per thread */
    pthread_mutex_unlock(&pt_mutex1);

    if (pp->task<0) break;
    // printf("DBG: thread %d starts work\n",pp->num);
    pp->worker( (void *)pp->data ); /* call worker function for this thread */

    pthread_mutex_lock(&pt_mutex1); /* ToDo: check for race condition? */
    pp->task=0; /* ready */
    pthread_cond_signal(&pt_cond2); /* wake up master */
    pthread_mutex_unlock(&pt_mutex1);
    /* ToDo16: should we give the master a chance to lock here? race-locks? */
    /*         no. pthread_cond_wait does unlocking, which is enough */
  }
  pthread_exit(0);
#endif /* _OPENMP */
#endif /* CONFIG_PTHREAD */
} /* main_pthread() */

/* return VM in [4kB-]pages, output debug mem info, 2018-11 */
/*  also: /proc/$$/status VmPeak VmSize,
 *    no fork, therefore less mem than system()
 *    just simple IO, no output if bad or no /proc
 */
ULong getmem(){ /* 2^32=4e9*4KB=16TB */
 FILE *f1; int rc=0; ULong vm=0,rss=0,shr=0; char s1[64];
 f1=fopen("/proc/self/statm","r"); /* linux only, BSD? */
 if (f1) { if (fscanf(f1,"%lu %lu %lu %*s",&vm,&rss,&shr)<3) rc=2; }
   else rc=1;
 if (f1) fclose(f1);
 printf("\nMEM.statm: rc=%d vm,rss,shr[4KB]= %lu %lu %lu ",rc,vm,rss,shr);

 f1=fopen("/proc/self/status","r"); /* linux only, BSD? */
 if (f1) while(!feof(f1)) { s1[0]=0;
     if (!fgets(s1,63,f1)) break;
   for(rc=0;rc<64 && s1[rc];rc++) if(s1[rc]=='\n') s1[rc]=0;
   if (strstr(s1,"VmPeak")
    || strstr(s1,"VmSize")) printf("\nMEM.status: %s", s1);
 } if (f1) fclose(f1);

 f1=fopen("/proc/meminfo","r"); /* linux only, BSD? */
 if (f1) while(!feof(f1)) { s1[0]=0;
     if (!fgets(s1,63,f1)) break;
   for(rc=0;rc<64 && s1[rc];rc++) if(s1[rc]=='\n') s1[rc]=0;
   if (strstr(s1,"MemAvailable")
    || strstr(s1,"MemFree")) printf("\nMEM.status: %s", s1);
 } if (f1) fclose(f1);
 return (vm);
}

extern int sym_bn_m_ini_flag;  /* see hilbert.c */
void sym_bn_m_ini(int pt);
extern int chkpt_time; /* diag.c checkpoint interval in seconds (4d) 2012 */
/* ---------------------------- MAIN --------------------------------- */
int    main(int argc,char *argv[]){
  const char *name_file_i="daten.i", *name_file_m="daten.def";
  static char s1[500], *ss;
  ULong n1max=1000;
  int i, ii=0, i1, j, itr=0, iv, nev, ne0,
    nruns=0, /* count Lanczos rounds, use for srand of randvec() */
    block;
  FILE *in1=NULL; int in1line=0; long in1pos=0; /* 2015-02-26 +pos */
  /* 2016-04-15 startvec moved from global to local var, see randvec() */
  int  startvec=4; /* see vector.c randvec(), dflt: 4=random_v */
  int max_ea=1;	/* maximum expected degeneracy of ew[NE0] test-loop,dflt=1 */
  /* int? in2 for inlined .i-files, removed for simplicity 2012-10 */ 
  double r1;
  /* dynamic allocation of SH-buffer per thread, done before 
   *   allocation of rest memory for matrix, as thread-local mem */
  thbuf * (hbuf)[B_NUM];
  scplx *v0, *v1, *ev[NEV]; /* large vectors */
  minfo_t m_v0, m_v1, m_ev[NEV],  /* memory descriptors for vectors */
          m_l2 = {NULL,0,0,TMPDIR "l2.map",NULL};
#if defined(CONFIG_PTHREAD) && !defined(_OPENMP)
  /* stack memory is local, memory + filesystem is shared */
  pthread_t pt[B_NUM];
#endif
#ifdef CONFIG_MPI /* accept both (MPI+MPP) together */
  int mpi_thread=0; /* 2017-01-17 always output result for debugging */
  int mpi_thread_req=0; /* requested */
  const char *mpi_t[]={"SINGLE","FUNNELED","SERIALIZED","MULTIPLE"}; 
#if defined(CONFIG_PTHREAD) && defined(MPI_THREAD_SERIALIZED) /* supermuc */
  /*  needs OMPI compiled with --enable-mpi-thread-multiple (not default!)
   *  check: ompi_info | grep -i thread
   *         Thread support: posix (mpi: yes, progress: no)
   *    not on comp2, t100 y2016
   * try setenv LAM_MPI_THREAD_LEVEL=3
   * 0: MPI_THREAD_SINGLE      (works with thread-sync on comp2,kautz etc)
   * 1: MPI_THREAD_FUNNELED    (only main thread does MPI, max on kautz)
   * 2: MPI_THREAD_SERIALIZED  (only one does MPI at a time)
   * 3: MPI_THREAD_MULTIPLE  (this was wanted until y2016, but failes)
   *      test2017.01 mixed: crashes without thread-synchronization
   *                         how to use correctly? one dup_COMM_WORLD/thread?
   * y2016: ToDo: how data flow is with mixed code? intra PT inter MPI only?
   *  ToDo: faster/simpler/minMPI: one MPI/node using multiple threads y2016
   *  ToDo: finally it hould be better using (better) multithread on each node 2017.1
   */
  mpi_thread_req = MPI_THREAD_SERIALIZED;
  if (MPI_Init_thread(&argc, &argv, mpi_thread_req, &mpi_thread)
     != MPI_SUCCESS) exit(1);
  /* WARNING: provided thread-level may be below required thread level! */
#else
  if (MPI_Init(&argc, &argv) != MPI_SUCCESS) exit(1);
  i=0;
#endif
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_n);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_i);
  for (mpi_n_bits=0;mpi_n_bits<32;mpi_n_bits++)
     if (1<<mpi_n_bits > mpi_n-1) break;
#endif

{ /* global mpi-nodename since 2016-04-08 for better MPI-errors */
  char *s1=NULL; /* 2014-07 get hostname for debugging/errors */
// char nodename[21]="\0HOSTNAME..........."; /* 2016-04-08 set global for err() */
// #if  _BSD_SOURCE || _XOPEN_SOURCE >= 500 || _POSIX_C_SOURCE >= 200112L 
#if defined __USE_BSD || defined __USE_UNIX98  /* 2015-09-23 */
  /* ToDo: set global var and for MPI compute num remote/local connects */
  gethostname(nodename,20); /* Since glibc 2.12, unistd.h */
#endif                 
  /* not set on CentOS, call system("hostname -s > hostname"); ? no */
  s1=NULL; if (nodename[0]==0) s1=getenv("HOSTNAME");
  if (s1) strncpy(nodename,s1,20); /* getenv may return NULL, cpy(NULL)? */
  s1=NULL; if (nodename[0]==0) s1=getenv("SLURMD_NODENAME");
  if (s1) strncpy(nodename,s1,20); /* 2016-04-07 */
  nodename[21]=0; /* ensure string has an end */
}
  if (!mpi_i) { /* output only once per MPI cluster */
    const char *s1;
    printf("\n"
     /* __DATE__ __TIME__ avoids reproducable binaries, use file date */
      "SPINPACK version %s %s\n"
      " Copyright (C) 2001-2019 Joerg (dot) Schulenburg at ovgu.de\n"
      " released under the GNU General Public License\n"
      " base coding for %s-model (%s) NN=%d\n",
      version_string, version_date,
      CM("Heisenberg","t-J","Hubbard"), CM("JJ","tJ","tU"), NN);
    printf("Host: %s\n",nodename); /* 2014-07: debugging speed */
    s1="SLURM_CLUSTER_NAME";
              if (getenv(s1)) printf("DBG: %s=%s\n",s1,getenv(s1));
    /* for debugging only, file problems, two jobs at same path? */
    s1="PWD"; if (getenv(s1)) printf("DBG: %s=%s\n",s1,getenv(s1));
#if NOSZSYM // NoSzSym
 printf("\nWARNING: new feature NOSZSYM configured, code not tested!\n\n");
#endif
#if   defined(__INTEL_COMPILER)
    printf("Compiler: INTEL_COMPILER %.1f\n",__INTEL_COMPILER*.01);
    /* 1010 = 10.1,  __INTEL_COMPILER_BUILD_DATE 20070913 */
    /* __OPTIMIZE__ 1 (-O1)   CC+DBG=icce+idbe -E -dM */
    /* thread local storage (TLS) model -ftls-model= ??? */
    /* 2017-01 strange hangs using pthread -t64, see bugs  ToDo */
#ifdef __MIC__
    printf(" INTEL-CPU-Features: MIC" /* 2015-02 performance relevant */
#ifdef  __KNC__
        " KNC" /* Knight Corner = Xeon Phi */
#endif /* KNC */
#ifdef __INTEL_OFFLOAD
         " OFFLOAD" /* CoPro-offload */
#endif /* __INTEL_OFFLOAD */
         "\n");
#endif /* __MIC__ */
#endif /* icc */
#if defined(__GNUC__)
    printf("Compiler: GNUC %d.%d.%d\n",
                  __GNUC__, __GNUC_MINOR__, __GNUC_PATCHLEVEL__);
    /* gcc-4.7 AVX2=8*i32 BMI2=pext64=bit_extract Haswell-2013
               bitperm64= 6 times (pext(v,p))<<32|pext(v,~p)
               -mavx2 -mbmi2 -mlzcnt == -march=core-avx2 */
#endif
#ifdef __clang_version__ /* clang -dM -E -x c /dev/null */
    printf("Compiler frontend: Clang " __clang_version__ "\n");
#endif
#ifdef __llvm__  /* clang -dM -E -x c /dev/null */
    printf("Compiler backend: LLVM\n");
#endif
#if defined(__STDC_VERSION__) /* 199901L p.e. gcc -std=c99 +_Complex +// */
    printf("Compiler: STDC %ld\n",__STDC_VERSION__);
#endif
#ifdef __DECCXX_VER
    printf("Compiler: DECCXX %d\n",__DECCXX_VER);
#endif
#ifdef __SUNPRO_CC
    printf("Compiler: SUNPRO_CC 0x%x\n",__SUNPRO_CC);
#endif
#if defined(__mips) && defined(_COMPILER_VERSION)
    printf("Compiler: MIPS_CC %d\n",_COMPILER_VERSION);
#endif
#if defined(__PATHSCALE__)
    printf("Compiler: PATHCC " __PATHSCALE__ "\n"); /* SiCortex "3.2n" */
#endif
    printf(" CPU-Features:" /* 2014-06 performance relevant */
/* gcc: include/immintrin.h __MMX__ __SSE__ ... */
/* gcc: include/cpuid.h     bit_MMX ... bit_XOP ... */
#ifdef __MMX__  // 2i32 4i16 8i8 integer only (we need min. 64bit vectors)
  //   " MMX"  // P5_1997  registers: mm0..mm7
  // MMX:  pandn(dst,src): dst=(~dst)&src;
#endif
#ifdef __SSE__
     " SSE"  // pentium3  xmm0..xmm7 128bit=4f32  ADDPS, ADDSS, MULPS, MULSS
#endif
#ifdef __SSE2__
     " SSE2"  // pentium-m pentium4  XMM+2f64,16i8,8i16,4i32,2i64
     //  __m128d == 2dbl   __m128i == int8-64-arrays  _mm_*
     //  gcc -ftree-vectorizer-verbose=2 -ftree-vectorize -O2
     //  gcc -ftree-vectorizer-verbose=2 -O3
     //  icc-10.1
#endif
#ifdef __SSE3__
     " SSE3"  // complex-arith ADDSUBPD ...PS 
#endif
#ifdef __SSSE3__
//     " SSSE3"  // atom core2  PMULHRSW, PMADDUBSW
#endif
#ifdef __SSE4a__  // see no use exept POPCNT,LZCNT
     " SSE4a"  // core2++?  PMULLD PMULDQ
     // AMD-barcelona++ EXTRQ/INSERTQ mask-shift (bit-range get/set constidx)
     // SSE4.1: PMINUD, PMAXUD packed min/max unsigned double=64bit
     //   not usefull because we need index of max/min for sign
     //     BLEND/PBLEND conditional move (vector a=(b)?c:d) b=imm or reg.
     //     for sign + cfg transfer of vmin(vcfg)
     // SSE4.2: POPCNT,LZCNT count 1-bits i64, leading zeros 64bit
#endif
#ifdef __AVX__  // 4*f64 8*f32 4*i32(only)   quanta1+2 -mavx no256bitFMA?
     " AVX"    // 256bit, needs GCC-4.6 + k2.6.30 (AVX task switch!)
    // only low128bit for integer usable!  2*get_symconfig speed
    //  __m256, __m256d und __m256i   _mm256_*  = registers ymm0...15
    // Intel-Sandy-Bridge-Mikroarchitektur 32nm Jan2011
    // Intel-Ivy-Bridge-Mikroarchitektur   22nm Apr2012 +RDRAND f16 cfgTDP
    // AMD-Bulldozer 2011Q4, Piledriver 12Q4, Jaguar 13Q2, Steamroller 14Q1
#endif
#ifdef __AVX2__  // 4*f64 4*i64(!?) 8*i32(!) 32*i8  haswell-2013 noAMD2014
     " AVX2"     // 256bit-SSE-iOPs, 256bit for integers, +FMA3(3register)
     // VPSLLVD,VPSLLVQ shift left logical, variable shifts for each word
     //   ToDo: lNv-code uses variable-shift? scaling?
     // 3Op-3bit-table-logic  
     // /usr/lib/gcc/x86_64-linux-gnu/4.7.2/include/avx2intrin.h
     // Intel-Haswell-Mikroarchitektur 22nm Jun2013 (vec=Ivy+8%, single+6%)
     // gcc -march=haswell -dM -E -x c /dev/null # __haswell__
     // haswell E5-2630v3 2(FMA) *4(256bitAVX/64bitFloat) *2(ALUs?)= 16FLOP
#endif
#ifdef __AVX512F__  // ??? 8*f64  gcc-4.9 -mavx512f  Xeon-Phi no256bitFMA?
     " AVX-512F"     // AVX512F: only 32bit/64bit-vectors, register zmm0..31
     // + k0..k7 mask registers
     //  VPERMT2PD, VPERMT2Q  Full 64-bit permute overwriting first source.
     //                         Todo: try it !!!
     //  VPTERNLOGD, VPTERNLOGQ Bitwise Ternary Logic (512*3bit to 8bit-table)
     //              incl. 0x80=dblAND,0xFE=dblOR,0xca=BitwiseBlend...XOR
     //             In: 3 Regs + 1 8bit-table, Out: 1 Reg
     //             ~i1 | i2 | i3 made in one step = 3OPs/cycle for benes
     // ToDo: 128bit = (PLH64|PHH64)<<64 | (PLL64|PHL64) = 4P64+2
     //              = 4(4P32+2)+2 = 16P32+10 = 64P16+42 = 256P8+170 = bad
     // __AVX512BW__ byte/word extensions
     // __AVX512CD__ Conflict Detection to allow more loops to be vectorized,
     //              Knights Landing and Skylake Xeon
     // __AVX512PF__ Prefetch capabilities, Knights Landing
#endif
#ifdef __BMI__  //  bit manipulation instruction ANDN,BEXTR
     " BMI"
     // BEXTR bitfield extract with register x>>start & ((1<<len)-1)
     //    one clock instead (x>>i)&1
#endif
#ifdef __BMI2__  // -mbmi2 bit manipulation instruction 2 64bit pext+pdep
     " BMI2"     // ToDo: use it to speed up getsymcfg()
     // PDEP,PEXT parallel bits deposit + bits extract (ld(64)=6 bitperm)
// ToDo: BMI2: intel-haswell-2013 amd-excavator-2015(planed) gcc-4.8
//    Parallel bit deposit and extract  (for const. runtime bitperm)
//    6 times  pext(v,m)<<32 | pext(v,~m)   6*14cycles ? vs 64*4cycles = 1:3
//   what about VBPERM 32*1B-Perm. (256bit)
// unsigned __int64 _pext_u64(unsigned __int64 src, unsigned __int64 mask);
//   see http://en.wikipedia.org/wiki/Bit_Manipulation_Instruction_Sets
//  /usr/lib/gcc/x86_64-linux-gnu/4.7.2/include/bmi2intrin.h
#endif
#ifdef __3dNOW__  /* mm3dnow.h AMD only, obsolete */
     " 3dNOW"   // k8  mmx+2f32  PFADD, PFMUL, PAND=x&y(2*64bit)
#endif     
#ifdef __XOP__ //  AMD2009 (also SSE5)  opcode 0x8f  include/x86intrin.h
     " XOP"    // VPCMOV = Vector conditional move
               // VPPERM = Packed Permute Byte (select byte-list of 2 regs)
               // VPSHLQ  Packed Shift Logical Quadwords 64bit +/-reg !!!
     // http://en.wikipedia.org/wiki/XOP_instruction_set 0x8f+3byte
#endif
// ToDo: CVT16 = F16C 16bit-float-support, f16/f32-conversion only AMD 2009
#ifdef __MIPSEB__
     " MIPS_big-endian"    /* highest first, like 68k TCP */
#endif
#ifdef __MIPSEL__ /* lowest first, like 6502 Z80 x86 ... */
     " MIPS_little-endian" /* lowest first, simple-carry-propagation */
#endif
/* sometimes the binary is copied, not tuned for the platform, debug: */
/* gcc -dM -E - < /dev/null */
// gcc492 -mtune=native -dM -E -x c /dev/null | grep tune
#ifdef __tune_atom__  // gcc 4.7.2 atom 2008 N455-1.66GHz L2=512KB 2HT 6.5W
     " ATOM"         // N455.FSB=200MHz L1d=24kB delay=3cycles SSSE3
    // in-order + 16-stage-pipeline (cant change cmds in pipeline)
    // 2iALUs = ALU^SHFT + ALU (intel_atom_cpu*.pdf)
#endif
#ifdef __tune_core2__  // intel core2 gcc4.9.2  penryn-uarch? core-arch? 2007/2008
     // ToDo: ri2015 i5-3470 gcc472 no _tune!
     " CORE2"     // +SSE4.1
     // https://en.wikipedia.org/wiki/File:Intel_Core2_arch.svg 
     // 3ports=(ALU^FMUL/FDIV128+SSE_SHFL/ALU)
     //       +(ALU^FADD128^SSE_SHFL/MUL)
     //       +(ALU^SSE_ALU)
     // eff064=2(ALU^SHFL)+ALU      (i64)    shift?
     // eff128=1(ALU^SHFL)+SHFL+ALU (v2i64)  vshift?
     //   shuffle: reorder bytes within register (since penryn) ToDo: use?
     // J-PC-2008 E5400 2009-2011 core=Wolfdale-3M
#endif
#ifdef __tune_corei7__  
      // corei7     2009 SSE4.2
      // corei7-avx      SSE4.2+AVX+PCLMUL gcc472
      // core-avx-i      SSE4.2+AVX+PCLMUL+RDRAND+F16C gcc472 ri15.i5-3470
     " COREi7"       // advantages? +SSE4.2? more cores/clock/L2 than i3, i5
     //  (i7 for all uarchs=nehalem...haswell = business-version)
#endif
#ifdef __tune_sandybridge__  // intel sandy-bridge gcc 4.9.2 (succ_of_nehalem)
     " SandyBridge"  // advantages? AVX=256bit-FP
     // 3ports=(ALU^LEA^SHIFT^FMUL256^SIMD_VMUL/VSHIFT128)
     //       +(ALU^LEA^MUL^SIMD_VALU/VSHUF128)
     //       +(ALU^SHIFT^SHFL256^SIMD_VALU/VSHUF128)
     //  eff064=2(ALU^SHIFT)+1ALU
     //  eff128=1SHIFT+2ALU (=SIMD)
     //  PMOVSMKB helpful?
     // http://www.intel.com/content/dam/doc/manual/64-ia-32-architectures-optimization-manual.pdf
#endif
#ifdef __tune_haswell__ // haswell 2014/2015 boost (succ_of_sandybridge)
     " HASWELL"   // AVX2 = AVX + i64-SIMD256, BMI2
     // 4Aports=(ALU^SHFT^FMA256^VMUL^VSHFT)
     //        +(ALU^MUL^FMA256^VALU) +(ALU^FSHFL256^VALU^VShuffle)
     //        +(ALU^SHFT)
     // i64:  2(ALU^SHFT) + 2ALU
     // v256: VSHFT + 2VALU + i64(ALU^SHFT)
     // not in gcc4.8.3
#endif
#ifdef __tune_core_avx2__ // gcc-4.8.3 --march=core-avx2 # =haswell
    " core_avx2"
#endif
#ifdef __k8__  // AMD K8 SSE2-path 64bit only 128bit/2clks??? L2=128bit/clk
     " K8"   // L1=L2=128bit-bus L1 w=2*64bit 3ALUs64bit 2SSE-ALUs
     // 12-stage-pipeline (k7=10-stages)
     //   no exceptions explored running k10-binary on k8 (comp1)
     //   only bn8 is much slower! k8=28clks vs. k10=13clks
#endif
#ifdef __tune_amdfam10__   // AMD K10  2SSEread=2*128bit/clk = 4*f64/clk
     " K10"    // +LZCNT,POPCNT,EXTRQ,INSERTQ  L1=L2=256bit-bus L1=64k-2way
     // L1 w=2*64bit r=2*128bit  3ALUs64bit 2ALUs128bit(1FADD+1FMUL)
     //   gcc may generate same binary for k8 and k10
#endif
#ifdef _MIPS_ARCH   // Sicortex-MIPS64 "5kf" (MIPS incl. FP+caches) 2009
     " MIPS_ARCH=" _MIPS_ARCH  // string "5kf"
     //  MIPS-III = 64bit = R4k-series
     //   Power challenge R10k 1996 super-scalar + out-of-order-exec
     //  MIPS-IV R8000 2ALUs +1SHIFT-Unit (shared? alu^shift?)
     //           +1Mul/Div(not pipelined,4clk/MUL32,6/MUL64,DIV=21..73)
     //  nor: d=~(s|t)
     //   see http://en.wikipedia.org/wiki/MIPS_instruction_set + refs
     //  gcc412  -dM -E -x c /dev/null | grep MIPS # == pathcc
     //   _MIPS_ARCH="5kf" _MIPS_TUNE="5kf" _R4000=1 _ABI64=3 _MIPS_ARCH_5KF=1
     //   -mpaired-single __mips_paired_single_float=1 2f32 in f64-reg
     //   try: -mbranch-likely
     //  gcc492 -x c -E -dM a.h | grep MIPS # _MIPS_ARCH "mips3"
     //  gcc492 -march=native -mtune=native -x c -E -dM a.h # "mips3"
     //  gcc492 -march=mips4  -mtune=mips4  -x c -E -dM a.h # _MIPS_TUNE_MIPS4
     //  gcc492 -march=5kf    -mtune=5kf    -x c -E -dM a.h # "5kf" 2017-09
#endif
#ifdef __SICORTEX__
     " SICORTEX"
#endif
#if defined(__LP64__) || defined(__x86_64__)  /* __WORDSIZE == 64 */
    " int64"
#endif
#if defined(__SIZEOF_INT128__)  /* __int128 available gcc4.7-4.9, ~gcc4.4 */
    " int128"
#endif
     "\n");
#if (!defined(__LP64__)) && (!defined(__x86_64__)) && NN*CM(1,2,2)>32
    printf("WARNING: NN>32 on a 32bit machine may be about 4 times slower\n");
#endif
  } /* !mpi_i */

/* we tested OpenMP very earlier, but performance was poor */
/* ToDok 2016 implementing pthread-structure using OpenMP-v3.tasks? */
#if defined(_OPENMP)  /* OpenMP (former used, but not now, later again?) */
  if(!mpi_i)printf(" OpenMP-Support:"
#if   _OPENMP >= 201511  /* YYYYMM of openmp-spec */
         " OMP-4.5" /* */
#elif _OPENMP >= 201307 
         " OMP-4.0" /* gcc 4.9 +simd+acceleraturs+... */
#elif _OPENMP >= 201107 
         " OMP-3.1" /* gcc 4.7 */
#elif _OPENMP >= 200805 
         " OMP-3.0" /* gcc 4.4  +tasks (omp task is needed 2017-01-08) */
#elif _OPENMP >= 200505  /* 2000=fortran 2002=cc/c++ 2005=f/c.combined */
         " OMP-2.5" /* gcc 4.2, icc 10 for+sections? (2017-01-09) */
         /* test g++4.1.2 warn only + exe is hanging 2017-01 */
#elif _OPENMP >= 199810  /* YYYYMM of openmp-spec */
         " OMP-1.0 (but need 2.5)" /* gcc 4.2, icc 10 */
#else /* 199810 = 1.0 */
         " unknown OMP version (but need 2.5)"
#endif /* _OPENMP values */
         /* icc10.1=200709 -openmp -E -dM _OPENMP=200505 = OMP 2.x? no task */
         /* icc10 -openmp-stubs = dummy-openmp lib, sequentiell, no macro */
         /* icc10 -cluster-openmp needs extra-LIC, 2010 discontinued */
         /* gcc4.1  _OPENMP=200505 */
      " OPENMP=%d\n",_OPENMP);
#endif /* _OPENMP defined */


/* 2016-02-25 see defined constants for MPI-1.1 MPI_SUCCESS=0 */
#if (!defined(CONFIG_MPI)) && (defined(OMPI_MAJOR_VERSION)\
      || defined(MPICH2) || defined(MPICH_VERSION) || defined(MPICH2_VERSION)\
      || defined(MPI_VERSION))
  if(!mpi_i)printf("WARNING: MPI-Compiler detected, but no CONFIG_MPI set\n");
#endif

/* ToDo2017: need to build own MPI_IAlltoallv, to replace MPI_Alltoallv */
/* ompi-1.4.0/mpi.h MPI_VERSION=2.1 no  MPI_IAlltoallv CentOS5 */
/* ompi-1.6.5/mpi.h MPI_VERSION=2.1 no  MPI_IAlltoallv, has MPI_Alltoallv */
/* ompi-1.7.4/mpi.h MPI_VERSION=3?  has MPI_IAlltoallv, has MPI_Alltoallv */
/* ompi-1.8.4/mpi.h MPI_VERSION=3.0 has MPI_IAlltoallv, has MPI_Alltoallv */
#ifdef CONFIG_MPI
  { int version, subversion; const char *s1=NULL;
    MPI_Get_version(&version,&subversion); 
    // ToDo18 compile version may differ from runtime version (dyn. libs)
    //       but: MPI-3.0 OMPI-1.8.4 using MPI-3.0 problem is 1.10.0 usage
    if (!mpi_i) {
      printf("MPI:");
#if defined(MPI_SUBVERSION) /* 2016-02-25 seen in ompi-1.8.4/mpi.h */
      printf(" MPI-%d.%d",MPI_VERSION,MPI_SUBVERSION);
#elif defined(MPI_VERSION) /* +MPI_SUBVERSION */
      printf(" MPI-%d",MPI_VERSION);
#endif
#if defined(OMPI_MINOR_VERSION)  /* using stringification of macros */
      printf(" OMPI-%d.%d",OMPI_MAJOR_VERSION,OMPI_MINOR_VERSION);
#ifdef OMPI_RELEASE_VERSION
      printf(".%d",OMPI_RELEASE_VERSION);
#endif
#elif defined(MPICH_VERSION)
      printf(" MPICH-" MPICH_VERSION);
#elif defined(MPICH2_VERSION)
      printf(" MPICH2-" MPICH2_VERSION);
#elif defined(MPICH2)    /* sicortex pathscale mpicc-3.2n */
      printf(" MPICH2 (unknown)");   
#else
      printf(" non-OMPI/MPICH"); 
#endif
      printf(" using MPI-%d.%d\n", version, subversion);
      printf("MPI: MPI_MAX_NUM=%d  mpi_n_bits=%d  mpi_n= %d\n",
                   MPI_MAX_NUM, mpi_n_bits, mpi_n);
      /* 2017.01 moved here */
#if defined(MPI_THREAD_MULTIPLE)
      printf("MPI: MPI_THREAD_MULTIPLE supported\n");
#else
      printf("DBG: MPI_THREAD_MULTIPLE not supported\n"); /* just info */
#endif
      printf("MPI: mpi_thread_level=%d (MPI_THREAD_%s) requested=%d \n",
        mpi_thread, mpi_t[mpi_thread&3], mpi_thread_req);
      if (mpi_thread != mpi_thread_req)
        printf("DBG: provided != requested mpi_thread_level\n");
      printf("MPI: NZ_BUFSZ1/AH=%.3f  AH=%d\n",
               MPI_NZ_BUFSZ1*1./NUM_AHEAD_LINES, NUM_AHEAD_LINES);
#if USE_Alltoallv_Emulation /* useful on SiCortex 2009, 15x faster */
      printf("MPI: using [I]Alltoallv_Emulation (for bad scaling A2av)\n");
#else  /* ~USE_Alltoallv_Emulation, old loop over sendrecv was slower */
      printf("MPI: using better scaling Alltoallv (since 2017-01-16)\n");
#endif /* USE_Alltoallv_Emulation */
      printf("MPI: partly nonblocking Ialltoallv (since 2017-02-24)\n");
#if 1
      s1="PBS_NUM_NODES"; 
      if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
      s1="PBS_NUM_PPN";
      if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
      s1="SLURM_NNODES";
      if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
      s1="SLURM_TASKS_PER_NODE";
      if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
      s1="OMPI_COMM_WORLD_SIZE"; /* NODES */
      if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
      s1="OMPI_COMM_WORLD_LOCAL_SIZE"; /* PPN */
      if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
      /* OMPI: Aggressive or Degraded performance mode */
      /* default: degraded on oversubscribe mode */
      s1="OMPI_MCA_mpi_yield_when_idle";
      if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
#if defined(OMPI_MINOR_VERSION)  /* power saving */
      else printf("MPI.HELP: try setting %s=1 (busy_wait=off)\n",s1);
#endif
      s1="I_MPI_WAIT_MODE"; /* intel MPI */
      if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
      s1="OMPI_MCA_hwloc_base_binding_policy"; /* oversubscribe mode? */
      if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
#if defined(OMPI_MINOR_VERSION)
      else printf("MPI.HELP: try setting %s=none\n",s1);
#endif
      s1="I_MPI_PIN"; /* intel MPI 0=idle 1=busy wait*/
      if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1)); 
      /* ToDo: OMPI_MCA_orte_hnp_uri for DBG network problems? */
#endif
    } /* !mpi_i */
  } /* config_mpi */
  if (mpi_n>MPI_MAX_NUM) {
    if (!mpi_i) printf("ERROR: please enlarge MPI_MAX_NUM"); 
    sleep(2);MPI_Abort(MPI_COMM_WORLD,-1); }
#endif

#if 0
  /* 2013-04-14 set to maximum checkpoint, from latest run */
  /*    may cause errors if l1[] is changed, so
   *    its better to set by hand to what is intended by the user (option)
   */
  if(!mpi_i){ /* read checkpoint state, may be replaced by option or ifile */
    FILE *f; int i=0;
    f=fopen(FCHKPT,"rt");
    if (f) {i=fgetc(f);
      if      (i>='0'&&i<='9') checkpoint=i-'0';
      else if (i>='a'&&i<='z') checkpoint=i-'a'+10;
      else i=0;
      fclose(f);
    }
    if (checkpoint)
      printf("WARNING: set checkpoint to %d from "FCHKPT"\n", checkpoint);
  }
#endif

#ifdef _OPENMP  /* override CONFIG_PTHREAD */
  if (!mpi_i) {  /* 2015-12 ToDo: add support, start on SH */
    int omp_n=1; // omp_set_num_threads(2);
    /* printf("DBG: compiled for OpenMP but unsupported until 2017-01\n"); */
    /* 2017-01-10 fixed mixed code MPI + OMP */
#pragma omp parallel
    omp_n= omp_get_num_threads();  /* must be called within parallel pragma */ 
    /* next line is outside parallel region */
    printf("DBG: OpenMP threads=%d\n", omp_n); /* added 2016-05-03 */
    pt_n=omp_n; /* set as default num threads */
  }
#ifndef CONFIG_PTHREAD
#error "CONFIG_PTHREAD is needed for _OPENMP (2017-01 but subject of change)"
#endif /* PTHREADS */
#endif /* _OPENMP */

  for (i=0;i<B_NUM;i++) hbuf[i]=NULL; /* better debugging 2015-07-08 */
  /* 2016-05-18 add checking OMP_NUM_THREADS, overwrite by -t */
  if (getenv("OMP_NUM_THREADS")) {
    pt_n=atoi(getenv("OMP_NUM_THREADS"));
    if (!mpi_i) printf(" set pt_n=$OMP_NUM_THREADS=%d\n", pt_n);
  }
#ifdef CFG_FTLM /* debugging, 2018-08-22 */
  if (!mpi_i) { printf("CFG_FTLM enabled\n"); }
#endif  
#ifdef LAPACK /* debugging, 2018-08-22 */
  if (!mpi_i) { printf("LAPACK enabled\n"); }
#endif  
  
  // Bcast_args: dat,sz,typ,snode,..
  MPI_Bcast(&checkpoint, 1, MPI_INT, 0, MPI_COMM_WORLD);
  i=0;while(++i<argc){ ss=argv[i];
    if (vvv && !mpi_i) printf("DBG: parse argument %d: %s\n",i,ss);
    if (ss[0]=='+' && ss[1]=='v'){vvv|= atoi(ss+2);continue;} /* 2017-02-21 */
    if (ss[0]=='-'){
      if (ss[1]=='v' && ss[2]!=0){vvv=  atoi(ss+2);continue;} /* verbose */
      /* -i daten.i, ToDo: stdin */
      /* 2017-02-21 add "-i a.i" version, which supports bash + TAB */
      if (ss[1]=='i' && ss[2]!=0) {    name_file_i=ss+2;   continue;}
      if (ss[1]=='i' && ss[2]==0) {i++;name_file_i=argv[i];continue;}
      /* daten.def */
      if (ss[1]=='m' && ss[2]!=0) {    name_file_m=ss+2;   continue;}
      if (ss[1]=='m' && ss[2]==0) {i++;name_file_m=argv[i];continue;}
//  2018-11-14 do not split output (see above prints, simplify dbg)
//      if (ss[1]=='o' && ss[2]!=0){ /* redirect stdout */
//                      if (!mpi_i) /* avoid chaos (May08) but is Nov18 */
//                       if (freopen(ss+2,"a",stdout)==NULL) E0("freopen");
//                                         continue;}
      /* threads per node, no spaced version */
      if (ss[1]=='t' && ss[2]!=0) {    pt_n=atoi(ss+2);   continue;}
    }
    /* short run, to get best estimation of performance 2017-01-13
     *   this is if CPUh are limited and we want to now how much it will eat */
    if (strncmp(ss,"--perf_test=",12)==0){
      opt_perf_test= (ULong) atof(ss+12); /* allow 1e5, 1e6 etc. */
//      if (opt_perf_test && opt_perf_test<100000UL) opt_perf_test=1000000UL;
//      if (opt_perf_test<2*NUM_AHEAD_LINES) opt_perf_test=2*NUM_AHEAD_LINES;
      if(!mpi_i)printf("DBG: set opt_perf_test=%lu\n",opt_perf_test);
      continue;}
    /* packed nonzero elements = sum nzr of same index, ignore nzr=0 */
    /* sq40_2+38 packed=0 [40 12][10 35],
     *           packed=1 [20  5][ 5 15]
     * sq40_6+34 packed=0 [197410 56564][57620  222031]  sum=asym +1..5%
     *           packed=1 [195420 56409][56409  217375]  sum=sym   */
    if (strncmp(ss,"--pack_nz=",10)==0){
      opt_pack_nz= atoi(ss+10);
      if(!mpi_i)printf("DBG: set opt_pack_nz=%d\n",opt_pack_nz);
      continue;}
    /* chkpt_mode: set checkpoint behavior, 0=no, 1=save, 2=restore */
    /* chkpt_load: set resume checkpoint, 0=start, 1=ns, 2=SH ?, 4=SH.it */
    if (strncmp(ss,"--chkpt_load=",13)==0){checkpoint=atoi(ss+13);continue;}
    if (strncmp(ss,"--chkpt_time=",13)==0){chkpt_time=atoi(ss+13);
      if(chkpt_time<15000 && !mpi_i) // 2014-09-19
        printf("WARNING: chkpt_time<25m is very small ...\n"
               " ... good choice is between 8x iteration time"
               " and 1/8 failure freq.\n");
      if(!mpi_i) /* 2015-02-21 */
        printf("DBG: int chkpt_time = checkpoint interval in seconds = %d\n",
         chkpt_time); 
      continue;
    }
//    if (strncmp(ss,"--chkpt_mode=",13)==0){chkpt_mode=atoi(ss+13);continue;}
    if (!mpi_i) printf("ERROR: unknown argument %d \"%s\"\n", i, ss);
    if (!mpi_i) /* ToDo18: unify daten.def + daten.i (simply concat?) */
      printf("options: [-v<verbose>] [-i<daten.i>] [-m<daten.def>]"
             "\n... [-t<threads_per_node>]" /* 2018-11 -o<outfile> removed */
             "\n... [--chkpt_load=4] [--chkpt_time=3600] [--perf_test=1e2]"
//             " [--chkpt_mode=3]"
//          " [--scratch_disk=tmp/] [-s./tmp/]" // shared|local scratch disk
             "\n");
    if (!mpi_i) /* 2016-05-18 added, grep getenv */
      printf("environment: OMP_NUM_THREADS\n");
    sleep(2); MPI_Abort(MPI_COMM_WORLD,1); return 1;
  } /* while ++i<argc */
  setvbuf(stdout, (char *)NULL, _IONBF, 0); /* no buffer */

  if (opt_perf_test==0 && !mpi_i) /* 2017-01-13 ToDo also set in daten.i?? */
    printf("INFO: try option --perf_test=100 for a n1/100 performance test\n");
  if (opt_perf_test!=0 && !mpi_i) /* 2017-01-13 ToDo also set in daten.i?? */
    printf("\n * * *\nWARNING: performance test only, no results!\n * * *\n");

  if (pt_n<1) pt_n=1;  /* silent correction to minimum 2014-06 */

#if defined(CONFIG_PTHREAD) && B_NUM==1
  if (!mpi_i)  /* 2014-06 */
    printf("WARNING: spinpack is compiled for pthreads"
           " but B_NUM=1 limits the tasks to 1\n");
  /* you should set B_NUM to 2 or higher if you want to use pthreads */
  /* 2016++ B_NUM can be set via CONFIG_PTHREAD set in Makefile */
#endif

  if (pt_n>B_NUM) {
      pt_n=B_NUM; /* set to maximum (1 if single-cpu code) */
    if (!mpi_i)
      printf("WARNING: number of threads is set back to maximum (B_NUM)\n");
  }
  for (pt_n_bits=0;pt_n_bits<32;pt_n_bits++)
     if (1<<pt_n_bits > pt_n-1) break;
  if (pt_n_bits==32) F0("pt_n do not fit into 32bits");

#if CFG_CPUSET < 2 /* ToDo17: set as default or set at runtime! */
  if (pt_n*mpi_n>1 && vvv && !mpi_i) printf( /* 2017.01 autoset bits */
    " please try option -DCFG_CPUSET=2 -D_GNU_SOURCE on multi-core systems\n"
           "   or use taskset, numactl, omplace, dplace to pinpoint tasks\n");
  /* ToDo: -openmp as replacement for pthread_create/start_parallel ? */
#endif
  if (pt_n==1) pt_cpuset(0); /* 2017-01 show + set affinity of thread/task */
#if defined(CFG_CPUSET)
  /* wait for outputs of worker threads */
  if (pt_n> 1) usleep(((mpi_n>1)?4:1)*20000);
#endif
  /* give vectors blocks different names for shared filesystems */
  m_v0.p=NULL; m_v0.l=0; m_v0.mode=0; m_v0.ff=NULL;
  sprintf(m_v0.name,TMPDIR "v0_%04d.map", mpi_i);
  m_v1.p=NULL; m_v1.l=0; m_v1.mode=0; m_v1.ff=NULL;
  sprintf(m_v1.name,TMPDIR "v1_%04d.map", mpi_i);

  v0=v1=(scplx *)NULL;	/* NULL = not allocated */
  for(i=0;i<NEV;i++){
    m_ev[i].p=NULL; m_ev[i].l=0; m_ev[i].mode=0; m_ev[i].ff=NULL;
    sprintf(m_ev[i].name,TMPDIR "ev%d_%04d.map", i, mpi_i);
    ev[i]=(scplx *)NULL;
  }

  l2b=NULL; l1=NULL; /* initialization */
  iv=0; algo=0; x_p=3; n1=0l; nn=NN; nw=0*Nw; ne0=0;
  sym.subl=0; /* if 0 auto detection will be used */
  hmem_nzmax=0;	/* no limit for hxy.tmp size 2015-01 */
  x_max=5; if(MAXx<x_max) x_max=MAXx;   /* max. num. of output-configs */
  /* bug: x_max=min(5,MAXx); - this gives warnings on SGI-CC, JS Mar99 */
  pew=2; nev=min(pew,NEV); /* added 2016-05 */ cfg_sisj=1;
  itime();	/* init itime, maxmem=0(infty, removed 2013) */
  /* output some info for possible bug reports */

  ini_handler(mpi_i, vvv); /* signalhandler (sighand.h) after pthread_create? */

#ifdef CONFIG_PTHREAD
  /* parallel but share everything exxcept local variables */
  if (!mpi_i)
   printf("PTHREAD max=B_NUM=%d threads/mpi-task, pt_bits=%d  pt_n= %d\n"
          "PTHREAD use option -t<threads> [-v4]\n", B_NUM, pt_n_bits, pt_n);
  MPI_Barrier(MPI_COMM_WORLD); /* sync "I am here" output */
  /* ToDo17: OPENMP-alternative to pthread? define pthread-dummies? */
#ifndef _OPENMP
  /* IRIX, True64 and Linux works with NULL=default, but SunOS does not */
  pthread_attr_init (&pt_attr);
  pthread_attr_setscope (&pt_attr, PTHREAD_SCOPE_SYSTEM);
  pthread_mutex_init(&pt_mutex1,NULL);
  pthread_mutex_init(&pt_mutex2,NULL);
  pthread_cond_init(&pt_cond1,NULL); /* correct init? */
  pthread_cond_init(&pt_cond2,NULL); /* correct init? */
  if (pthread_barrier_init(&pt_barrier, NULL, pt_n))
    { printf("PTHREAD.ERROR: Could not create a barrier\n");ERR(9999); }
#endif
// ToDo: #pragma parallel (mother-thread) + #pragma openmp for (child-threads)
// #pragma omp parallel num_threads(pt_n+1)
// {  
//    #pragma omp for schedule(static,1)
//    for (0..pt_n) { 
//       if (i>0) pthread-main-thread();
//       if (i=0) { // master_thread
//    ......
//    }} // close master + omp_for
// }
// // end-of-PG
// --OR-- tested v2017-01:
//#pragma omp parallel num_threads(pt_n+1) // starts next blk on all threads
//{
//#pragma omp master // master=thread0.single, single=ok? after for?
//{
//  for (i=0; i<pt_n; i++) {
//#pragma omp task  // needs OpenMP v3
//    worker(i);
//  }  // for
//  ....
//}} // end master + parallel
// --OR-- untested v2017-01:
//#pragma omp parallel num_threads(pt_n+1) // starts next blk on all threads
//{
//  for (i=0; i<pt_n; i++) {
//#pragma omp single nowait // needs OpenMP v2.5 ???
//    worker(i);
//  }  // for
//#pragma omp single nowait // also: omp master as thread0.single
//{
//  ....
//}} // end master + parallel
  /* create main child-threads */
// pragma parallel opens all threads and starts them
//#pragma omp parallel for schedule(static,1)
#pragma omp parallel private(i) shared(main_pt) num_threads(pt_n+1)
{
//#pragma omp master /* master = single where thread=0 but no nest = OK */
//#pragma omp single /* single like master = OK */
//#pragma omp task
//{   // moved below! 2017-01++
 int i; /* 2017 needed for OMP */
// #pragma omp for schedule(static,1) private(i)
  for (i=0;i<pt_n;i++) {
#ifndef _OPENMP
    pthread_mutex_init(&token_ring[i       ],NULL);
    pthread_mutex_init(&token_ring[i+  pt_n],NULL);
    pthread_mutex_init(&token_ring[i+2*pt_n],NULL);
    pthread_mutex_init(&token_ring[i+3*pt_n],NULL);
    pthread_mutex_lock(&token_ring[i       ]);
    pthread_mutex_lock(&token_ring[i+  pt_n]);
    pthread_mutex_lock(&token_ring[i+2*pt_n]);
    pthread_mutex_lock(&token_ring[i+3*pt_n]);
#endif
//#pragma omp task  /* needs OMP-3 or higher */
#pragma omp single nowait /* needs OMP-2.5 only !? */
{
#pragma omp critical
{
    main_pt[i].num=i;     /* put thread number to child */
    main_pt[i].data=NULL; /* for first do nothing */
    main_pt[i].task=0;    /* initialize */
} /* pragma omp critical */
#ifdef _OPENMP
    if (i<8 && !mpi_i) printf("DBG_OMP: create thread %d\n",i);
    main_pthread((void *)&(main_pt[i]));  /* child = worker thread */
#else
    if( pthread_create((pthread_t *)&(pt[i]),
                       (const pthread_attr_t *)&pt_attr,
                       (void * (*)(void *)) &main_pthread,
                       (void *)  &(main_pt[i]) ) !=0 )  F0("pthread_create");
#endif
} /* pragma omp task */
  } // for i=0...
#pragma omp single
{ // master thread moved here 2017-01++ 
#ifdef _OPENMP
  usleep(30000); /* DBG_OMP better readable if ordered */
  if (!mpi_i) printf("DBG_OMP: master still alive, go on ...\n");
#endif
  if (vvv&4) // reduce DBG output
  { /* coordinate the output of "I am here!" messages */
    int ii;
    for (ii=0; ii<pt_n; ii++)
      parallel_start(ii, test_pthread, (void *)&main_pt[ii]);
    for (ii=0; ii<pt_n; ii++)
      parallel_wait(ii);
  }
  pthread_mutex_unlock(&token_ring[0     ]); /* put token into the ring1 */
  pthread_mutex_unlock(&token_ring[2*pt_n]); /* put token into the ring2 */
#else  /* ifdef CONFIG_PTHREAD */
  if (!mpi_i)
   printf("PTHREAD _not_ used, %d thread (per node, max=B_NUM=%d, bits=%d)\n",
      pt_n, B_NUM, pt_n_bits);
  if (pt_n!=1) {
    pt_n=1; pt_n_bits=0;
    if (!mpi_i) printf("Warning: reset number of threads to 1\n");
  }
#endif
  
  MPI_Barrier(MPI_COMM_WORLD); /* sync "I am here" messages */
  if (!mpi_i) { /* to have better DBG overview, output macros */
    printf(" Hcfg: H_JZ=%d Jadd=%3.1f Ham2=%d Ham3=%d HFIELD=%d SZSYM=%d\n",
           H_JZ, H_Jadd, Ham2, Ham3, HFIELD, 1-NOSZSYM);
    printf(" using  L1_PACKED=%d VecType=%d Tnorm2=%c%3s%d\n",
      L1_PACKED,  VecType, /* L1=tbase_vector */
      (+1==(int)(((Tnorm2)-1)*((Tnorm2)-1))?'_':'u'),
      ((((Tnorm2)1)/2==0)?"int":"flt"),8*(int)sizeof(Tnorm2));
//  printf(" Tnorm2 [unsigned][int,flt][bits]= %c%c%d\n", /* ui64 or _f64 */
//     (+1==(int)(((Tnorm2)-1)*((Tnorm2)-1))?'_':'u'),
//     ((((Tnorm2)1)/2==0)?'i':'f'), (int)sizeof(Tnorm2)*8 );
    printf(" h_file.h: TYPE_SIZE XY=%d BY=%d HR=%d in Byte, Hsize=%.0fkB\n",
     (int)sizeof(XY_TYPE), (int)sizeof(BY_TYPE), (int)sizeof(HR_TYPE), 
     Hsize/1024.);
    printf(" using CKPT_HELP=%d (pkill -SIGUSR1 spin  for safe ckpt)\n",
           CKPT_HELP);
#ifdef HAVE_ZLIB    // defined per Makefile DFLAGS as 1 (or not)
    printf(" using ZLIB=%d\n", HAVE_ZLIB); // old: defined(HAVE_ZLIB)
#endif
#ifndef ReserveMB
#ifdef  OMPI_MINOR_VERSION
#define ReserveMB 150 /* comp2 22*16cores=142MB+4MB ompi-1.8 ToDo: checkit! */
#elif   CONFIG_MPI
#define ReserveMB 26  /* kautz=5800cores/18*64kB=21MB+4MB(checkpoint) mpich2 */
#else
#define ReserveMB 5   /* 4MB(checkpoint)+1MB(local data) 2015-06-10 */
#endif /* OMPI, MPI */ 
#endif /* ReserveMB */
#ifdef _Vlint_h
    printf("WARNING: using class Vlint for %d-bit emulation"
           " (about 10 times slower)\n", (int)(sizeof(vlint)*8));
#endif
    printf("use symmetric base configurations (mandatory)\n");
#ifdef CONFIG_S1SYM
    /* ToDo: replace by set of commutating symmetries */
    printf("S1SYM: use S1-symmetric base configurations (default)\n");
#else
    printf("NOS1SYM: no S1-symmetries (slow on S1-models, debug)\n");
#endif
#ifdef AddSS
    // ist das full matrix? nein!? oder nur N^2/4+Nw per line
    printf(" using AddSS=" #AddSS " (H+float_const*Op_SS)\n");
#endif
#ifdef DEBUG
    printf(" using DEBUG=%d (more checks, lower speed)\n", DEBUG);
#endif
  /* BY_TYPE defined in h_file.h is ushort (Jun08) = 16bit = max65536 threads
   *  (badly triggered 2018-05: 2048nodes*2tasks/node*16threads)
   * since v2.36 2008.08 BY is used for MPI only, threads work on same vblk
   *   can be expanded to 32bit to allow 2e9 threads */
#if 0 /* bad since v2.36 */
    if (((BY_TYPE)(pt_n*mpi_n))!=pt_n*mpi_n)
       printf("\nWARNING: pt_n*mpi_n dont fit to BY_TYPE (h_file.h)\n\n");
    if (((((BY_TYPE)mpi_n)<<pt_n_bits)>>pt_n_bits)!=mpi_n)
       printf("\nWARNING: mpi_n<<pt_n_bits dont fit to BY_TYPE (h_file.h)\n\n");
#else
    if (((BY_TYPE)(mpi_n))!=(unsigned)mpi_n)
       printf("\nWARNING: mpi_n dont fit to BY_TYPE (h_file.h)\n\n");
#endif
    /* ToDo test by BY_TYPE = char + 255 and more tasks */
    if (((BY_TYPE)(mpi_n-1))!=(unsigned)mpi_n-1) { // makes sence only for BY_TYPE<INT
      printf("ERROR: mpi_n-1 dont fit to BY_TYPE (see h_file.h)\n");
      ERR(9999); }
  } /* if !mpi_i */

  /* create the child early to save memory on pipe or fork */
  MPI_Barrier(MPI_COMM_WORLD);

  {int i;for (i=0;i<pt_n;i++){  /* ini memH old.STOREH==0 */
    memH[i].ptr=NULL; memH[i].len=0; /* memH[i].used=0;*/ }} /* ini nzr_mem */
  sub_ini();		/* v1.6.1 start child for gzip, ini memH */
  /* relax nfs, use cache 2016-03-16 kautz 5000 tasks may cause OOM */
  if (mpi_i) usleep(mpi_i/512*1000+10000); /* 5120 tasks = 0.2s */

  /* output helpful configuration for support/debug,
   * moved from after modeldef..ini_sym.{k,w,l} to before modeldef 2016-04-06
   * add Ahead */
  if (!mpi_i) {
    const char *s1;
#ifdef CplxInfo
    printf("CplxInfo: " CplxInfo "\n"); /* compiler specifique cplx def */
#endif
    printf(" long=%d int=%d  double=%d sdouble=%d  cplx=%d scplx=%d\n"
           "  hxy=%d/%d=%d tbase1=%d tbase=%d",
    (int)Lsize, (int)sizeof(int), (int)sizeof(double), (int)sizeof(sdouble),
                                  (int)sizeof(cplx),   (int)sizeof(scplx),
    (int)Hsize,HBLen,(int)Hsize/HBLen,(int)sizeof(tbase1),(int)sizeof(tbase));
    /* added 2016-04-06 for debugging/support */
    printf("\n NumAheadLines= %d= %dK\n",NUM_AHEAD_LINES,NUM_AHEAD_LINES>>10);
    s1="SLURM_CPU_BIND"; /* 2018-11-11 */
              if (getenv(s1)) printf("DBG: %s=%s\n",s1,getenv(s1));
    s1="SLURM_CPU_BIND_LIST";
              if (getenv(s1)) printf("DBG: %s=%s\n",s1,getenv(s1));
  }
  if (sizeof(int)  != S_int) {
      printf ("Error: sizeof(int)   != S_int\n");    ERR(9999); } 
  if (sizeof(long) != S_long) {
      printf ("Error: S_long  != sizeof(long)\n");   ERR(9999); } 
            

  if (initvars(mpi_i, name_file_m)) { /* modeldef.c:read(daten.def) */
    E0("load model failed, sleep8s + 2nd try");sleep(8);
    if (initvars(mpi_i, name_file_m)) F1("load model from %s",name_file_m);
  }
  initmasks(nn);	/* initialize bitmasks for lintab/sym/etc. */
  p_mask_ini(nn);       /* initialize bitmasks for parity() */

  // MPI_Barrier(MPI_COMM_WORLD);  // removed 2016-03-16

  /* initialize ud-symmetry data, HFIELD or TBC? */
  sym.wud= +1*(1-NOSZSYM)*(1-HFIELD);     /* initialize symmetry data */
  /* must be set to zero in case of nu!=nd and set back 2018.05 ToDo */
#if defined(CONFIG_TBC) /* sym.wud=0 */
  sym.wud=0;		/* do not use ud-sym */
#endif     /* reset sym.wud */
  sym.wud_saved= sym.wud;
#ifdef Sud
# error "macro Sud removed, switch off sym_ud by setting sym.wud=sym_ud=0"
#endif
  
  for (i=0;i<MaxSym;i++){ /* init sym */
    sym.k[i]=0;  /* phase factor in units of 2pi/l */
    sym.w[i]=+1; /* sign = exp(i*2pi*k/l) */
    sym.l[i]=1;  /* possible phase quantum numbers old2015: 2 */
    for (j=0;j<NN;j++) sym.i[i][j]=j; /* no permutation P=1 */
  }

  /* check bond data for errors */
  for (i=0;i<2*nw;i++) if(iw[i>>1][i&1]>=nn) F1("site>=NN iw[%u]",i>>1);
  /* initialize phase parameters */
  for (i=0;i<nw;i++){  ww.ex[i]=1;
    if (iw[i][0]>iw[i][1]){		/* lowest index first */ 
      j=iw[i][0]; iw[i][0]=iw[i][1]; iw[i][1]=j;
      ww.si[i]=-ww.si[i]; ww.ex[i]=-1;
    }
  }

  MPI_Barrier(MPI_COMM_WORLD);
  /* relax nfs, use cache 2016-03-16 kautz 5000 tasks may cause OOM */
  if (mpi_i) usleep(mpi_i/512*1000+10000);
  
  in1=fopen(name_file_i,"rt"); 
  /* fopen may fail after edit nfs-file on another node 2016-01 t100 */
  if (in1==NULL) { 
    if (!mpi_i) E1("ropen \"%s\", sleep 8s + 2nd try",name_file_i);
    sleep(8); in1=fopen(name_file_i,"rt"); } /* try again */
  if(in1==NULL) {
    if (!mpi_i) E1("ropen \"%s\", example doc/daten.i",name_file_i);
    MPI_Abort(MPI_COMM_WORLD,1);
  }

  /* ini should be called after (potential) n change, JJ,tU n=const */
  /*  n may be changed after loading new model.def */
  /* pascal triangle = binomial coefficent initialization */
  binomial_ini(nn);
  
  sprintf(FILE_l1, TMPDIR "l1_%04d.dat", mpi_i); /* initialize the name */

  // ToDo: simpler syntax: [a-z]*[0-9]*= for setting parameters
  // ToDo:                 .[a-z]*[0-9]* for procedures
  for(;;){  /* load and execute input file */
    if (!in1) {
      if((in1=fopen(name_file_i,"rt"))==NULL) { /* was closed for better C/R */
        if (!mpi_i)
          printf("\nERROR: reopen \"%s\" failed\n",name_file_i);
        sleep(1); MPI_Abort(MPI_COMM_WORLD,1);
      }
      fseek(in1,in1pos,SEEK_SET); /* continue on old position */
    }
    if (feof(in1)){ break; } /* removed2012-09: restore .i file */
    s1[0]=0; if (fgets(s1,500,in1)==NULL) break; /* 2012-05 +err-handling */
    in1line++; // infile line number for errors/debugging output
    /* read without \n ?? , .begin ... .end */
    if(s1[0]=='#' || s1[0]=='%' || s1[0]==0 || s1[0]=='\n'){ /* nop */ } else
    if(s1[0]=='!'){ // task 0 only exec commands
      // ToDo: reopen in1 to allow infile-manipulations?
      /* the tmp/pid check dont work on MPI + network_filesystems */
      // 2012-05 printf to fputs (less lib-overhead)
      if (!mpi_i) { 
        // ToDo: produce pid by echo $$PPID | tee spins.pid
        //   for sending signals like CKPT_HELP_REQ
        if (fputs(s1,stdout)<=0) E0("main.scan.fputs");
        if (system((char *)&s1[1])) E0("main.system(s1+1)");/* exec arg */}
    } else
    if(s1[0]==':'){ // task 0..3, use to debug task0!=task1 problems
      if (mpi_i<4) {  // srun -l + ":ps -o pid,spid,psr,pcpu,time,cmd -p $PPID -m"
        if (!mpi_i) if (fputs(s1,stdout)<=0) E0("main.scan.fputs");
        if (system((char *)&s1[1])) E0("main.system(s1+1)");/* exec arg */}
    } else
    if(!scan_int(s1,"max_ea",&max_ea)){max_ea=max(1,max_ea);} else
//    if(!scan_int(s1,"chkpt_mode",&chkpt_mode)){ } else
    if(!scan_int(s1,"nev",&nev)){nev=min(max(0,nev),NEV);} else
    if(!scan_int(s1,"ne0",&ne0)){ne0=min(max(0,ne0),NEW-1);} else
    if(!scan_int(s1,"pew",&pew)){pew=min(max(1,pew),NEW);} else
    if(!scan_int(s1,"startvec",&startvec)){} else
    if(!scan_int(s1,"startvec+",&i1)){ startvec+=i1; } else
    if(!scan_int(s1,"chkpt_load",&checkpoint)){ } else
    if(!scan_int(s1,"chkpt_time",&chkpt_time)){} else // 2012-05-22 checkpointing
    if(!scan_int(s1,"sym_lm",&sym.type_lm)){ sym.type_lm &= 1; } else // 2016-01
    if(!scan_int(s1,"sym_ud",&sym.wud)){ if(abs(sym.wud)>1)ERR(9999);
      // C: defined() expansion is undefinded, at GNU its 0 or 1
#if HFIELD || defined(CONFIG_TBC) || NOSZSYM
      if (!mpi_i)
        printf("\nError: sym_ud can't be used for field, TBC and ~Sz\n\n");
#endif
      sym.wud_saved= sym.wud; // 2018.05 if some bad code follows!?
    } else
    if(!scan_int(s1,"sym_k",NULL)){ /* in daten.k schreiben? */
      for(i=0;i<MaxSym;i++)sym.kgen[i]=0;
      scan_inta(s1,"sym_k",sym.kgen,MaxSym);
    } else
    if(scan_var(s1,"SL")){ /* set sublattice 0=A 1=B etc. */
      char *s2=scan_var(s1,"SL");
      printf(" set sublattice %d: ", sym.subl);
      for (i=0;i<nn;i++) {
        sym.sl[sym.subl][i]=((s2[i]&~0x20)-'A')%nn;
        printf("%c",'A'+sym.sl[sym.subl][i]);
      } printf("\n");
      /* old: scan_inta(s1,"SL",sym.sl[sym.subl],nn); */
      sym.subl++; sym.subl%=MaxNumSL; /* handle overflow as reset */
    } else
    /* removed2013: if(!scan_dbl(s1,"maxmem" ,&r1)){ maxmem=(ULong) r1;} else */
    if(!scan_dbl(s1,"maxmem", &r1)){ opt_maxmem=r1; } else /* Byte */
    if(!scan_dbl(s1,"maxfile",&r1)){ opt_maxfile=r1;  /* p.e. 2e9 = 2GB */
      hdsk_nzmax=      (ULong)(opt_maxfile*HBLen/Hsize);
      if( (double)hdsk_nzmax+1<opt_maxfile*HBLen/Hsize ) ERR(612); } else
    if(!scan_dbl(s1,"xy_anisotropy+",&r1)){ h_xypara+=r1;} else
    if(!scan_dbl(s1,"z_anisotropy+", &r1)){ h_zpara+=r1; } else
    if(!scan_dbl(s1,"xy_anisotropy", &r1)){ h_xypara=r1; } else
    if(!scan_dbl(s1,"z_anisotropy",  &r1)){ h_zpara=r1;  } else
    if(!scan_int(s1,"verbose",  &vvv)){} else
    if(!scan_int(s1,"sisj",&cfg_sisj)){} else
    if( (ss=scan_var(s1,"nud+"  ))!=NULL ){ scan_ud(ss,1);   } else
    if( (ss=scan_var(s1,"param+"))!=NULL ){ scan_para(ss,1); } else
    if( (ss=scan_var(s1,"nud"   ))!=NULL ){ scan_ud(ss,0);   } else
    if( (ss=scan_var(s1,"param" ))!=NULL ){ scan_para(ss,0); } else
    /* allow "p1=-1.0", "p1= -1.0" or " p1= -1.0" as input per line */
    if (sscanf(s1," p%d= %lf",&i1,&r1)==2) {
        para[i1]=r1; printf("set parameter p%d= %f\n",i1,r1); } else
 // if(s1[0]=='x'){ sscanf(s1,"%*c%d%lu",&x_p,&x_max); } else /* ToDo: replace by function xout= */
    if((ss=scan_var(s1,"xout"))!=NULL){ sscanf(ss,"%d%lu",&x_p,&x_max); } else
 // if(s1[0]=='l'){ loadvec(ev[0],s1+1,-1,-1); } else
    if((ss=scan_var(s1,"loadvec"))!=NULL){ loadvec(ev[0],ss,-1,-1); } else
    if(strncmp(s1,"newtime",7)==0){starttime=0l;itime();} else	/* reset clock */
    if(strncmp(s1,".out",4)==0){if(!mpi_i)outpara(stdout);} else
    if(strncmp(s1,"exit",4)==0){exit(0);} else /* 2015-09-08 for testing */
    if(strncmp(s1,"quit",4)==0){exit(0);} else /* 2015-09-08 for testing */
    // removed2012-10: pout could also be done by !echo "text"
    if(s1[0]=='a' && s1[1]>='0' &&  s1[1]<='9'){ /* add 0-9 check 2012-10 */
      /* start computation a=chooze_algorithm */
      sscanf(s1,"%*c%d%lu",&algo,&n1max);
      check_ud(); /* 2018-08-27 without nud in i-file, nu+nd may be out-of-range */
      if (!mpi_i) outpara(stdout);
      /* 2015-02-26: getfilepos + close + work(+C/R) + reopen in1 */
      in1pos= ftell(in1); fclose(in1); in1=NULL; /* closed for simpler C/R */
      
      MPI_Barrier(MPI_COMM_WORLD); /* debug */
      /* --- test and if neccessary create Sym symmetries P --- */
      symcreate(mpi_i);
      sym_bn_m_ini_flag=1;
      sym_bn_m_ini(0); /* 2015-09-22 was buggy for 2nd sym_k */

      MPI_Barrier(MPI_COMM_WORLD); /* debug */

      if (vvv&1) testsym();

      MPI_Barrier(MPI_COMM_WORLD);

      /* wud=(+1,-1, 0) to kud=(0,1,* old=-1), wud=exp(kud*I*Pi) */
      if (nu!=nd) { /* we will have simpler inner code, setting wud=0 */
        /* save + setting back if set nu=nd later 2018.05 */
        if (sym.wud) sym.wud_saved= sym.wud; 
        sym.wud=0; /* 2017-09 ud-sym can not be used, ToDo: setback */
      }
      /* setting wud back from saved state 2018.05 */
      if (nu==nd) sym.wud= sym.wud_saved;

      if (!mpi_i) { /* output on node 0 */
#if NOSZSYM
        printf("WARNING: for NOSZSYM=1 nu,nd are ignored\n");
#endif
        printf("\n nn,nu,nd %d %d %d   nw= %d"
              "  (number of sites, up/down-spins and bounds)\n",nn,nu,nd,nw);
        /* now only wud must be checked before sudconfig can be used */
        if (sym.wud) printf(" k_ud %d", 1-(sym.wud+1)/2);
                else printf(" k_ud *"); /* for the eyes */
        printf(" k_T ");
        for (i=0;i<sym.numsym;i++) if (sym.g[i]) printf("_%01d",sym.k[i]);
      } // node 0 output only
#if   CM(1,0,0)
      sym.type_lm=0; /* 1 not possible in JJ model, 2018-05 outside !mpi_i */
#else  /* 2016-01-16 LM-type 0=maxS(S+1)  1=maxJ(J+1) */
      if (sym.num_lm != nn && !mpi_i) printf(" type_lm %d", sym.type_lm);
#endif // CM(1,0,0)
      if (!mpi_i) { /* output on node 0 */
        /* ToDo: move output of parameters to a function? */
        printf("\n parameters ");
        //  for (j=0;j<nw;j++){ printf("t[%d]=%d ",j,ww.wt[j]); }
        for(i=1;i<numpara;i++){ int j,k=0;
          /* output human readable names, ex: J1=J2= 0.3 t1= -0.7 ... */
          for (j=0;j<nw;j++) if (ww.wj[j]==i) {printf("J%d=",(k=i)); break;}
          for (j=0;j<nw;j++) if (ww.wjz[j]==i) {printf("Jz%d=",(k=i)); break;}
          for (j=0;j<nw;j++) if (ww.wt[j]==i) {printf("t%d=",(k=i)); break;}
          for (j=0;j<nw;j++) if (ww.w[j]==i)  {printf("w%d=",(k=i)); break;}
          for (j=0;j<nw;j++) if (ww.wv[j]==i) {printf("V%d=",(k=i)); break;}
          for (j=0;j<nw;j++) if (ww.wh[j]==i) {printf("h%d=",(k=i)); break;}
          for (j=0;j<nn;j++) if (ww.we[j]==i) {printf("e%d=",(k=i)); break;}
          for (j=0;j<nn;j++) if (ww.wu[j]==i) {printf("U%d=",(k=i)); break;}
          for (j=0;j<nn;j++) if (ww.wfz[j]==i) {printf("fz%d=",(k=i)); break;}
          for (j=0;j<nn;j++) if (ww.wd[j]==i) {printf("D%d=",(k=i)); break;}
# if NOSZSYM  /* added 2018-01-09 */
          for (j=0;j<nn;j++) if (ww.wfx[j]==i) {printf("fx%d=",(k=i)); break;}
          for (j=0;j<nn;j++) if (ww.wfy[j]==i) {printf("fy%d=",(k=i)); break;}
          if (mpi_i==0 && j<nn && (VecType & 4)==0) E0("fy needs complex");
          for (j=0;j<nw;j++) if (ww.wjx[j]==i) {printf("Jx%d=",(k=i)); break;}
          for (j=0;j<nw;j++) if (ww.wjy[j]==i) {printf("Jy%d=",(k=i)); break;}
# endif
          if (!k) printf("p%d=",i); /* unused or unknown parameter */
          printf(" %g  ",(double)para[i]);
        } printf("\n");
        printf(" Bsize=%d (size of each basecfg, NN/8[Byte]), tbase[bits]=%d\n",
               (int)Bsize, (int)sizeof(tbase)*8);
        // printf(" Bsize=%d (size of each basecfg, NN/8[Byte]) t+=%s\n",(int)Bsize,sclock(~0l));
      } // task0 output only

      /* alloc thread local buffers (not during iterations (old)!)  2014-07-17 */
      if (pt_n > B_NUM) F0("pt_n > B_NUM");
      for (block=0; block<pt_n; block++) {
        /* alloc_hbuf depends from sym */
        hbuf[block] = alloc_hbuf(); /* better local? no, but 1st access */
        if (!hbuf[block]) {
          F2("ERROR: alloc_hbuf failed, mpi=%d thread=%d", mpi_i, block); 
          exit(1);// 2014-07 continue;
        }
        // ToDo: hbuf[block]->root = hbuf; // pointer to thbuf *(hbuf)[B_NUM]
        //    to access hbufs of other threads ??
        hbuf[block]->block = block;
        hbuf[block]->v0 = NULL;
        hbuf[block]->v1 = NULL;
      }

      /* early speed tests for better tuning: 2014-05-26 speed_mpi.gpl */
      // 2015-03-26 add pthread too see cache coherence effects
      // 2015-07-07 AHEAD=2K segfault (meggie, ulimit -s=10240, 10240000)
      // 2016-11-04 segfault if ulimit -s 18000 (stacksize) to small
      // 2016-12-20 hangs on LRZ.icpc B_NUM=128 -D CFG_CPUSET=0xffffffff -t32
      // ToDo replace by omp-version? gcc only 600% (no pinning, t100?)
      for(block=0; block<pt_n; block++)
              /* hbuf[] + .block is be initialized above */
              parallel_start(block, speed_tests, hbuf[block]);
      for(block=0; block<pt_n; block++) parallel_wait(block);
      /* ompi will hang on MPI_Barrier here on oversubscription above 16 */
      /*  also using --mca mpi_yield_when_idle 1 --mca mpi_preconnect_mpi 1 */
      /* MPI_Barrier(MPI_COMM_WORLD); */ /* removed for oversubscr. 2016-03 */
      /* ri2016 tri96+4 4c: 64.bnV4=510ns SH=0.02m 128.bn4=1000ns SH=0.03m
                           256.bnV4=1850ns SH=crash 255,257=ok 256=crash
                   98+2 4c: 8-9=ok 10,16,32=crash 10:n1=51=8*6+1*3+1*0 */
      
      /* ===================== new-method l1+H storage ================ */
      /* generate hilbert space+matrix by applying hilbert Op
       *  on single configs iterative, ToDo: check speed
       *  ToDo: may be better with hashed_l1[]
       */	
      if (algo&8) {  // ToDo: needs to review this code (buggy for sure)
        if (mpi_n!=1 || pt_n!=1) { 
          printf("WARNING: (mpi_n!=1 || pt_n!=1) is untested, please report\n");
        } /* not tested, ToDo */
        n1=(n1max)?n1max:100000;	/* n1max<=n1 variiere Basis! */
        if (((XY_TYPE)(n1/(mpi_n*pt_n)))!=n1/(mpi_n*pt_n)) {
          printf(" block_n1 dont fit to XY_TYPE, see h_file.h\n");
          ERR(9999);
        }
        for (n1_bits=0;n1_bits<(int)sizeof(ULong)*8;n1_bits++) 
           if (((ULong)1)<<n1_bits > n1-1) break;
        b_ini(n1,vvv);          /* init b_len[], node_n1 (algo&8) */
        if((algo&7)<2){
          /* malloc by thread 0, first touched by working threads for ccNUMA */
          if ((!mpi_i) && (!v0)) printf(" a8.malloc v0,v1 ...\n");
          v0=(scplx *)mymalloc(b_ofs[pt_n]*Csize,&m_v0);
          v1=(scplx *)mymalloc(b_ofs[pt_n]*Csize,&m_v1);
          if ((!mpi_i) && ((!v0) || (!v1)))
            F0(" malloc failed, consider mmap?");
          touchvec(v0); // 2015-03-09 touch before storeH eats local memory
          touchvec(v1);
        }
        {
          if (1.*(node_n1*(off_t)Bsize)!=node_n1*1.*Bsize)
            printf("\nERROR mymap(n1*Bsize): off_t overflow");
          fl1=fopen(FILE_l1,"r+b");
          if (!fl1) fl1=fopen(FILE_l1,"w+b");
          if (!fl1) F1("fopen %s",FILE_l1);
          l1 = (unsigned char *)
                mymap(FILE_l1, fl1, node_n1 *(off_t)Bsize,
                                    node_ofs*(off_t)Bsize);
          fclose(fl1); fl1=NULL;  /* 2015-02-24 close for BLCR, mmap ok? */
          if (node_n1 && !l1) {
            E1("mymap(l1) failed for mpi=%d, skipping", mpi_i);
#ifdef CONFIG_MPI
            if (mpi_n>1) MPI_Abort(MPI_COMM_WORLD,1);
#endif
            continue;
          }
        }
        for(i=0;i<nev;i++){
          int vvv2=vvv; if (pt_n*mpi_n>4) vvv&=~2; /* avoid to much output */
          if ((!mpi_i) && (!ev[i])) printf(" malloc ev%d", i);
          ev[i]=(scplx *)mymalloc(b_ofs[pt_n]*Csize,&m_ev[i]);
          vvv=vvv2;
        }
// sprintf(m_n2.name, TMPDIR "n2_%04d.map", mpi_i);
// norm2=(ATnorm2 *)mymalloc(node_n1*sizeof(ATnorm2),&m_n2);base2ok=0;
        l2b  =  (ULong *)mymalloc(node_n1*Lsize,&m_l2); /* hilfsarray2 */
        n1max=node_n1;       /* save memory block len before n1 changed */
        if (mpi_n!=1 && !mpi_i) E0("storeh2 will fail for mpi");
        storeh2(0, hbuf[0]);	/* ??? provisorium - testphase */
        for (n1_bits=0;n1_bits<(int)sizeof(ULong)*8;n1_bits++) 
           if (((ULong)1)<<n1_bits > n1-1) break;
        /* l2b=(ULong *)myfree((void *) l2b,&m_l2); / * zur Zeit noch!!! */
        /* (algo&8)!=0 storeh2 */
        randvec(v1,startvec,0);itr=inc1(nev,ne0,v0,v1,hbuf); /* EW(T),EV(T)  */
        randvec(v1,startvec,0);incv1(nev,v0,v1,ev,itr,hbuf); /* EV(H)   */
        /* repeat computation with new start vector */
        if(n1==n1max && nev>0)F0("Fatal, should not break here");
        if (nev) for(j=0;j<nev;j++){
          if (!mpi_i) printf("\n\nstart chk3.ev%d ...", j);
          chk3(ev[j],v1,0,hbuf); /* (inp_vec, work_vec, ?, workbuf) */
        }
        /* ToDo: reduce to / replace by better code structure:
         *   if (algo&8) { alloc } else { numsymconf + alloc }
         *   ...
         *   if (algo&8) { storeh2 } else { storeh }
         */
      } /* (algo&8)!=0  ====== experimental storeh2 (one-step-space+matrix) */
      if ((algo&8)==0){	/* ============ old-good-ns+storeh-method ========= */
        /* generating Hilbert space, and Hilbert matrix in two steps */

        // MPI_Barrier(MPI_COMM_WORLD);

        /* 2015-12-30 l1-generation only for high S, one task recursive */
        if((!mpi_i) && (algo&64)) printf("\nDBG: algo&64 stop after l1 ...");
        n1=(ULong)numsymconf();  /* slow, calculate n1, write l1 */
        /* ToDo17: store to memory or disk like matrix STOREH */
        /* ToDo: use l1 as argument, not global var, better code reading */
//        if (opt_perf_test) { /* assuming speed/n1 is constant */
//          ULong perf_n1=opt_perf_test; /* 1e6 = about size of sq40 9+31 */
//          if (2*NUM_AHEAD_LINES > perf_n1) perf_n1= 2*NUM_AHEAD_LINES;
//          if (n1<perf_n1) perf_n1=n1; /* not reduced */
//          if (n1!=perf_n1 && !mpi_i)
//            printf("\nWARNING: n1 reduced to %lu for perf_test"
//                  " (check speed/n1)\n\n", perf_n1);
//          n1=perf_n1; /* 1st idea, see if it works */
//        }
        // checkpoint2 old=1
        if((!mpi_i) && (algo&64))
          printf("\nDBG: algo&64 stop, use --ckpt_load=2 to continue.");
        if (algo&64) MPI_Abort(MPI_COMM_WORLD,0);
        
        for (n1_bits=0;n1_bits<(int)sizeof(ULong)*8;n1_bits++) 
           if (((ULong)1)<<n1_bits > n1-1) break;
        
        if (!n1) { if (!mpi_i) E0("n1==0"); checkpoint=0; continue; }
        if (!mpi_i) {
          printf("\n N,nu,nd= %u  %u %u  n1= %lu  n1_bits= %d",
                  nn, nu, nd, n1, n1_bits); /* n1_bits */
          /* Tru64@alpha: log2(0) segfault, 2017-01 log2 replaced */
          /* (long+sdouble*2)*n1/nodes, Dsize=sizeof(sdouble) */
          /* 2017-01 more info for large scale problems, 1000++ nodes */
          /*  how much nodes are needed on mpi-cluster? mem_needed/nodes_mem
           *  ... but use big enough buffers to saturate network
           *  ... there will be an optimum in buffersize ? ca. 10% of total mem?
           *  more nodes will cost more I/O overhead, so use min num nodes
           */
          printf("\nMEM: vector size = n1*%5dB = %.2e (all nodes, needed)",
              (int)(Bsize+Dsize*(2+nev)), (Bsize+Dsize*(2+nev))*1.*n1);
          printf("\nMEM: matrix size = n1*nzx*%dB = %.2e (all nodes, optional)",
              (int)((Hsize+HBLen-1)/HBLen), n1*1.*NZXMAX*Hsize/HBLen);
              /* 2017.01.22 size for { rr(2B),yy(4B),by(2B) } ? */
          printf("\nMEM: ... assuming nzx=NZXMAX=%d,"
                 " real value nzx will be lower", NZXMAX); /* 2017-01-22 */
            /* call storeh with mpi_n=1 with rnd-cfg, to get better nzx? */
            /* ToDo: or set global nzx_estimated? or ising-erg? */
            /* HELP: for multiple runs, think on nzx depends from nu+nd */
          printf("\nMEM:  PT-buf =  pt_n*AH*nzx  = %.2e (alloc per MPI)",
              pt_n*1.*sizeof(thbuf) ); /* 2017-02-05 */
          printf("\nMEM: MPI-buf = B_NUM*AH*nzx  = %.2e (per node, * 4..8B)",
              B_NUM*NUM_AHEAD_LINES*1.*NZXMAX); /* 2017-01-22 { rr(4-8B),yy(4B) } ? */
          printf("\nMEM: ... recommend num_tasks*1MB = "
                 "%d MB/node for buffers",mpi_n); /* AH*B_NUM propto mpi_n */
          printf("\nMEM: ... but max. 10%%*PhysMem, please check \"size ./spin\"");
          printf("\nMEM: MPI-A2a-table size = n1/AH*mpi_n*4B = %.2e (per node, minimize)",
              n1*1./NUM_AHEAD_LINES*mpi_n*4); /* 2017-01-22 */
          printf("\nMEM: ... AH=%d should be multiple of mpi_n=%d"
             " (ToDo2017-01 mpi-speed)",NUM_AHEAD_LINES,mpi_n);
          printf("\nPRF: mean_mpi_packet_size: n1*nzx/AH/mpi_n=%lu ",
             n1*NZXMAX/NUM_AHEAD_LINES/mpi_n);
          /* 2017-01-22 */
          printf("\nMEM: maxmem  = % .2e (Bytes per task, -1.0=auto)",
                  opt_maxmem); /* hmem_nzmax*8B */
          printf("\nDSK: maxfile = % .2e (Bytes per task,  0.0=no_disk)",
                  opt_maxfile); /* hdsk_nzmax*8B */
          /* ToDo2017: maxfile should also be per node (write bigger blocks) */
          printf("\n");
        }
        /* we need a bit for the XY_FLAG, ToDo: change for MPI */
        if (((n1<<1)>>1)!=n1) { if (!mpi_i) E0("n1>max"); continue; }
        if (((XY_TYPE)(n1/mpi_n))!=n1/mpi_n) {
          printf(" node_n1 dont fit to XY_TYPE, see h_file.h\n");
          ERR(9999);
        }

        MPI_Barrier(MPI_COMM_WORLD);
        b_ini(n1,vvv);	/* block initialization b_len[], b_ofs[] node_n1 ... */
        /* 2017-02-15 prepare new matrix storage in AH-blks */
        /* blks*AH round off to multiple of AH */
        /* is node_n1 same on all nodes? smaller on last node? mpimax later */
  /* ToDo17 better round up node_n1 to multiple pt_n*AH from beginning!? */
        node_ah_blks= (node_n1+NUM_AHEAD_LINES-1)/NUM_AHEAD_LINES;
        /* round up to multiple of pt_n (+mpi_max) */
        node_ah_blks= ((node_ah_blks+pt_n-1)/pt_n)*pt_n;
        if (((XY_TYPE)(node_n1))!=node_n1   // 2017-02-20 no unidentified bugs
            || (node_n1>>(sizeof(XY_TYPE)*8-1))>1) {
          printf(" node_n1 dont fit to XY_TYPE, see h_file.h\n");
          ERR(9999);
        }
#if CONFIG_MPI
        { ULong t1=node_ah_blks; MPI_Allreduce(&t1, &node_ah_blks, 1,
              MPI_UNSIGNED_LONG, MPI_MAX, MPI_COMM_WORLD); }
#endif
        ah_nz=(int *)realloc((void *)ah_nz, node_ah_blks*sizeof(int));
        if(!ah_nz) F0("malloc.ah_nz failed");
        {ULong i;for(i=0;i<node_ah_blks;i++)ah_nz[i]=0;} /* 16.02.17 */
         /* must be and is round up to multiple of pt_n (see above) */
        if(!mpi_i) printf("DBG.node_ah_blks= %6lu =%d*%6lu (mpi_max)\n",
          node_ah_blks, pt_n, node_ah_blks/pt_n);

        /* --- l1 is now initialized by numsymconf --- */
#ifndef CFG_STREAMS
#define CFG_STREAMS 1
#endif
        if (!l1) {
          double tm1, tm2;
          int opt_streams=1;
          opt_streams= ((mpi_n<(CFG_STREAMS))?mpi_n:(CFG_STREAMS));
          tm1=fstime();
          if (!mpi_i) printf(" mymap(%s + node_ofs to *l1)\n",FILE_l10);
          /* Oct07: every node maps only its part of l1[]!
           * may be this is not necessary because only accessed 
           * parts loaded into mem? but if mmap failes, its loaded!
           */
          if (1.*(node_n1*(off_t)Bsize)!=node_n1*1.*Bsize)
            printf("ERROR mymap(n1*Bsize): off_t overflow\n");
// ToDo: may fail on cached NFS and fast write_task0/read_task3 cycle
// void *mymap(const char *nam, FILE *f1, off_t len, off_t ofs)
// 5700*parallel read is bad idea via nfs,
//     zuletzt kautz 4.3.2016 nfs-node-OOM nach 8Tagen alte version 2015-09
// ToDo17: but now its IB-bottleneck (upto config_num_io=8 inputstreams?)
// ToDo17: MPI_File_Iread? MPI-3? fallback=seq-read+MPI-send
#ifdef CONFIG_MPI  // until 2016-01-20 token was nearly useless, serialized
          if (mpi_i%(mpi_n/opt_streams)>0) { int token=0; MPI_Status status;
            MPI_Recv(&token,1,MPI_INT,mpi_i-1,1,MPI_COMM_WORLD,&status);
            if (0==(mpi_i&(mpi_i-1))) printf(" l1_load.node=%d\n",mpi_i); }
#endif
          fl1=fopen(FILE_l10,"r+b"); /* 2016-01 moved behind token for nfs */
          if (!fl1) { F0("main.fopen_l1"); }
          l1 = (unsigned char *)
               mymap(FILE_l10, fl1, node_n1 *(off_t)Bsize,
                                    node_ofs*(off_t)Bsize);
          fclose(fl1); fl1=NULL;  /* 2015-02-24 close for BLCR, mmap ok? */
#ifdef CONFIG_MPI
          if (mpi_i<mpi_n-1 && mpi_i%(mpi_n/opt_streams)<mpi_n/opt_streams) {
            int token=0; MPI_Send(&token,1,MPI_INT,mpi_i+1,1,MPI_COMM_WORLD); }
#endif
          // ToDo: malloc + load via task0 serial read
          if (node_n1 && !l1){ E1("mymap(l1) %2d failed, skipping", mpi_i);
            if (mpi_n>1) MPI_Abort(MPI_COMM_WORLD,1);
            else         continue; }
          tm2=fstime()-tm1;
          if (mpi_i==mpi_n-1)
            printf("PRF.load_l1 t[min]= %.2f  streams=%d\n",
               tm2/60., opt_streams);
        } /* if(!l1) */
// ToDo18: all cfgs zero after load, maybe check here first 2 states n1>1

        MPI_Barrier(MPI_COMM_WORLD);
        /* ToDo: add vec_malloc[_block](tvec,m_v0) to vector.h */
        if((algo&7)<2){ /* we need this 2 vectors for Lanczos algorithm */
          int localerror=0, sumerror=0;
          /* ToDo: alloc within threads or remove thread-blocks */
          if ((!mpi_i) && (!v0))
            printf(" a0.malloc %s%3s(%dB) v0,v1,%d*ev a %8.2eB/mpitask\n",
                   ((VecType&4)?"c":" "), ((VecType&1)?"dbl":"flt"),
                   (int)Csize, nev, 1.*b_ofs[pt_n] * (int)Csize);
          /* JS2012-10 for mpi_n > n1  b_ofs */
          // 2015-03-05 touching memory to avoid vm overbooking by storeH
          v0=(scplx *)mymalloc(b_ofs[pt_n]*Csize,&m_v0); touchvec(v0);
          v1=(scplx *)mymalloc(b_ofs[pt_n]*Csize,&m_v1); touchvec(v1);
          // ToDo: mpi_sum num_errors 0=ok, 1...mpi_n = num failes
          // compute log2 group of nodenumber or simply sum of errors
          // 2012-11-01: no errors if n1 < mpi_n (b_ofs==0)
          if (b_ofs[pt_n] && ((!v0) || (!v1))) sumerror=localerror=1;
#ifdef CONFIG_MPI
          MPI_Allreduce(&localerror, &sumerror, 1,
                        MPI_INT, MPI_SUM, MPI_COMM_WORLD);
#endif
          if ((!mpi_i) && sumerror)
            F2(" malloc failed on %d of %d nodes", sumerror, mpi_n);
          // if (sumerror) continue;
          // test 8MB allocatable for MPI 32?
        }
        MPI_Barrier(MPI_COMM_WORLD);
        if(((algo&7)==2) && nev<1){ nev=1; E0("nev is 0, set 1"); }
        for(i=0;i<nev;i++){
          int vvv2=vvv; if (mpi_i>3) vvv&=~2; /* avoid to much output */
          if ((!mpi_i) && (!ev[i]) && (vvv&128)) printf(" malloc ev%d", i);
          ev[i]=(scplx *)mymalloc(b_ofs[pt_n]*Csize,&m_ev[i]);
          touchvec(ev[i]);
          vvv=vvv2;
        }
// sprintf(m_n2.name, TMPDIR "n2_%04d.map", mpi_i);
// norm2=(ATnorm2 *)mymalloc(node_n1*sizeof(ATnorm2),&m_n2);base2ok=0;
        /* if( checkpoint < 3 ) */ 
        { /* norm2 is not mapped! change it ??? */
          double tm1=fstime();
          ini_norm2();
          if (!mpi_i) printf("PRF: n2.t[m]= %6.2f", (fstime()-tm1)/60.);
        } /* else load_l1(l1,n1); --- mmap used */
        MPI_Barrier(MPI_COMM_WORLD);
        outbasis(stdout,hbuf[0]);

#ifdef CFG_DBG_BadMemEmulation  /* 2017-04-05 testing MPT debug output  */
        /* simulate memory bitflips to see error handling 2017-04 */
        printf("\n\nWARNING: !!! CFG_DBG_BadMemEmulation enabled !!!\n\n");
        ((unsigned char *)l1)[node_n1 *(off_t)Bsize/2] ^= 4;
        ((unsigned char *)l1)[node_n1 *(off_t)Bsize/4] ^= 8;
        ((unsigned char *)l1)[node_n1 *(off_t)Bsize/2+3] ^= 9;
#endif
        hdsk_nzmax= ((ULong)(opt_maxfile*HBLen/Hsize)); // 2017-01-28
/* 2015-04-08 malloc Hbuffer (rest of memory), fail means its slower */
/* ReserveMB need for libMPI
  # sc072.asgard extrapolated to kautz-mpich2: 972*6/18*64KB=20.7MB 5+49
  srun -p sca -n 72 bash -c "(ulimit -v 23232;./spin)"
  srun -p sca -n 55 bash -c "(ulimit -v 23232;./spin)"  # +64K/18tasks
  srun -p sca -n 54 bash -c "(ulimit -v 23168;./spin)"
  srun -p sca -n 37 bash -c "(ulimit -v 23168;./spin)"  # +64K/18tasks
  srun -p sca -n 36 bash -c "(ulimit -v 23104;./spin)"
  srun -p sca -n 19 bash -c "(ulimit -v 23104;./spin)"  # +64K/18tasks
  srun -p sca -n 18 bash -c "(ulimit -v 23040;./spin)"
  srun -p sca -n  1 bash -c "(ulimit -v 23040;./spin)"
  # comp2.ompi-1.6.5-ft reg-mem-warning ???
  mpiexec -np $((4* 8)) bash -c "(ulimit -v 185000;./spin)" # OK +4M/16 +MPIw
  mpiexec -np $((4* 4)) bash -c "(ulimit -v 181000;./spin)" # OK +4M/8 +MPIwarn
  mpiexec -np $((4* 2)) bash -c "(ulimit -v 177000;./spin)" # OK +2M/4 +MPIwarn
  mpiexec -np $((4* 1)) bash -c "(ulimit -v 175000;./spin)" # OK + reg mem limit low
  # comp2.ompi-1.8.4 8+32 2015-04
  mpiexec -np $((6*16)) bash -c "(ulimit -v 255000;./spin)" # OK +138MB
  mpiexec -np $((5*16)) bash -c "(ulimit -v 259000;./spin)" # OK +142MB
  mpiexec -np $((4*16)) bash -c "(ulimit -v 259000;./spin)" # OK +142MB
  mpiexec -np $((3*16)) bash -c "(ulimit -v 254000;./spin)" # OK +137MB
  mpiexec -np $((2*16)) bash -c "(ulimit -v 254000;./spin)" # OK +137MB
  mpiexec -np $((1*16)) bash -c "(ulimit -v 254000;./spin)" # OK 
  mpiexec -np $((1* 2)) bash -c "(ulimit -v 256000;./spin)" # OK +139MB
  mpiexec -np $((6* 1)) bash -c "(ulimit -v 121000;./spin)" # OK +4MB
  mpiexec -np $((4* 1)) bash -c "(ulimit -v 122000;./spin)" # OK 
  mpiexec -np $((3* 1)) bash -c "(ulimit -v 122000;./spin)" # OK 
  mpiexec -np $((2* 1)) bash -c "(ulimit -v 123000;./spin)" # OK +6MB
  mpiexec -np $((1* 1)) bash -c "(ulimit -v 117000;./spin)" # OK 
  bash -c "(ulimit -v 117000;./spin)" # OK (but mpicc)
  # comp1.ompi-1.8.4 5+49
  mpiexec -np 8 bash -c "(ulimit -v 245000;./spin)" # OK +4M/8
  mpiexec -np 7 bash -c "(ulimit -v 241000;./spin)" # OK +4M
  mpiexec -np 6 bash -c "(ulimit -v 237000;./spin)" # OK +4M
  mpiexec -np 5 bash -c "(ulimit -v 233000;./spin)" # OK +3M
  mpiexec -np 4 bash -c "(ulimit -v 230000;./spin)" # OK +4M
  mpiexec -np 3 bash -c "(ulimit -v 226000;./spin)" # OK +2M
  mpiexec -np 2 bash -c "(ulimit -v 223355;./spin)" # OK +156M
  mpiexec -np 2 bash -c "(ulimit -v 223354;./spin)" # hanging SH!
  mpiexec -np 1 bash -c "(ulimit -v  67464;./spin)"
  (ulimit -v 84000;./spin -t8) # ca. 8MB/task
  (ulimit -v 67000;./spin -t6)
  (ulimit -v 51000;./spin -t4)
  (ulimit -v 34000;./spin -t2)
  (ulimit -v 26000;./spin -t1)
 */     /*  sub_ini() must be called before! */
        if ((algo&7)<3) /* 2016-11-10 malign alloc big Hbuf for fulldiag */
        { size_t memHbuf= ((size_t)opt_maxmem); /* 2015-05-11 32bit */
          /* ToDo2016: replace NZXMAX by real nzxmax as better approx */
          /* estimate upper border per Core, do not alloc more than that */
          size_t maxHbuf= (n1/(mpi_n*pt_n)+1)*NZXMAX*(size_t)Hsize/HBLen*120/100
                          +2*sizeof(struct thxy); 
          /* 2017-01-12 override maxHbuf by opt_maxmem (if we want no SH) */
 // ToDo some disk + memory 2017
          /* ToDo: do a lower estimation of NZXMAX */
          /* ToDo2016: check size_t overflow! 32bit.max=16G */
          int lok=0, gok=0; char *reserve;
          /* biggest we have is 512GB/32=16GB per core comp2_2015.mpi=32 */
          if (opt_maxmem<0.0) /* autoset */ memHbuf = maxHbuf;
          if (opt_maxmem>=0. && maxHbuf>memHbuf) maxHbuf= memHbuf; // 2017-01-28
//   if (memHbuf>maxHbuf) memHbuf= maxHbuf; // 2017-01-28 set above need
          if (opt_maxmem>0. &&  opt_maxmem > 1.000001*memHbuf && !mpi_i)
              E0("maxmem does not fit to size_t");
          if (!mpi_i) getmem(); /* 2018-11 output vm rss shr */
          if (!mpi_i){
if (mpi_n>1) {
  printf("\nWARNING: spinpack may crash here, if libmpi causes OutOfMemory");
  if (pt_n<B_NUM) printf(
  "\nWARNING: ... but B_NUM=CFG_PTHREAD is set bigger than needed,"
            " please reduce!");
  else printf(
  "\nWARNING: ... if so minimize NZXMAX, reduce NUM_AHEAD_LINES or maxmem");
   // old: maxfile
}
#ifdef ReserveMB
  printf("\nDBG: ReserveMB= %d (ompi chkpt, not eaten by SH)", ReserveMB); 
#if defined(OMPI_MAJOR_VERSION) /* 2017-01 do not see effects */
            if (mpi_n>1) /* check using !ps -o pid,psr,vsz,rss,cmd -p $PPID */
              printf("\nToDo: try mpiexec --mca mpi_preconnect_mpi 1");
#endif
#endif
            printf("\nDBG: maxfile[B/core]=maxmem*%d= %8.2e autoMax= %8.2e",
                   (int)(Hsize/HBLen), 1.*memHbuf, 1.*maxHbuf);
          }
          /* better let some free memory for mpich xmalloc calls! + min16MB
           *   ompi-1.4 + pthread? failed for free=16MB
           *  allocating a reserve and free it for mpi-use does not
           *    seems to help in every case (bugs 2015-04-06), but no choice
           *   hopefully it crashs at start and not later as in v2.49
           * ToDo: replace NZXMAX by nzxmax (computed on runtime)
           */
          gok=lok= 1; /* gok: global num threads with reserve = OK */
#ifdef ReserveMB
          reserve=(char *) malloc(ReserveMB<<20); /* MPI reserve memory */
          /* if no MPI, reduce reserve for stack + functions */
          gok=lok= (reserve) ? 1 : 0; /* thread-global and local OK */
#endif
#ifdef CONFIG_MPI
          MPI_Allreduce(&lok, &gok, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
#endif
          if (gok<mpi_n && !mpi_i) {
            printf("\nWARN: OOM %d no reserve %dMB ...", mpi_n-gok, ReserveMB);
#ifdef CONFIG_MPI
            /* 2015-04-09 8MB was not enough ompi-1.4 */
            printf("\nWARN: ... on OOM MPI_Sendrecv may hang later ...");
#endif
          }
          if (reserve && !mpi_i)
            printf("\nDBG: reserve=OK, try Hbuf/thread (min(1/16...1)* %.2e",
                 1.*maxHbuf);
          if ((!reserve) && !mpi_i) printf("\nDBG: reserve=FAIL");
          gok=0; /* reused to count mpi_n*pt_n Hbuffer-mallocs */
          if (reserve) /* no Hbuffer without reserveMB
[odin:05295] [ 0] /lib64/libpthread.so.0 [0x35e260ebe0] # CentOS-5.8
[odin:05295] [ 1] /usr/lib64/openmpi/1.4-gcc/lib/libopen-pal.so.0(\
                 opal_memory_ptmalloc2_int_malloc+0xb21) [0x35e2236f31]
web:   mpirun -x OMPI_MCA_memory_ptmalloc2_disable=1 ...   # hilft! or ...
       OMPI_MCA_memory_ptmalloc2_disable=1 mpirun  ...  # set in /etc/bashrc
        bad ptmalloc2 part of glibc? 
      https://www.open-mpi.org/community/lists/users/2012/09/20249.php */
          while (memHbuf > maxHbuf/16
             || (opt_maxmem && memHbuf>opt_maxmem/8) || memHbuf>400e6){
            if (!mpi_i)
              printf("\nDBG: try alloc %4d*%2d* %8.2e Byte for Hbuffer",
                    mpi_n, pt_n, 1.*memHbuf);
            /* ToDo2016: for OMP better malloc(pt_n*memHbuf) +local_access */
            for (lok=0,i=0;i<pt_n;i++) {
              if (memH[i].ptr!=NULL || memH[i].len) ERR(9999); /* 2018-11 */
              memH[i].ptr= (char *) malloc(memHbuf); // malloc per thread
              if (memH[i].ptr) lok++;
            } /* MPI_BAND = bitwise and */
            gok=lok; /* number of MPI-threads, where malloc succeeded */
#ifdef CONFIG_MPI
            MPI_Allreduce(&lok, &gok, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
            // ToDo18 better use MPI_Alloc_mem() ?
            //    but unclear if all tasks get error if one task fails
            //    == automatic error reduce OR/SUM
#endif
            if (gok==mpi_n*pt_n && !mpi_i) printf(" = OK");
            if (gok==mpi_n*pt_n) break; /* fine, leave the loop here */
            /* malloc all or nothing! */
            if (gok!=mpi_n*pt_n && !mpi_i) 
              printf(" = FAIL %d of %d", mpi_n*pt_n-gok, mpi_n*pt_n);
            for (i=0;i<pt_n;i++) {
              if (memH[i].ptr) {
                free(memH[i].ptr);
                memH[i].ptr=NULL;
                memH[i].len=0; }
            }
            // 2015-06 better next try (ubound+lbound)/2 in case of gok>0
            if (gok>0) {
              memHbuf = ((2*gok+1)*memHbuf/(pt_n*mpi_n))/2; // +0.5
              if (memHbuf<maxHbuf) maxHbuf = memHbuf; // shift lower bond
              /* slowdown on mpi-systems where one node has less free mem */
            }
            /* remove lower 1 bits, binary rounding */
            else {
              while ((memHbuf&(memHbuf-1))>memHbuf-memHbuf/4)
                      memHbuf &= (memHbuf-1);
              if (memHbuf&(memHbuf-1)) memHbuf&=(memHbuf-1);
              else                     memHbuf>>=1;  /* -25%...-50% */
            }
            if (!mpi_i) 
              printf(", next= %.2e", 1.*memHbuf);
          } /* end_while try to alloc rest of mem for Hbuffer */
          if (gok==mpi_n*pt_n) for (i=0;i<pt_n;i++) memH[i].len= memHbuf;
          hmem_nzmax= memHbuf/(Hsize/HBLen); /* ToDo17 ??? */
          if (gok==mpi_n*pt_n && vvv && !mpi_i)
            printf("\nDBG: allocated %4d*%2d* %8.2e Byte for Hbuffer",
                 mpi_n, pt_n, 1.*memHbuf);
          if (reserve && gok!=mpi_n*pt_n && vvv && !mpi_i)
            printf("\nDBG: alloc %4d*%2d* %8.2e Byte Hbuffer failed",
                   mpi_n, pt_n, 1.*memHbuf);
          if (reserve) free(reserve);
        } // local memHbuf=f(opt_maxmem)

        {int i;for(i=0;i<16*16;i++) mpi_xy_traffic[i]=0; }
        
        /* ToDo: free Hbuffer after computation?, error handling statH_out */
        if ((algo&7)<3) { storeh(hbuf); } /* compute H + store to mem/disk */
        ERR(-1);	/* reset error + print old errors */
        
        /* to see difference option packed_nz */
        if (mpi_n>1 && (mpi_n<=8 || (vvv&4))) { /* see above, not computed else */
          int i,j,n=((mpi_n>16)?16:mpi_n); ULong sum[16*16];
          MPI_Allreduce(mpi_xy_traffic, sum, 256,
                        MPI_UNSIGNED_LONG, MPI_SUM, MPI_COMM_WORLD);
          if (!mpi_i) for(i=0;i<n*n;i++) {
              if (i%n==0) printf("\nMPIsz %2d:",i/n);
              if (n<=8){ if (sum[i]<1e8) printf(" %7lu",sum[i]);
                         else        printf(" %7.1e",1.*sum[i]); } // 2018-05
              else { for(j=0;j<64 && (sum[i]>>j);j++) {}; /* j=log(sum[i]) */
                     printf(" %2d",j); } // 2Byte_log(y) statt 7Byte_y
            } }
        MPI_Barrier(MPI_COMM_WORLD);
        test_all_sym();	/* only array-ov test Mai99 */
#if 0
        if((algo&7)<2){ /* ToDo: remove! its twice!? */
          if ((!mpi_i) && (!v0)) printf("\n malloc v0,v1");
          /* ToDo: alloc within threads or remove thread-blocks */
          if(!v0) v0=(scplx *)mymalloc(b_ofs[pt_n]*Csize,&m_v0);
          if(!v1) v1=(scplx *)mymalloc(b_ofs[pt_n]*Csize,&m_v1);
          if ((!mpi_i) && ((!v0) || (!v1))) E0(" malloc failed");
        }
        MPI_Barrier(MPI_COMM_WORLD);
#endif
        /* degeneracy loop for different startvectors (2016-04: +FTLM) */
        for(iv=1;iv<=max_ea;iv++){ /* iv = index number of EA start vector */
          if (err_fatal) break;   /* stop on fatal errors */
          if (iv>1) checkpoint=5; /* reset checkpoint old=3 */
          if((algo&7)==0){ /* a0 = Lanczos tridiag */
            /* pseudo random startvector depends only on iv! */
            /* here is the right place to choose iv2=random instead of iv */
            randvec(v1,startvec,iv-1+(startvec&8)*nruns); /* bug? svec&7? */
            itr=inc1(nev,ne0,v0,v1,hbuf);               /* EW(T),EV(T)  */
#ifdef CFG_FTLM /* max_ea is used to get different startvectors */
            if (!mpi_i) printf("\nFTLM: next round %d/%d",iv,max_ea);
            /* do not slow recompute H, ToDo: better add dirtyH=0 */
            if (nev==0) continue; /* 2018-08 continue is bad for make test */
            /* Todo2016: tridiag enough for degeneracy too? */
#endif
            if (checkpoint<9 || iv>1) {	/* v1.7.14=5, v2.44=6 */
              if(nev){ randvec(v1,startvec,iv-1+(startvec&8)*nruns);
                       incv1(nev,v0,v1,ev,itr,hbuf); }          /* EV(H) */
              if ( iv<=1 ) {checkpoint=9;if(!mpi_i)printf(" checkpoint9");}
            } else { // checkpoint9 && iv==1 = load_ev
              if (!mpi_i) printf("\n checkpoint9=%d=reload_ev,",checkpoint);
              if(nev>0)loadvec(ev[0],FileGS,iv,-1);
            } /* reload j? chkpt */
          } else if((algo&7)==1){ /* a1 = 1-step-lanzcos */
            randvec(v0,   startvec,iv-1+(startvec&8)*nruns);
            itr=lanz1(v0,v1,ev[0],hbuf);   /* 1-step-lanzcos EW,EV(H) */
          } else if((algo&7)==2 && nev>0){ /* a2 = 2x2 diag, only needs ev[0] */
            randvec(ev[0],startvec,iv-1+(startvec&8)*nruns);
            itr=lanz2(ev[0],hbuf);         /* 2x2-Matrix     EW,EV(H) */
          } else if((algo&7)==3 && nev>0){ /* a3 load_EV from disk */
            printf("\n algo=3: load vectors from disk");
            /* %NEV to avoid warning "array access above bounds" for NEV=1 */
            if (loadvec(ev[0],FileGS,iv,-1)) continue; /* v1.7.13+ */
            if (nev>1)
            if (loadvec(ev[1%NEV],FileEX,iv,-1)) continue; /* v2.28 */
          } /* <O> zu H' */
          else if((algo&7)==4){
             if (fulldiag(ev,nev,hbuf[0],ew)) break; }/* a4, 2016-11 +break */
          /* 2016-11 a4 malloc fail gives no output erg= because break */
          if (!mpi_i) {
            printf("\n erg= "); outew(stdout,ew,NEW);
            if (vvv) fprintf(stdout," t+=%s",sclock(~0l));
            printf("\n E0/N= %12.8f E0/nw= %12.8f",D(ew[0]/nn),D(ew[0]/nw)); /* Feb99 */
            if (ne0)
            printf("\n E%d/N= %12.8f E%d/nw= %12.8f E%d= %12.8f",ne0,D(ew[ne0]/nn),ne0,D(ew[ne0]/nw),ne0,D(ew[ne0])); /* Feb99 */
          }
          MPI_Barrier(MPI_COMM_WORLD);

          if ((algo&7)==4 || (!nev))break; /* brk here if noEV or a4-diag */
          /* continue with EV-handling and degeneracy loop */
          
          if (v0) /* not for a2 2013-01 */
          for(ii=0,j=1;j<iv;j++){ /* orthogonalize ev against saved vectors */
            loadvec(v0,FileGS,j,-1);
            if ((ii=iortho(ev[0],v0))==1) break;
          } /* stop if we have found all orthogonal ev's */
          if (ii!=0) break;
          
#ifdef CONFIG_Fidelity  /* better check for option? ToDo: test value */
          /* 1 - <\Psi(a-da)|\Psi(a)> (or 1-|<p1|p2>| because of sign) */
          /* try this with J1-J2-model, j2= 0.09,0.1,0.19,0.2 */
          if (v0) { double r1;
            /* JS-2011-03-25 Fidelity = Overlap to other GS vector */
            loadvec(v0,FileGS,iv,-1);  /* tmp/d%d.vec  */
            /* if (!mpi_i) printf("\n old vector:"); x_out(stdout,v0,0,x_p); */
            /* if (!mpi_i) printf("\n new vector:"); x_out(stdout,ev[0],0,x_p); */
            /* older versions work because complex.h+fabs(cplx) was used
             * instead of the (faster) internal complex type */
            /* 2018-11-06 replace fabs by cabs=hypot(z) for cplx */
            r1 = 1 - Cabs( sprod(v0, ev[0]) );
            savevec(ev[0],FileGS,iv,-1);
            if (iv>1 && !mpi_i)
              printf("WARNING: ea>1 + FIDELITY does not work!\n");
            if (!mpi_i)
              printf("\n fidelity= %12.9f (1-overlap to last computed EV)\n",
                     r1);
          }
#endif
          /* we only save if its necessary (very slow) */
          /* %NEV to avoid warning "array access above bounds" for NEV=1 */
          if (max_ea>1)   savevec(ev[0    ],FileGS,iv,-1);
          if (nev>1) {    savevec(ev[1%NEV],FileEX,iv,-1); } /* nev<=NEV! */
          /* 0,1,2 */
          if((algo&7) <= 4)
            for(j=0;j<nev;j++){
              if (!mpi_i) printf("\n\nstart chk3.ev%d (nev=%d NEV=%d) ...",
                                 j, nev, NEV);
              chk3(ev[j],v1,iv,hbuf);
              if((algo&7)==3) break; /* ev0 only */
            }

#ifdef CONFIG_DIMER_CHECK
          if (nev>2) { /* Richter extension to check for dimers */
            /* we break the symmetry by adding lowest two states */
            printf("\n dimer-check+ (ev0+ev1)/2:");
            clrvec(ev[2]);
            add_av(ev[2],+sqrt(0.5),ev[0]);
            add_av(ev[2],+sqrt(0.5),ev[1]);
            normvec(ev[2]); // not needed?  
            chk3(ev[2],v1,1,hbuf);
            printf("\n dimer-check- (ev0-ev1)/2:");
            clrvec(ev[2]);
            add_av(ev[2],+sqrt(0.5),ev[0]);
            add_av(ev[2],-sqrt(0.5),ev[1]);
            normvec(ev[2]);
            chk3(ev[2],v1,1,hbuf);
          }
#endif
        } /* loop: iv=1...max_ea */
        if (max_ea>1 && !mpi_i)
          printf("\n degree_of_degeneracy=%d<%d\n",iv-1,max_ea);
      } /* a0 ============ old-good-ns+storeh-method ========= */
 /* label_end: */

     // ?? pthread_barrier_wait(&pt_barrier); /* 2017-01-23 output */
      usleep(10000);
      MPI_Barrier(MPI_COMM_WORLD);

      /* 2015-04-08 free Hbuffer */
      for (i=0;i<pt_n;i++){
        if (memH[i].ptr) free(memH[i].ptr);
        memH[i].ptr=NULL; /* 2018-11 */
        memH[i].len=0;
      }
      
// norm2=(ATnorm2 *)myfree((void *) norm2, &m_n2);base2ok=0;
      l2b=    (ULong *)myfree((void *) l2b, &m_l2);
      if (l1) { // ??? 2015-02-26, change fl1 to l1 2016-03-22 fix memleak
        fl1=fopen(FILE_l10,"r+b"); /* 2016-03 added to fix reenable myunmap */
        if(algo&8) myunmap(l1,fl1,n1max*(off_t)Bsize,0);
        else       myunmap(l1,fl1,n1   *(off_t)Bsize,0);
      } // ???
      if (fl1) fclose(fl1);
      l1=NULL; fl1=NULL;
      {
        int vvv2=vvv; if (pt_n*mpi_n>4) vvv&=~2; /* avoid to much output */
        if ((!mpi_i) && (!ev[0])) printf(" free ev");
        for(i=0;i<nev;i++) {
          ev[i]=(scplx *)myfree((void *)ev[i],&m_ev[i]);
        }
        if ((!mpi_i) && (!v0)) printf(" free v0,v1");
        v1=(scplx *)myfree((void *)v1,&m_v1);
        v0=(scplx *)myfree((void *)v0,&m_v0);
        vvv=vvv2;
      }
      n1=0l; /* reset n1 */
      /* reset checkpoint, next start will do new computation */
      // echo 0 > tmp/chkpt
      if (!mpi_i) 
        printf(" reset checkpoint %d to 0 for next cycle", checkpoint);
      checkpoint=0;
      nruns++;  /* can be used for other srand on next computation */
      if(!mpi_i){FILE *f;f=fopen(FCHKPT,"wt");fputs("0\n",f);fclose(f);}
      
      // --- JS2012-05 rausgenommen, nur laestige outputs ohne viel Sinn
      //MPI_Barrier(MPI_COMM_WORLD);
      //if (!mpi_i) printf("\n --- next run + mpi-balance-check:\n");
      //MPI_Barrier(MPI_COMM_WORLD);
      // if (mpi_i<8 || mpi_i>=mpi_n-8) // first and last 8
      //   printf(" %d", mpi_i);
      //MPI_Barrier(MPI_COMM_WORLD);
      if (!mpi_i) printf("\n");  /* flush */

      /* 2017-09 moved inside a0 because MIPS alloc_hbuf depend from sym */
      for (block=0; block<pt_n; block++) free_hbuf(hbuf[block]); /* 2014-07-17 */
    } /* end a#-algo-command if(s1)-lines */
    else {
      if (!mpi_i)
        E2("infile L%d: unknown parameter or command: %s", in1line, s1);
    }

    if (err_fatal){
      err_fatal=0; 
      if ((mpi_i&(mpi_i-1))==0 && mpi_i<32) /* 2015-02-26 log2 output */ 
        printf("\n reset err_fatal (mpi=%d)\n", mpi_i);
    }

  } /* for(;;) scan input lines from daten.i */ 
  if (!mpi_i) printf(" close(%s)", name_file_i);
  MPI_Barrier(MPI_COMM_WORLD);
  if (in1) { fclose(in1); in1=NULL; }
  if (!mpi_i) printf(" ok\n");
  b_free(); /* release dynamic arrays b_ofs[], b_len[] block.c dummy */
  MPI_Barrier(MPI_COMM_WORLD); /* sync output of threads 2012-05 */

#ifdef CONFIG_PTHREAD
#ifdef _OPENMP
#pragma omp critical
{ for (i=0;i<pt_n;i++) main_pt[i].task=-1;    /* finalize */ }
#else /* OPENMP */
  /* send no further tasks to threads */
  pthread_mutex_lock(&pt_mutex1);
  for (i=0;i<pt_n;i++) main_pt[i].task=-1;    /* finalize */
  pthread_cond_broadcast(&pt_cond1);
  pthread_mutex_unlock(&pt_mutex1);
  for (i=0;i<pt_n;i++) {          /* join main threads */
    pthread_join(pt[i],NULL);
  }
#endif /* ~OPENMP */
} /* pragma omp master|single test at L4539,4547 2017-01-09 */
} /* pragma omp parallel */
#endif /* PTHREAD */
  sub_end();	/* kill child, dummy pipe.c */
#if defined(CONFIG_MPI)
  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Finalize();
#endif
  return(0);
}

#ifdef F2C
/* lapack-3.0 for linux needs f2c and f2c needs MAIN__() */
/* if you know a better solution, please tell me */
#if (VecType & 4) || defined(__cplusplus)
extern "C"
#endif
int MAIN__(){ return main(0,NULL); }
#endif
